/*************************************************************************
 * The contents of this file are subject to the MYRICOM ABSTRACTION      *
 * LAYER (MAL) SOFTWARE AND DOCUMENTATION LICENSE (the "License").       *
 * User may not use this file except in compliance with the License.     *
 * The full text of the License can found in LICENSE.TXT                 *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2012 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

/*
 * These are mal_ utility definitions that do not need tons of other
 * includes like mal.h does.
 */

#ifndef _mal_utils_h_
#define _mal_utils_h_

#ifdef __WORDSIZE
#define MAL_WORDSIZE __WORDSIZE
#elif defined(__x86_64) || defined(_M_IA64) || defined(_M_AMD64)
#define MAL_WORDSIZE 64
#else
#define MAL_WORDSIZE 32
#endif

#if MAL_WORDSIZE == 64
#define MAL_PIOWORDSIZE 8
typedef uint64_t mal_pioword_t;
#else
#define MAL_PIOWORDSIZE 4
typedef uint32_t mal_pioword_t;
#endif

typedef uintptr_t mal_wcsum;

uint32_t mal_piocopy_csum(uintptr_t *piobuf, const uint8_t *src,
                          uint32_t len, uint32_t *offp);
MAL_FUNC(int) mal_get_macaddr(const char *ifname, uint8_t mac_addr[6]);
MAL_FUNC(uint16_t) mal_get_vlan_tag(const char *ifname);

#if !MAL_OS_WINDOWS
#define add_carry32(sum, v)  __asm__("addl %1, %0 ;"                   \
                                     "adcl $0, %0 ;"                   \
                                     : "=r" (sum)                      \
                                     : "g" ((uint32_t) v), "0" (sum))

#define add_carry64(sum, v)  __asm__("addq %1, %0 ;"                   \
                                     "adcq $0, %0 ;"                   \
                                     : "=r" (sum)                      \
                                     : "g" ((uint64_t) v), "0" (sum))
#else
#define add_carry32(sum, v) add__carry32(&sum, v)
#define add_carry64(sum, v) add__carry64(&sum, v)

static inline
void add__carry32(uint32_t *sum, uint32_t v)
{
  *sum += v;
  if (*sum < v) {
    (*sum)++;
  }
}

static inline
void add__carry64(uint64_t *sum, uint64_t v)
{
  *sum += v;
  if (*sum < v) {
    (*sum)++;
  }
}
#endif

static inline
uint32_t
do_add_carry32(uint32_t csum, uint32_t val)
{
  add_carry32(csum, val);
  return csum;
}

#endif /* _mal_utils_h_ */
