/*
 * common test stuff
 */

#ifdef _WIN32
#include <winsock2.h> 
#include <process.h>
#endif
#include "maltest_common.h"

char *Recv_mode_string[] = {
  "poll",
  "epoll"
};
char Recv_mode_supported[NUM_RECV_MODE];
sock_recv_mode_t Recv_mode_dflt = NUM_RECV_MODE;

static arg_t *master_list, *product_list, *user_list, *lp;
static char master_string[MAX_ARGS] = {'\0'};
static char product_string[MAX_ARGS] = {'\0'};
static char user_string[MAX_ARGS] = {'\0'};
static char set_string[MAX_ARGS] = {'\0'};
static char *sp;
static int *ip;
static int master_cnt = 0;
static int product_cnt = 0;
static int user_cnt = 0;
static int set_cnt = 0;
static char opt_string[2*MAX_ARGS] = {':', '\0'};
static int opt_cnt = 1;
static char *test_desc;
static struct option longopts[MAX_ARGS];
static saved_t saved_list[MAX_ARGS];
static int num_saved = 0;

#ifdef _WIN32
struct thread_ctx
{
  int timeout;
  void (*timesup)(int);
};

unsigned __stdcall timeout_helper(void *p)
{
  struct thread_ctx *ctx;

  ctx = p;
  Sleep(1000 * ctx->timeout);
  ctx->timesup(0);
  return 0;
}
#endif

int
setup_timeout(int timeout, void (*timesup)(int))
{
#ifndef _WIN32
  struct itimerval it;
  struct sigaction sa;
  int rc;

  if (timeout > 0) {
    memset(&it, 0, sizeof(it));
    it.it_value.tv_sec = timeout;
    rc = setitimer(ITIMER_REAL, &it, NULL);
    if (rc != 0) {
      perror("setitimer");
      return rc;
    }
    memset(&sa, 0, sizeof(sa));
    sa.sa_handler = timesup;
    rc = sigaction(SIGALRM, &sa, NULL);
    if (rc != 0) {
      perror("sigaction");
      return rc;
    }
  }
  return 0;
#else
  static struct thread_ctx ctx;
  unsigned thrdaddr;

  if (timeout > 0) {
    ctx.timeout = timeout;
    ctx.timesup = timesup;
    _beginthreadex(NULL, 0, timeout_helper, &ctx, 0, &thrdaddr);
  }
  return 0;
#endif
}

/* Common function pointers */
void set_isserver(testargs_t *args, char *str, char c){ args->isserver = (str == NULL); }
void set_remote_addr(testargs_t *args, char *str, char c){ args->remote_addr = str; }
void set_port(testargs_t *args, char *str, char c){
  if (atoi(str) >= 1024 && atoi(str) < 65536) args->port = atoi(str);
  else { 
    fprintf(stderr, "Error: Port %d is invalid.\n", atoi(str)); 
    exit(1); 
  }
} 
void set_extra_remotes(testargs_t *args, char *str, char c) {
  char *p;
  int i;
  if (!strcmp(str, MALTEST_PARSE_EXTRA_ADDRS_S)) {
    for (i = 0; i < MALTEST_PARSE_EXTRA_ADDRS; i++) {
      args->extra_remote_addrs[i] = NULL;
      args->extra_remote_ports[i] = -1;
    }
    args->extra_remote_num = 0;
    return;
  }
  if (args->extra_remote_num >= MALTEST_PARSE_EXTRA_ADDRS) return;
  args->extra_remote_addrs[args->extra_remote_num] = str;
  p = strchr(str, ':');
  if (p != NULL) {
    *p = '\0';
    args->extra_remote_ports[args->extra_remote_num] = atoi(p+1);
  }
  args->extra_remote_num++;  
}
void set_join_addr(testargs_t *args, char *str, char c){ args->join_addr = str; }
void set_local_addr(testargs_t *args, char *str, char c){ args->local_addr = str; }
void set_sock_local_addr(testargs_t *args, char *str, char c){ args->sock_local_addr = str; }
void set_extra_locals(testargs_t *args, char *str, char c) {
  char *p;
  int i;
  if (!strcmp(str, MALTEST_PARSE_EXTRA_ADDRS_S)) {
    for (i = 0; i < MALTEST_PARSE_EXTRA_ADDRS; i++) {
      args->extra_local_addrs[i] = NULL;
      args->extra_local_ports[i] = -1;
    }
    args->extra_local_num = 0;
    return;
  }
  if (args->extra_local_num >= MALTEST_PARSE_EXTRA_ADDRS) return;
  args->extra_local_addrs[args->extra_local_num] = str;
  p = strchr(str, ':');
  if (p != NULL) {
    *p = '\0';
    args->extra_local_ports[args->extra_local_num] = atoi(p+1);
  }
  args->extra_local_num++;  
}
void set_localport(testargs_t *args, char *str, char c){ 
  if ((atoi(str) == -1) || (atoi(str) >= 1024 && atoi(str) < 65536)) args->localport = atoi(str);
  else { 
    fprintf(stderr, "Error: Port %d is invalid.\n", atoi(str)); 
    exit(1); 
  }
}
void set_iterations(testargs_t *args, char *str, char c){ args->iterations = atoi(str); }
void set_warmup(testargs_t *args, char *str, char c){ args->warmup = atoi(str); }
void set_delay(testargs_t *args, char *str, char c){ args->delay = atoi(str); }
void set_spintime(testargs_t *args, char *str, char c){ args->spintime = atoi(str); }
void set_timeout(testargs_t *args, char *str, char c){ args->timeout = atoi(str); }
void set_start_size(testargs_t *args, char *str, char c){ args->start_size = atoi(str); }
void set_end_size(testargs_t *args, char *str, char c){ args->end_size = atoi(str); }
void set_check_order(testargs_t *args, char *str, char c){ args->check_order = (str == NULL); }
void set_verify(testargs_t *args, char *str, char c){ args->verify = (str == NULL); }
void set_modulus(testargs_t *args, char *str, char c){ args->modulus = atoi(str); }
void set_num_threads(testargs_t *args, char *str, char c){ args->num_threads = atoi(str); }
void set_num_sockets(testargs_t *args, char *str, char c){ args->num_sockets = atoi(str); }
void set_af(testargs_t *args, char *str, char c){ args->af = (str != NULL) ? MALTEST_DEFAULT_DOMAIN : AF_INET6; }
void set_sock_type(testargs_t *args, char *str, char c) { 
  args->sock_type = ((str != NULL) ? SOCK_STREAM : SOCK_DGRAM);
}
void set_reuse(testargs_t *args, char *str, char c){ args->reuse = (str == NULL); }
void set_recv_mode(testargs_t *args, char *str, char c){ args->recv_mode = (str != NULL) ? MALTEST_RECV_DEFAULT : MALTEST_RECV_BLOCKING; }
void set_check_accel(testargs_t *args, char *str, char c){ args->check_accel = (str == NULL); }
void set_sock_recv_mode(testargs_t *args, char *str, char c) {
  args->sock_recv_mode = atoi(str);
  if (args->sock_recv_mode >= NUM_RECV_MODE || args->sock_recv_mode < 0) {
    fprintf(stderr, "Error: Unknown receive mode\n");
    exit(1);
  } else if (!Recv_mode_supported[args->sock_recv_mode]) {
    fprintf(stderr, "Error: Receive mode '%s' not supported\n", Recv_mode_string[args->sock_recv_mode]);
    exit(1);
  }
}
void set_call_receive(testargs_t *args, char *str, char c){ args->call_receive = (str != NULL); }
void set_loopback(testargs_t *args, char *str, char c){ args->loopback = (str == NULL); }
void set_seed(testargs_t *args, char *str, char c){ args->seed = atoi(str); }
void set_periodic_data(testargs_t *args, char *str, char c){ args->periodic_data = (str == NULL); }
void set_verbose(testargs_t *args, char *str, char c){ args->verbose = (str == NULL); }
void set_feed_unit_id(testargs_t *args, char *str, char c){ args->feed_unit_id = atoi(str); }

/* Product and user function pointers */
void set_value(testargs_t *args, char *str, char c){ save_arg(c, str, 0); }
void set_value_o(testargs_t *args, char *str, char c){ save_arg(c, str, 1); }
void set_true(testargs_t *args, char *str, char c){ save_arg(c, (str == NULL) ? "1" : str, 1); }
void set_false(testargs_t *args, char *str, char c){ save_arg(c, (str == NULL) ? "0" : str, 1); }

void add_args(arg_t args_list[], int num_entries, const char *caller)
{
  int i, pos;
  char *p;

  /* Allocate memory */
  if (master_list == NULL) master_list = (arg_t *) malloc(sizeof(arg_t) * MAX_ARGS);
  if (product_list == NULL) product_list = (arg_t *) malloc(sizeof(arg_t) * MAX_ARGS);
  if (user_list == NULL) user_list = (arg_t *) malloc(sizeof(arg_t) * MAX_ARGS);

  /* Add to the appropriate list */
  if (!strcmp(caller, "user")) {
    lp = user_list;
    sp = user_string;
    ip = &user_cnt;
  } else if (!strcmp(caller, "product")) {
    lp = product_list;
    sp = product_string;
    ip = &product_cnt;
  } else if (!strcmp(caller, "common")) {
    lp = master_list;
    sp = master_string;
    ip = &master_cnt;
  } else {
    fprintf(stderr, "Error: unknown caller to add_args() -- %s\n", caller);
    exit(1);
  }

  for (i = 0; i < num_entries; i++) {
    if (isalnum(args_list[i].option)) {
      sp[*ip] = args_list[i].option;
      lp[(*ip)++] = args_list[i]; 
    } else {
      fprintf(stderr, "Error: option is not alphanumeric -- %c\n", args_list[i].option);
      exit(1);
    }
  }

  /* Merge everything for the last call */
  if (!strcmp(caller, "common")) {

    /* Product and common should not have option collision */
    for (i = 0; i < product_cnt; i++) {
      if (strchr(master_string, product_string[i]) == NULL) {
        if (master_cnt >= MAX_ARGS) {
          fprintf(stderr, "Error: too many args\n");
          exit(1);
        }
        master_string[master_cnt] = product_string[i];
        master_list[master_cnt++] = product_list[i]; 
      } else {
        fprintf(stderr, "Error: common-product option collision -- %c\n", product_string[i]);
        exit(1);
      }
    }

    /* Overwrite with user options */
    for (i = 0; i < user_cnt; i++) {
      p = strchr(master_string, user_string[i]);
      if (p == NULL) {
        if (master_cnt >= MAX_ARGS) {
          fprintf(stderr, "Error: too many args\n");
          exit(1);
        }
        master_string[master_cnt] = user_string[i];
        master_list[master_cnt++] = user_list[i]; 
      } else {
        pos = (int)(p - master_string);
        master_list[pos] = user_list[i];
      }
    }

    /* Create string for use in getopt */
    for (i = 0; i < master_cnt; i++) {
      opt_string[opt_cnt++] = master_string[i];
      if (master_list[i].takes_arg) opt_string[opt_cnt++] = ':';
    }

    /* Add help option */
    opt_string[opt_cnt++] = '?';
  }
}

void add_desc(char *desc) {
  test_desc = desc;
}

void maltest_usage(const char *progname, const char *opts, const char *prod_name)
{
  int i, dflt, env;
  char earg[40];

  fprintf(stderr, "Usage: %s [options]\n", progname);
  fprintf(stderr, "\t-?, --help\n\t\tPrint help information and usage.\n");
  /* Go through master_string instead of opts to control the order of printing */
  for (i = 0; i < master_cnt; i++) {
    if (strchr(opts, master_string[i]) != NULL) {
      fprintf(stderr, "\t-%c, --%s %s\n", master_list[i].option, master_list[i].long_option, master_list[i].desc);
      /* Options with arguments print defaults */
      if (master_list[i].takes_arg) {
        dflt = (master_list[i].dflt_value != NULL 
                && strncmp(master_list[i].dflt_value, "0", 2)
                && master_list[i].dflt_value[0] != '-');
        env = (master_list[i].env != NULL);
        if (env) sprintf(earg, master_list[i].env, prod_name);
        if (dflt && env) {
          fprintf(stderr, "\t\tDefault: %s, or %s if undefined\n", earg, master_list[i].dflt_value);
        } else if (!dflt && env) {
          fprintf(stderr, "\t\tDefault: %s\n", earg);
        } else if (dflt && !env) { 
          fprintf(stderr, "\t\tDefault: %s\n", master_list[i].dflt_value);
        }
      }
    }
  }

  exit(-1);
}

void maltest_usage_verbose(const char *progname, const char *opts, const char *prod_name)
{
  if (test_desc != NULL) fprintf(stderr, "%s\n\n", test_desc);
  maltest_usage(progname, opts, prod_name);
}

void init_longopts() {
  int i;

  /* Allow duplicates since earlier(user) entries have precedence */
  for (i = 0; i < master_cnt; i++) {
    longopts[i].name = master_list[i].long_option;
    longopts[i].has_arg = master_list[i].takes_arg;
    longopts[i].flag = NULL;
    longopts[i].val = master_list[i].option;
  }

  longopts[master_cnt].name = "help";
  longopts[master_cnt].has_arg = 0;
  longopts[master_cnt].flag = NULL;
  longopts[master_cnt].val = '?';
}

int check_common(char c, const char *opts) {
  return (strchr(opts, c) != NULL && strchr(user_string, c) == NULL);
}

void maltest_parseargs(const char *opts, int argc, char **argv,
                       testargs_t *args, const char *prod_name)
{
  char c;
  char *env = NULL, *m_pos = NULL;
  extern char* optarg;
  char earg[40];
  int i, m_index, pos, end_index, long_index;

  arg_t common_list[] = {
    {'s', "server", "\n\t\tRun as server.", 0, "0", NULL, &set_isserver},
    {'h', "remote", "<remote_address>\n\t\tSpecify remote address.", 1, NULL, "%s_REMOTE",
          &set_remote_addr},
    {'p', "port", "<port>\n\t\tUse [remote] port for sending/receiving (1024-65535).", 1,
          MALTEST_DEFAULT_PORT_S, NULL, &set_port},
    {'H', "xremote", "<address:port>\n\t\tSpecify a remote address/interface.\n\t\tOption may be used multiple times, up to a limit.", 1, MALTEST_PARSE_EXTRA_ADDRS_S, NULL, &set_extra_remotes},
    {'j', "mcast", "<mcast_address>\n\t\tSpecify the multicast address to join.", 1, "224.1.1.50", "%s_MCAST",
          &set_join_addr},
    {'l', "local", "<local_address>\n\t\tSpecify address of local product interface.", 1, NULL,
          "%s_LOCAL", &set_local_addr},
    {'a', "local", "<local_address>\n\t\tSpecify address of local socket interface.", 1, "0.0.0.0",
          "%s_LOCAL", &set_sock_local_addr},
    {'P', "localport", "<local_port>\n\t\tUse local port (1024-65535).", 1, "-1", NULL, &set_localport},
    {'L', "xlocal", "<address:port>\n\t\tSpecify a local address/interface.\n\t\tOption may be used multiple times, up to a limit.", 1, MALTEST_PARSE_EXTRA_ADDRS_S, NULL, &set_extra_locals},
    {'i', "iterations", "<iterations>\n\t\tSet number of send/receive iterations.\n\t\tMust be set equally on client and server.", 1, MALTEST_DEFAULT_ITERATIONS_S, "%s_ITERATIONS", &set_iterations},
    {'w', "warmup", "<warmup_count>\n\t\tSet number of warmup sends to do from client before starting main loop.", 1, MALTEST_DEFAULT_WARMUP_S, NULL, &set_warmup},
    {'g', "delay", "<usecs>\n\t\tSet delay.", 1, MALTEST_DEFAULT_DELAY_S, NULL, &set_delay},
    {'T', "spintime", "<usecs>\n\t\tSpecify time in usecs to wait between sends.", 1,
          MALTEST_DEFAULT_SPINTIME_S, NULL, &set_spintime},
    {'t', "timeout", "<seconds>\n\t\tSet time to run before terminating application.", 1, "0", NULL, 
          &set_timeout},
    {'S', "size", "<size>\n\t\tSet size of packet in bytes to send/receive (min).", 1, MALTEST_DEFAULT_START_SIZE_S,
          NULL, &set_start_size},
    {'E', "endsize", "<size>\n\t\tSet size of packet in bytes to send/receive (max).", 1,
          MALTEST_DEFAULT_END_SIZE_S, NULL, &set_end_size},
    {'o', "order", "\n\t\tCheck packet order.", 0, "0", NULL, &set_check_order},
    {'v', "verify", "\n\t\tVerify packets.", 0, "0", NULL, &set_verify},
    {'c', "accel", "\n\t\tCheck whether sockets are appropriately accelerated.", 0, "0", NULL,
          &set_check_accel},
    {'m', "modulus", "<num_packets>\n\t\tPrint status after some number of packets has been received.", 1,           MALTEST_DEFAULT_MODULUS_S, NULL, &set_modulus},
    {'n', "threads", "<num_threads>\n\t\tSpecify number of threads.", 1, MALTEST_DEFAULT_THREADS_S, NULL,
          &set_num_threads},
    {'k', "sockets", "<num_sockets>\n\t\tSpecify number of sockets.", 1, MALTEST_DEFAULT_SOCKETS_S, NULL,
          &set_num_sockets},
    {'6', "ipv6", "\n\t\tSet socket family to IPv6.", 0, "IPv4", NULL, &set_af},
    {'u', "udp", "\n\t\tSet socket type to UDP.\n\t\tMust be set on both client and server.", 0, "TCP", 
          NULL, &set_sock_type},
    {'r', "reuse", "\n\t\tReuse address.", 0, "0", NULL, &set_reuse},
    {'b', "blocking", "\n\t\tUse blocking receive.", 0, "0", NULL, &set_recv_mode},
    {'R', "recvmode", "<recv_mode>\n\t\tSelect receive mode.\n\t\t0 - Poll [default]\n\t\t1 - Epoll", 
          1, "0", NULL, &set_sock_recv_mode},
    {'x', "dead", "\n\t\tDo not call receive.", 0, "0", NULL,
          &set_call_receive},
    {'O', "loopback", "\n\t\tSet loopback.", 0, "0", NULL, &set_loopback},
    {'e', "seed", "<seed>\n\t\tUse seed for random numbers.", 1, "0", NULL, &set_seed},
    {'D', "periodic", "\n\t\tPrint periodic data.", 0, "0", NULL, &set_periodic_data},
    {'V', "verbose", "\n\t\tPrint verbose output for debugging.", 0, "0", NULL, &set_verbose},
    {'F', "unit", "\n\t\tID of a feed unit.", 1, "0", NULL, &set_feed_unit_id}
  };

  add_args(common_list, ARRAY_LENGTH(common_list), "common");

  init_longopts();

  /* Set defaults and env variables */
  Recv_mode_supported[RECV_MODE_POLL] = 1;

  for (i = 0; i < master_cnt; i++) {
    /* set_func sets appropriate default in case of empty string */
    master_list[i].set_func(args, master_list[i].dflt_value, master_list[i].option);
    if (master_list[i].env != NULL) {
      sprintf(earg, master_list[i].env, prod_name);
      env = getenv(earg);
      if (env != NULL) master_list[i].set_func(args, env, master_list[i].option);
    }
  }

  while ((c = getopt_long(argc, argv, opt_string, longopts, &long_index)) != EOF) {
    /* Option does not exist in opt_string */
    if (c == '?' && optopt != 0) {
      fprintf(stderr, "%s: unknown option -- %c\n", argv[0], optopt);
      maltest_usage(argv[0], opts, prod_name);
    /* User asking for help */
    } else if (c == '?') {
      maltest_usage_verbose(argv[0], opts, prod_name);
    /* Found option was not requested */
    } else if (!strchr(opts, c) && optopt == 0) {
      fprintf(stderr, "%s: invalid option -- %c\n", argv[0], c);
      maltest_usage(argv[0], opts, prod_name);
    /* Missing argument for option */
    } else if (c == ':') {
      fprintf(stderr, "%s: missing argument for -%c\n", argv[0], optopt);
      maltest_usage(argv[0], opts, prod_name);
    }

    /* Check argument syntax */
    if (optarg != NULL) {
      pos = strcspn(optarg, ":");
      if (pos == 0) {
        fprintf(stderr, "%s: argument cannot begin with ':'\n", argv[0]);
        maltest_usage(argv[0], opts, prod_name);
      }

      pos = strcspn(optarg, "=");
      if (pos == 0) optarg++;

      pos = strcspn(optarg, "@#%^*()~[]{}?");
      end_index = strlen(optarg);
      if (pos != end_index) {
        fprintf(stderr, "%s: invalid symbol in argument -- %c\n", argv[0], c);
        maltest_usage(argv[0], opts, prod_name);
      }
    }

    /* Find arg_entry and set function */
    if ((m_pos = strchr(master_string, c)) != NULL) {
      m_index = (int)(m_pos - master_string);
      master_list[m_index].set_func(args, optarg, c);
      set_string[set_cnt++] = c;
    } else {
      fprintf(stderr, "%s: requested option %c was not defined\n", argv[0], c);
      exit(1);
    }
  }

  /* Check for required input */
  if (check_common('h', opts)) {
    if ((args->remote_addr == NULL) && (args->isserver == 0)) {
      fprintf(stderr, "Error: Must specify remote address or set %s_REMOTE.\n", prod_name);
      maltest_usage(argv[0], opts, prod_name);
    } else if (args->remote_addr != NULL) {
      args->remote_sin.sin_family = AF_INET;
      args->remote_sin.sin_port = htons(args->port);
      if (inet_aton(args->remote_addr, &args->remote_sin.sin_addr) == 0) {
        fprintf(stderr, "Error: Remote address %s is not valid.\n", args->remote_addr);
        exit(1);
      }
    }
  }

  if (check_common('H', opts)) {
    for (i = 0; i < args->extra_remote_num; i++) {
      if (args->extra_remote_ports[i] == -1)
        args->extra_remote_ports[i] = args->port;
    }
  }

  if (check_common('j', opts)) {
    if (args->join_addr != NULL) {
      args->join_sin.sin_family = AF_INET;
      args->join_sin.sin_port = htons(args->port);
      if (inet_aton(args->join_addr, &args->join_sin.sin_addr) == 0) {
        fprintf(stderr, "Error: Multicast address %s is not valid.\n", args->join_addr);
        exit(1);
      }
      args->mcast = 1;
    } else 
      args->mcast = 0;
  }

  /* args->port is always set to a default */
  if (args->localport == -1) args->localport = args->port;

  if (check_common('l', opts)) {
    if (args->local_addr == NULL) {
      fprintf(stderr, "Error: Must specify local address or set %s_LOCAL.\n", prod_name);
      maltest_usage(argv[0], opts, prod_name);
    }
    args->local_sin.sin_family = AF_INET;
    args->local_sin.sin_port = htons(args->localport);
    if (inet_aton(args->local_addr, &args->local_sin.sin_addr) == 0) {
      fprintf(stderr, "Error: Local address %s is not valid.\n", args->local_addr);
      exit(1);
    }
  }

  if (check_common('L', opts)) {
    for (i = 0; i < args->extra_local_num; i++) {
      if (args->extra_local_ports[i] == -1)
        args->extra_local_ports[i] = args->localport;
    }
  }

  if (optind < argc) {
    fprintf(stderr, "%s: excess number (%d) of arguments.\n", argv[0], argc - optind);
    maltest_usage(argv[0], opts, prod_name);
  }
}

/* Prints alpha-numeric sequence of characters that are not in master string */
void get_leftovers() {
  int c, index = 0;
  char leftovers[64] = {'\0'};

  for (c = 48; c < 123; c++) {
    if ((c > 57 && c < 65) || (c > 90 && c < 97)) continue;
    if (!strchr(master_string, c)) {
      leftovers[index++] = c;;
    }
  }

  printf("Available options: %s\n", leftovers);
}

/* Save non-common args in a saved_t array */
void save_arg(char c, char *str, int overwrite) {
  int i;

  if (c == 0) return;

  if (overwrite) {
    for (i = 0; i < num_saved; i++) {
      if (saved_list[i].option == c) {
        saved_list[i].optarg = str;
        return;
      }
    }
  }

  /* If not found, save the arg */
  if (num_saved < MAX_ARGS) {
    saved_list[num_saved].option = c;
    saved_list[num_saved].optarg = str;
    num_saved++;
  } else {
    fprintf(stderr, "Error: excess number of saved arguments (%d).\n", MAX_ARGS);
    exit(1);
  }
}

/* Get the saved arg */
char *load_arg(char c, int index) {
  int i, c_index;

  if (index < 0) index = 0;

  c_index = -1;
  for (i = 0; i < num_saved; i++) {
    if (saved_list[i].option == c) {
      c_index++;
      if (c_index == index) 
        return saved_list[i].optarg;
    }
  } 

  fprintf(stderr, "Warning: No saved argument found (index: %d, max: %d) -- %c\n", index, c_index, c);

  return NULL;
}

/* Return number of times the option was specified */
int num_set(char c) {
  int cnt = 0;
  char *p;

  p = strchr(set_string, c);
  while (p != NULL) {
    cnt++;
    p = strchr(p + 1, c);
  }

  return cnt;
}

void
fail_th(int rc, char *s)
{
#ifndef _WIN32
  errno = rc;
  perror(s);
#else
  LPVOID lpMsgBuf;
  LPVOID lpDisplayBuf;

  FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        rc,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );


  printf ("%s: %s\n", s, lpMsgBuf);
  LocalFree(lpMsgBuf);

#endif
}

void
fail(int rc, char *s)
{
  fail_th(rc, s);
  exit(1);
}

void
timesup(
  int sig)
{
  printf("Exiting due to timeout\n");

  exit(0);
}

void
timesup_err(
  int sig)
{
  printf("Exiting with error due to timeout\n");

  exit(1);
}

char *
inadport_decimal(sad)
	struct sockaddr_in *sad;
{
	static char buf[32];
	int a;

	a = ntohl(0xffffffff & sad->sin_addr.s_addr);
	sprintf(buf, "%d.%d.%d.%d:%d",
			0xff & (a >> 24),
			0xff & (a >> 16),
			0xff & (a >> 8),
			0xff & a,
			0xffff & (int)ntohs(sad->sin_port));
	return buf;
}

int64_t
time_ns(void)
{
  struct timeval tv;
  int64_t usec;
  gettimeofday(&tv, NULL);
  usec = (int64_t) tv.tv_sec * 1000000;
  usec += (int64_t) tv.tv_usec;
  return usec;
}

#ifndef _WIN32
int
mal_init_sockets()
{
  return 0;
}
void
mal_finish_sockets()
{
}

#else	/* _WIN32 */

/*
 * Windows socket functions
 */
int

mal_init_sockets()
{
  WORD version;
  WSADATA data;
  int ret;

  version = (MAKEWORD(2, 2));
  ret = WSAStartup(version, &data);

  if (ret != 0) {
    ret = WSAGetLastError();
    if (ret == WSANOTINITIALISED) {
      printf("not initialised");
      return -1;
    }
  }

  return 0;
}

void
mal_finish_sockets()
{

  WSACleanup();

}

#if (_WIN32_WINNT < 0x0600)
int
mal_recv_all(
  mal_socket_t sock,
  char *buf,
  int len)
{
  int ret;
  int left;
  int rx;

  left = len;
  rx = 0;
  while (left > 0) {
    ret = recv(sock, buf+rx, left, 0);
    if (ret == -1) {
     if (WSAGetLastError() == WSAEINTR) continue;
     return ret;
    }

    rx += ret;
    left -= ret;
  }

  return rx;
}
#endif  /* (_WIN32_WINNT < 0x0600) */
#endif

enum sk_accel_mode
mal_accel_mode(
  int s)
{
  int amode;
  socklen_t len;
  int rc;

  amode = 0;
  len = sizeof(amode);
  rc = getsockopt(s, MALPROTO_IP, MAL_ACCELMODE, (char *) &amode, &len);
  if (rc == -1) {
    return SK_AMODE_NONE;
  } else {
    return amode;
  }
}

int
mal_is_accel(
  int s)
{
  return (mal_accel_mode(s) != SK_AMODE_NONE);
}

#ifndef __linux__
int mal_epoll_create(
  int n)
{
  fprintf(stderr, "mal_epoll_create not supported\n");
  exit(1);
  return -1;
}

int 
mal_epoll_ctl(
  int epfd,
  int op,
  int fd,
  struct mal_epoll_event *event)
{
  fprintf(stderr, "mal_epoll_ctl not supported\n");
  exit(1);
  return -1;
}

int 
mal_epoll_wait(
  int epfd,
  struct mal_epoll_event *events,
  int maxev,
  int timeout)
{
  fprintf(stderr, "No mal_epoll_wait available\n"); \
  exit(1);
}
#endif
