#ifndef _MAL_TEST_COMMON_H
#define _MAL_TEST_COMMON_H

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#ifndef _WIN32
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <poll.h>
#include <signal.h>
#include <getopt.h>
#include <assert.h>
#include <netdb.h>
#include <pthread.h>
#include <poll.h>
#ifdef __linux__
#include <linux/netdevice.h>
#include <sys/epoll.h>
#endif
#else

typedef signed __int8      int8_t;
typedef signed __int16    int16_t;
typedef signed __int32    int32_t;
typedef signed __int64    int64_t;
typedef unsigned __int8   uint8_t;
typedef unsigned __int16 uint16_t;
typedef unsigned __int32 uint32_t;
typedef unsigned __int64 uint64_t;
#ifdef _M_IX86
typedef unsigned __int32 uintptr_t;
#else
typedef unsigned __int64 uintptr_t;
#endif

#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include "getopt.h"
#define sleep(x_) Sleep((x_) * 1000)
#define close(x_) closesocket(x_)
#define usleep(x_) Sleep(x_/1000)
#ifndef ECONNREFUSED
#define ECONNREFUSED WSAECONNREFUSED
#endif
#ifndef ETIMEDOUT
#define ETIMEDOUT WSAETIMEDOUT
#endif
#ifndef EADDRNOTAVAIL
#define EADDRNOTAVAIL WSAEADDRNOTAVAIL
#endif
#ifndef EADDRINUSE
#define EADDRINUSE WSAEADDRINUSE
#endif
#define random rand
#define srandom srand

#endif

#include "mal.h"
#define MX_THREAD_SAFE 1
#include "mal_thread.h"

#define MALPROTO_IP 42
#define MAL_ACCELMODE 1
enum sk_accel_mode {
  SK_AMODE_NONE,
  SK_AMODE_MYRIUDP,
  SK_AMODE_BSDUDP,
  SK_AMODE_BSDTCP
};

int setup_timeout(int timeout, void (*timesup)(int));

#define MALTEST_PARSE_EXTRA_ADDRS 10
#define MALTEST_DEFAULT_PORT 3333
#define MALTEST_DEFAULT_ITERATIONS 10000
#define MALTEST_DEFAULT_WARMUP 10000
#define MALTEST_DEFAULT_MODULUS 100000
#define MALTEST_DEFAULT_START_SIZE 1
#define MALTEST_DEFAULT_END_SIZE 2048
#define MALTEST_DEFAULT_DELAY 0
#define MALTEST_DEFAULT_SPINTIME 1
#define MALTEST_DEFAULT_DOMAIN AF_INET
#define MALTEST_DEFAULT_THREADS 4
#define MALTEST_DEFAULT_SOCKETS 4
#define MALTEST_DEFAULT_WARMUP 10000

#define MALTEST_PARSE_EXTRA_ADDRS_S "10"
#define MALTEST_DEFAULT_PORT_S "3333"
#define MALTEST_DEFAULT_ITERATIONS_S "10000"
#define MALTEST_DEFAULT_WARMUP_S "10000"
#define MALTEST_DEFAULT_MODULUS_S "100000"
#define MALTEST_DEFAULT_START_SIZE_S "1"
#define MALTEST_DEFAULT_END_SIZE_S "2048"
#define MALTEST_DEFAULT_DELAY_S "0"
#define MALTEST_DEFAULT_SPINTIME_S "1"
#define MALTEST_DEFAULT_DOMAIN_S "AF_INET"
#define MALTEST_DEFAULT_THREADS_S "4"
#define MALTEST_DEFAULT_SOCKETS_S "4"
#define MALTEST_DEFAULT_WARMUP_S "10000"

#define TCP_DEFAULT_TOTAL_CONNS 2000
#define TCP_DEFAULT_OPEN_CONNS 64
#define TCP_DEFAULT_CLOSE_SEC 10

typedef enum {
  RECV_MODE_POLL,
  RECV_MODE_EPOLL,
  NUM_RECV_MODE
} sock_recv_mode_t;
extern char Recv_mode_supported[];
extern char *Recv_mode_string[];
extern sock_recv_mode_t Recv_mode_dflt;

typedef enum {
  DBL_RECV_MODE_DEFAULT,
  DBL_RECV_MODE_IBLOCK,
  DBL_RECV_MODE_POLL,
  DBL_RECV_MODE_NBLOCK,
  DBL_RECV_MODE_EBLOCK,
  NUM_DBL_RECV_MODE
} dbl_recv_mode_t;

enum maltest_recv_mode {
  MALTEST_RECV_DEFAULT = 0,
  MALTEST_RECV_BLOCKING = 1,
};

typedef struct testargs {
  int isserver;
  char *remote_addr;
  int port;
  char *extra_remote_addrs[MALTEST_PARSE_EXTRA_ADDRS];
  int extra_remote_ports[MALTEST_PARSE_EXTRA_ADDRS];
  int extra_remote_num;
  char *join_addr;
  int mcast;
  char *local_addr;
  char *sock_local_addr;
  int localport;
  char *extra_local_addrs[MALTEST_PARSE_EXTRA_ADDRS];
  int extra_local_ports[MALTEST_PARSE_EXTRA_ADDRS];
  int extra_local_num;
  struct sockaddr_in remote_sin;
  struct sockaddr_in join_sin;
  struct sockaddr_in local_sin;
  int iterations;
  int warmup; 
  int delay;
  int spintime;
  int timeout;
  int start_size;
  int end_size;
  int check_order;
  int verify;
  int check_accel;
  int modulus;
  int num_threads;
  int num_sockets;
  unsigned short af;
  int sock_type; 
  int reuse;
  int recv_mode;
  sock_recv_mode_t sock_recv_mode;
  int call_receive;
  unsigned char loopback;
  int seed; 
  int periodic_data;
  int verbose;
  int feed_unit_id;
} testargs_t;

typedef struct {
  char option;
  char *long_option;
  char *desc;
  int takes_arg;
  char *dflt_value;
  char *env;
  void (*set_func)(testargs_t *args, char *str, char c);
} arg_t;

typedef struct {
  char option;
  char *optarg;
} saved_t;

void add_args(arg_t args_list[], int num_entries, const char *caller);
void add_desc(char *desc);

void set_isserver(testargs_t *args, char *str, char c);
void set_remote_addr(testargs_t *args, char *str, char c);
void set_extra_remotes(testargs_t *args, char *str, char c);
void set_port(testargs_t *args, char *str, char c);
void set_join_addr(testargs_t *args, char *str, char c);
void set_local_addr(testargs_t *args, char *str, char c);
void set_sock_local_addr(testargs_t *args, char *str, char c);
void set_extra_locals(testargs_t *args, char *str, char c);
void set_localport(testargs_t *args, char *str, char c);
void set_iterations(testargs_t *args, char *str, char c);
void set_warmup(testargs_t *args, char *str, char c);
void set_delay(testargs_t *args, char *str, char c);
void set_spintime(testargs_t *args, char *str, char c);
void set_timeout(testargs_t *args, char *str, char c);
void set_start_size(testargs_t *args, char *str, char c);
void set_end_size(testargs_t *args, char *str, char c);
void set_check_order(testargs_t *args, char *str, char c);
void set_verify(testargs_t *args, char *str, char c);
void set_accelcheck(testargs_t *args, char *str, char c);
void set_modulus(testargs_t *args, char *str, char c);
void set_num_threads(testargs_t *args, char *str, char c);
void set_num_sockets(testargs_t *args, char *str, char c);
void set_ipv6(testargs_t *args, char *str, char c);
void set_sock_type(testargs_t *args, char *str, char c);
void set_reuse(testargs_t *args, char *str, char c);
void set_recv_mode(testargs_t *args, char *str, char c);
void set_sock_recv_mode(testargs_t *args, char *str, char c);
void set_call_receive(testargs_t *args, char *str, char c);
void set_loopback(testargs_t *args, char *str, char c);
void set_seed(testargs_t *args, char *str, char c);
void set_periodic_data(testargs_t *args, char *str, char c);
void set_verbose(testargs_t *args, char *str, char c);
void set_feed_unit_id(testargs_t *args, char *str, char c);

void set_value(testargs_t *args, char *str, char c);
void set_value_o(testargs_t *args, char *str, char c);
void set_true(testargs_t *args, char *str, char c);
void set_false(testargs_t *args, char *str, char c);

#define MAX_ARGS 64
#define ARRAY_LENGTH(array) (int)(sizeof(array) / sizeof(array[0]))

void maltest_parseargs(const char *opts, int argc, char **argv, 
		       testargs_t *args, const char* prod);
void maltest_usage(const char *progname, const char *opts, const char *prod);
void maltest_usage_verbose(const char *progname, const char *opts, const char *prod);

void save_arg(char c, char *str, int overwrite);
char *load_arg(char c, int index);
int num_set(char c);
void get_leftovers();

void timesup(int sig);
void timesup_err(int sig);
char *inadport_decimal(struct sockaddr_in *sad);
void fail(int rc, char *s);
void fail_th(int rc, char *s);

#define CHECK(S) if (rc != 0) fail(rc, S)
#ifndef _WIN32
#define XCHECK(S) if (rc != 0) fail(errno, S)
#define YCHECK(S) if (rc == -1) fail(errno, S)
#else
#define XCHECK(S) if (rc != 0) fail(WSAGetLastError(), S)
#define YCHECK(S) if (rc == -1) fail(WSAGetLastError(), S)
#endif

#define YCHECK_TH(S) if (rc == -1) do { fail_th(errno, S); return NULL; } while (0)
#define CHECK_COND(c,S) if (!(c)) fail(errno, S)

int64_t time_ns(void);

#ifdef _WIN32

typedef ULONG mal_nfds_t;
typedef struct pollfd mal_pollfd_t;

#if (_WIN32_WINNT < 0x0600)
int mal_recv_all(mal_socket_t sock, char *buf, int len);
#else
#define mal_recv_all(s,b,l) recv(s,b,l,MSG_WAITALL)
#endif

#define mal_poll WSAPoll


int gettimeofday (struct timeval *tv, void *tz);

static int gettimeofday (struct timeval *tv, void *tz)
{
#if 1
        static LARGE_INTEGER Frequency;
        LARGE_INTEGER CurrCounter;
        static int init = 0;
        FILETIME CurrentTime;
        static int UseHighResTimer = 1;
        LONGLONG tmp;
        tz = 0;

        if (!init) {
                if (!QueryPerformanceFrequency(&Frequency)) {
                        UseHighResTimer = 0;
                }
                init = 1;
        }
        if (!tv)
                return -1;

        if (UseHighResTimer) {
                GetSystemTimeAsFileTime(&CurrentTime);
                /* don't include any function overhead */
                QueryPerformanceCounter(&CurrCounter);

                CurrCounter.QuadPart *= 10000000;
                CurrCounter.QuadPart /= Frequency.QuadPart;

                CurrentTime.dwHighDateTime = CurrCounter.HighPart;
                CurrentTime.dwLowDateTime = CurrCounter.LowPart;
        } else {
                GetSystemTimeAsFileTime(&CurrentTime);
        }

        tmp = ((LONGLONG) CurrentTime.dwHighDateTime << 32);
        tmp += (LONGLONG) CurrentTime.dwLowDateTime;

        tv->tv_sec = (long) (tmp / 10000000);
        tv->tv_usec = (long) (tmp - ((LONGLONG)(tv->tv_sec) * 10000000)) / 10;

        return 0;
#else
  FILETIME ft;
  ULARGE_INTEGER base1601;
  ULARGE_INTEGER base1970;

  if (tv != NULL) {
    GetSystemTimeAsFileTime (&ft);
    base1601.LowPart = ft.dwLowDateTime;
    base1601.HighPart = ft.dwHighDateTime;
    base1970.QuadPart = base1601.QuadPart - 116444736000000000;
    tv->tv_sec = (long)(base1970.QuadPart / 10000000);
    tv->tv_usec = (long)((base1970.QuadPart % 10000000) / 10);
  }
  return 0;
#endif
}

static int inet_aton(const char *cp, struct in_addr *inp)
{
  unsigned long x;
  x = inet_addr(cp);
  if (x == INADDR_NONE) {
    return 0;
  }
  inp->S_un.S_addr = x;
  return 1;
}

#define mal_sock_perror(STR)							\
  do {									\
    int __err = WSAGetLastError();					\
    char __msg[1024];							\
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | 				\
                  FORMAT_MESSAGE_IGNORE_INSERTS | 			\
                  FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, __err,	        \
                  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),		\
                  (LPSTR) __msg, 1024, NULL);				\
    fprintf(stderr, "%s: %s\n", STR, __msg);				\
  } while (0)

/*****************************************************
 * Non-windows definitions for generic socket interface
 *****************************************************/
#else /* _WIN32 */

typedef nfds_t mal_nfds_t;
typedef struct pollfd mal_pollfd_t;

#define mal_poll poll
#define mal_sock_perror perror

#define mal_recv_all(s,b,l) recv(s,b,l,MSG_WAITALL)

#endif /* _WIN32 */

/*
 * epoll implementations
 */
#ifdef __linux__
#define mal_epoll_event epoll_event
#define mal_epoll_create epoll_create
#define mal_epoll_ctl epoll_ctl
#define mal_epoll_wait epoll_wait
#else
typedef union mal_epoll_data {
  void *ptr;
  int fd;
  uint32_t u32;
  uint64_t u64;
} mal_epoll_data_t;
struct mal_epoll_event {
  uint32_t events;      /* Epoll events */
  mal_epoll_data_t data;      /* User data variable */
};
int mal_epoll_ctl(int epfd, int op, int fd, struct mal_epoll_event *event);
int mal_epoll_wait(int epfd, struct mal_epoll_event *events, int maxev, int timeo);
int mal_epoll_create(int maxev);
enum EPOLL_EVENTS
  {
    EPOLLIN = 0x001,
#define EPOLLIN EPOLLIN
    EPOLLPRI = 0x002,
#define EPOLLPRI EPOLLPRI
    EPOLLOUT = 0x004,
#define EPOLLOUT EPOLLOUT
    EPOLLRDNORM = 0x040,
#define EPOLLRDNORM EPOLLRDNORM
    EPOLLRDBAND = 0x080,
#define EPOLLRDBAND EPOLLRDBAND
    EPOLLWRNORM = 0x100,
#define EPOLLWRNORM EPOLLWRNORM
    EPOLLWRBAND = 0x200,
#define EPOLLWRBAND EPOLLWRBAND
    EPOLLMSG = 0x400,
#define EPOLLMSG EPOLLMSG
    EPOLLERR = 0x008,
#define EPOLLERR EPOLLERR
    EPOLLHUP = 0x010,
#define EPOLLHUP EPOLLHUP
    EPOLLRDHUP = 0x2000,
#define EPOLLRDHUP EPOLLRDHUP
    EPOLLONESHOT = (1 << 30),
#define EPOLLONESHOT EPOLLONESHOT
    EPOLLET = (1 << 31)
#define EPOLLET EPOLLET
  };
#define EPOLL_CTL_ADD 1 /* Add a file decriptor to the interface.  */
#define EPOLL_CTL_DEL 2 /* Remove a file decriptor from the interface.  */
#define EPOLL_CTL_MOD 3 /* Change file decriptor epoll_event structure.  */

#endif /* not __linux__ */


extern int mal_init_sockets(void);
extern void mal_finish_sockets(void);
extern int mal_is_accel(int s);
extern enum sk_accel_mode mal_accel_mode(int s);

#ifndef MALTEST_OVERRIDE_ARGS
/* we aren't being called by one of the {prod}test_common.h files so
   define test_usage and test_parseargs */
#define test_parseargs(opts, argc, argv, args) maltest_parseargs(opts, argc, argv, args, "SOCK")
#define test_usage(progname, opts) maltest_usage(progname, opts, "SOCK")
#endif /* !MALTEST_OVERRIDE */

#endif /* _MAL_TEST_COMMON_H */
