/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2010 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#ifndef _mcp_global_h_
#define _mcp_global_h_

#ifdef MYRI_PHOENIX
#error "Do not include this file for Phoenix build"
#endif

#include "mcp_config.h"
#include "mcp_types.h"

#define MCP_STATUS_LOAD		0xBEEF1111
#define MCP_STATUS_INIT		0xBEEF2222
#define MCP_STATUS_RUN		0xBEFF3333
#define MCP_STATUS_PARITY	0xBEEF4444
#define MCP_STATUS_ERROR	0xBEEF5555

#define MCP_PARITY_NONE		0xCACA0000
#define MCP_PARITY_MULTIPLE	0xCACA1111
#define MCP_PARITY_PANIC	0xCACA2222
#define MCP_PARITY_RECOVER_NONE	0xCACA3333
#define MCP_PARITY_RECOVER_ALL	0xCACA4444

#define MX_DMA_INVALID_ENTRY	0xffffffff
#define MX_DMA_BOGUS_ENTRY	0x87654321



/* link_state word related definitions:
   reserve room for up to 16 port per NIC, (in practice at most 2)
    - 16 first bits of word represent the up status of each   port,
    - 16 upper bits are to signal a network mismatch
*/
#define MCP_LINK_STATE_UP 1
#define MCP_LINK_STATE_MISMATCH 0x10000


typedef struct
{
  /* Misc */
  uint32_t mcp_status;             /* Current status of the MCP. */
  uint32_t parity_status;
  uint32_t parity_address;
  uint32_t reboot_status;
  
  /* Parameters */
  uint32_t intr_coal_delay;        /* Minimum delay between interrupts (us). */
  uint32_t intr_msix;
  uint32_t ethernet_mtu;
  uint32_t ethernet_smallbuf;
  uint32_t ethernet_bigbuf;
  uint32_t random_seed;
  uint32_t endpt_recovery;
  uint32_t recvq_vpage_cnt;
  uint32_t eventq_vpage_cnt;
  uint32_t idt_war;
  
  /* Interface MCP/Driver */
  mcp_dma_addr_t sys_intr_dma;
  mcp_dma_addr_t endpt_intr_dma[MCP_ENDPOINTS_CNT];
  mcp_dma_addr_t raw_desc;
  mcp_dma_addr_t tx_done_vpage;
  mcp_dma_addr_t counters_vpage;
  mcp_dma_addr_t bogus_vpage;
  mcp_dma_addr_t snf_rx_bytes;


  uint32_t snf_cnt_recv;
  uint32_t snf_cnt_drop_flow;
  uint32_t snf_rx_doorbell_off;
  uint32_t pad1;



  /* ethernet */
  volatile uint32_t ether_tx_cnt;
  volatile uint32_t ether_rx_small_cnt;
  volatile uint32_t ether_rx_big_cnt;
  uint32_t pad2;
  volatile uint8_t ether_tx_ring[MX_ETHER_TX_SIZE];
  volatile uint8_t ether_rx_small_ring[MX_ETHER_RX_SIZE];
  volatile uint8_t ether_rx_big_ring[MX_ETHER_RX_SIZE];
#if MYRI_ETHER_PIO
  volatile uint8_t ether_tx_pio[MX_ETHER_PIO_SLOTS][MX_ETHER_PIO_SIZE];
#endif
  mcp_dma_addr_t ether_rx_desc_dma[MX_ETHER_RX_DESC_VPAGES];
  
  uint32_t counters_offset;
  uint32_t logging_offset;
  uint32_t print_buffer_addr;      /* Address of printf buffer */
  uint32_t print_buffer_pos;       /* Current index in printf buffer */
  volatile uint32_t print_buffer_limit;     /* Limit of printf buffer */
  uint32_t clock_freq;
  uint32_t local_peer_index;
  uint32_t dma_dbg_lowest_addr;
  uint32_t dma_dbg_pfn_max;
  uint32_t z_loopback;
  uint32_t pcie_down_on_error;
  uint32_t lxgdb;
  uint32_t pcie_rx_detect;
  uint32_t pcie_lane_select;
  uint32_t host_page_size;
  uint32_t ether_pause;
  uint32_t license_valid;
  uint32_t license_start_time;	/* timeofday driver loaded */
  uint32_t serdes_mode; /* passed down to mcp to influence serdes init */
} mcp_public_global_t;


#endif  /* _mcp_global_h_ */
