/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2011 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#ifndef _mcp_types_h_
#define _mcp_types_h_

#ifdef MYRI_PHOENIX
#error "Do not include this file for Phoenix build"
#endif

#define MCP_SYSBOX_OFF_CMD	0

#define MX_DMA_ADDR_SHIFT 3

#define MCP_CMD_ADDR 0x400000
#define MCP_KRAW_ADDR 0x440000


/* 8 Bytes */
typedef struct
{
  uint32_t high;
  uint32_t low;
} mcp_dma_addr_t;

typedef mcp_dma_addr_t nic_dma_addr_t;

/* Commands */
enum mcp_cmd_type {
  MCP_CMD_NONE = 0,
  MCP_CMD_PARAMS_READY,
  MCP_CMD_RESET_COUNTERS,
  MCP_CMD_OPEN_ENDPOINT,
  MCP_CMD_SET_ENDPOINT_SESSION,
  MCP_CMD_SET_ENDPT_RECVQ_MPL_SIZE,
  MCP_CMD_GET_ENDPT_SENDQ_OFFSET,
  MCP_CMD_GET_ENDPT_RECVQ_MPL_OFFSET,
  MCP_CMD_GET_ENDPT_EVENTQ_OFFSET,
  MCP_CMD_GET_USER_MMAP_OFFSET,
  MCP_CMD_GET_RDMA_WINDOWS_OFFSET,
  MCP_CMD_GET_HWCLOCK_OFFSET,
  MCP_CMD_SET_ENDPT_MAX,
  MCP_CMD_ENABLE_ENDPOINT,
  MCP_CMD_CLOSE_ENDPOINT,
  MCP_CMD_SET_PEER_MAX,
  MCP_CMD_ADD_PEER,
  MCP_CMD_REMOVE_PEER,
  MCP_CMD_ETHERNET_UP,
  MCP_CMD_ETHERNET_DOWN,
  MCP_CMD_SET_MAC,
  MCP_CMD_ENABLE_FLOW_CONTROL,
  MCP_CMD_DISABLE_FLOW_CONTROL,
  MCP_CMD_ENABLE_PROMISC,
  MCP_CMD_DISABLE_PROMISC,
  MCP_CMD_ENABLE_ALLMULTI,
  MCP_CMD_DISABLE_ALLMULTI,
  MCP_CMD_JOIN_MULTICAST,
  MCP_CMD_LEAVE_MULTICAST,
  MCP_CMD_LEAVE_ALL_MULTICAST,
  MCP_CMD_START_LOGGING,
  MCP_CMD_START_DMABENCH_READ,
  MCP_CMD_START_DMABENCH_WRITE,
  MCP_CMD_CLEAR_RAW_STATE,
  MCP_CMD_SET_DISPERSION,
  MCP_CMD_GET_DISPERSION,
  MCP_CMD_ADD_RX_FILTER,
  MCP_CMD_DEL_RX_FILTER,
  MCP_CMD_SNF_RING_ALLOC,
  MCP_CMD_SNF_RING_INIT,
  MCP_CMD_SNF_RING_FREE,
  MCP_CMD_SNF_RECV,
  MCP_CMD_INIT_LICENSE,
};

enum mcp_cmd_status {
  MCP_CMD_OK = 0,
  MCP_CMD_UNKNOWN,
  MCP_CMD_ERROR_RANGE,
  MCP_CMD_ERROR_BUSY,
  MCP_CMD_ERROR_CLOSED,
  MCP_CMD_ERROR_LICENSE,
};

/* 64 Bytes */
struct mcp_cmd
{
  uint32_t type;
  uint32_t data0;
  uint32_t data1;
  uint32_t data2;
  mcp_dma_addr_t done_addr;
  uint8_t pad[40];
};

struct mcp_cmd_done
{
  uint32_t data;
  uint32_t status;
};


/* 64 Bytes */
struct mcp_raw_tx
{
  uint32_t addr_high;
  uint32_t addr_low;
  uint32_t data_len;
  uint8_t route[8];
  uint8_t route_len;
  uint8_t port;
  uint8_t tx_id;
  uint8_t pad[41];
};


/* 32 Bytes */
typedef struct
{
  uint32_t length;
  uint32_t virt_high;
  uint32_t virt_low;
  uint32_t log_level;
  uint32_t seqnum;
  uint32_t refcnt;
  mcp_dma_addr_t table;
} mcp_rdma_win_t;


#define MX_ETHER_RX_DESC_VPAGES						\
  (2 * MX_ETHER_RX_SIZE / sizeof(mcp_dma_addr_t)			\
   * sizeof(mcp_ether_rx_desc_t) / MX_VPAGE_SIZE)

#define MCP_ETHER_RX_SMALL	1
#define MCP_ETHER_RX_BIG	2

/* 8 Bytes */
typedef struct
{
  uint16_t cksum;
  uint16_t len;
  uint8_t buf_cnt;
  uint8_t flags;
  uint8_t dmaop_cnt;
  uint8_t type;
#if MYRI_ENABLE_PTP
  uint32_t nticks_high;
  uint32_t nticks_low;
#endif
} mcp_ether_rx_desc_t;


#define MCP_RAW_TYPE_NONE	0
#define MCP_RAW_TYPE_TX		1
#define MCP_RAW_TYPE_RX		2

/* 8 Bytes */
typedef struct
{
  uint32_t length;
  uint8_t port;
  uint8_t pad;
  uint8_t tx_id;
  uint8_t type;
#if MYRI_ENABLE_PTP
  uint32_t nticks_high;
  uint32_t nticks_low;
#endif
} mcp_raw_desc_t;


#define MCP_INTR_ETHER_RX	(1 << 0)
#define MCP_INTR_ETHER_TX	(1 << 1)
#define MCP_INTR_ETHER_DOWN	(1 << 2)
#define MCP_INTR_LINK_STATE	(1 << 3)
#define MCP_INTR_DMABENCH	(1 << 4)
#define MCP_INTR_LOGGING	(1 << 5)
#define MCP_INTR_ENDPT		(1 << 6)
#define MCP_INTR_RAW		(1 << 7)

#define MCP_INTR_ENDPT_WAKE_RX	(1 << 8)
#define MCP_INTR_ENDPT_WAKE_TX	(1 << 9)
#define MCP_INTR_ENDPT_CLOSED	(1 << 10)
#define MCP_INTR_ENDPT_ERROR	(1 << 11)
#define MCP_INTR_ENDPT_HELPER	(1 << 12)

#define MCP_INTR_TIMESYNC       (1 << 13)

#define MCP_INTR_SNF_WAKE	(1 << 14)

/* 40 Bytes */
typedef struct
{
  uint32_t ether_tx_cnt;
  uint8_t endpt_bitmap[8];
  uint32_t link_state;
  uint16_t dmabench_cpuc;
  uint16_t dmabench_cnt;
  uint16_t logging_size;
  uint8_t seqnum;
  uint8_t endpt_flags;
  uint8_t pad[15];
  uint8_t mxge_flags;
} mcp_intr_t;

typedef mcp_intr_t nic_intr_t;

/* 16 Bytes */
typedef struct
{
  uint32_t data0;
  uint32_t data1;
  uint32_t seqnum;
  uint16_t index;
  uint8_t flag;
  uint8_t type;
} mcp_slot_t;
  
/* 24 Bytes */
typedef struct {
  uint32_t  nticks_high;
  uint32_t  nticks_low;
  uint32_t  ts_secs;
  uint32_t  ts_nanos;
  uint8_t   ts_status; /* status of time source */
  uint8_t   pad[3];
  uint32_t  nticks_seqnum;
} mcp_hwclock_update_t;



#define MCP_UMMAP_SIZE (16*1024)

#define SNF_TX_DESC_MAX         32
#define SNF_TX_DESC_PKT_MAX	64
#define SNF_TX_DESC_PKT_ALIGN	8

#define SNF_RX_DESC_PKT_MAX	((128 - 12) / 2)
#define SNF_RX_DESC_PAD_V1	(512 - 128 - 64 - (SNF_RX_DESC_PKT_MAX * 4) - 32 - 8)
#define SNF_RX_DESC_PAD_V2	(512 - 128 - 64 - (SNF_RX_DESC_PKT_MAX * 2) - 8)
#define SNF_RX_DESC_SEQNUM_MAX	255

#define SNF_TX_INTR 1
#define SNF_RX_INTR 2

#define MCP_SNF_RECVINFO_OFF	2048

/* Format in FW and lib as eventq elem */
typedef struct {
 uint16_t pkt_lens[SNF_RX_DESC_PKT_MAX];
 uint32_t timestamp_high;
 uint32_t timestamp_low;
 uint16_t total_len;
 uint8_t pkt_cnt;
 uint8_t seqnum; /* FW uploads the first 128B including seqnum */
 /* The whole desc becomes 512B.  The above is 128B.  The firmware uploads
  * the first 128B.  Eventually, this part needs to go into a separate ring.
  */
 struct {
   uint8_t seqnum;
   uint8_t pad0[63];
   uint64_t timestamp;
   union {
     struct {
       uint8_t start_idx[32];
       struct {
         uint16_t off;
         uint8_t pad;
         uint8_t next;
       } pkt[SNF_RX_DESC_PKT_MAX];
       uint8_t pad[SNF_RX_DESC_PAD_V1];
     } v1;
     struct {
       uint16_t hash_val[SNF_RX_DESC_PKT_MAX];
       uint8_t pad[SNF_RX_DESC_PAD_V2];
     } v2;
   } u;
 } prehash; /* lib uses this part to store hash values or ring IDs */
} mcp_snf_rx_desc_t;

typedef struct
{
  uint8_t desc[MCP_UMMAP_SIZE - 8];
  uint32_t eventq_host_flow;
  uint32_t recvq_host_flow;
} mcp_endpt_ummap_t;



#endif /* _mcp_types_h_ */
