/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2012 by Myricom, Inc.  All rights reserved.                 *
 *************************************************************************/

#ifndef _myri_mxgefw_h_
#define _myri_mxgefw_h_


#define MXGEFW_SEND_GO		0x380000
#define MXGEFW_SEND_STOP	0x3C0000
#define MXGEFW_IRQ_DEASSERT	0xF40000
#define MXGEFW_CMD		0xF80000

#define MXGEFW_RSS_MCP_SLOT_TYPE_MIN 0

enum mxgefw_cmd_type {
  MXGEFW_CMD_NONE = 0,
  MXGEFW_CMD_RESET = 1,
  MXGEFW_CMD_SET_INTRQ_DMA = 3,
  MXGEFW_CMD_SET_BIG_BUFFER_SIZE = 4,
  MXGEFW_CMD_SET_SMALL_BUFFER_SIZE = 5,
  MXGEFW_CMD_GET_TX_OFFSET = 6,
  MXGEFW_CMD_GET_SMALL_RX_OFFSET = 7,
  MXGEFW_CMD_GET_BIG_RX_OFFSET = 8,
  MXGEFW_CMD_GET_IRQ_ACK_OFFSET = 9,
  MXGEFW_CMD_GET_IRQ_DEASSERT_OFFSET = 10,
  MXGEFW_CMD_GET_TX_RING_SIZE = 11,
  MXGEFW_CMD_GET_RX_RING_SIZE = 12,
  MXGEFW_CMD_SET_INTRQ_SIZE = 13,
  MXGEFW_CMD_ETHERNET_UP = 14,
  MXGEFW_CMD_ETHERNET_DOWN = 15,
  MXGEFW_CMD_SET_MTU = 16,
  MXGEFW_CMD_GET_INTR_COAL_DELAY_OFFSET = 17,
  MXGEFW_CMD_ENABLE_PROMISC = 20,
  MXGEFW_CMD_DISABLE_PROMISC = 21,
  MXGEFW_CMD_SET_MAC_ADDRESS = 22,
  MXGEFW_CMD_ENABLE_FLOW_CONTROL = 23,
  MXGEFW_CMD_DISABLE_FLOW_CONTROL = 24,
  MXGEFW_CMD_DMA_TEST = 25,
  MXGEFW_CMD_ENABLE_ALLMULTI = 26,
  MXGEFW_CMD_DISABLE_ALLMULTI = 27,
  MXGEFW_CMD_JOIN_MULTICAST_GROUP = 28,
  MXGEFW_CMD_LEAVE_MULTICAST_GROUP = 29,
  MXGEFW_CMD_LEAVE_ALL_MULTICAST_GROUPS = 30,
  MXGEFW_CMD_SET_STATS_DMA_V2 = 31,
  MXGEFW_CMD_GET_MAX_TSO6_HDR_SIZE = 44,
  MXGEFW_CMD_SET_TSO_MODE = 45,
  MXGEFW_CMD_SET_RSS_MCP_SLOT_TYPE = 52,
  MXGEFW_CMD_GET_DCA_OFFSET = 56,
};

enum mxgefw_cmd_status {
  MXGEFW_CMD_OK = 0,
  MXGEFW_CMD_UNKNOWN = 1,
  MXGEFW_CMD_ERROR_RANGE = 2,
};

struct mxgefw_cmd {
  uint32_t type;
  uint32_t data0;
  uint32_t data1;
  uint32_t data2;
  mcp_dma_addr_t done_addr;
};

struct mxgefw_cmd_done {
  uint32_t data;
  uint32_t status;
};

typedef struct {
  /* add new counters at the beginning */
  uint32_t future_use[1];
  uint32_t dropped_pause;
  uint32_t dropped_unicast_filtered;
  uint32_t dropped_bad_crc32;
  uint32_t dropped_bad_phy;
  uint32_t dropped_multicast_filtered;
  /* 40 Bytes */
  uint32_t send_done_count;

#define MXGEFW_LINK_DOWN 0
#define MXGEFW_LINK_UP 1
#define MXGEFW_LINK_MYRINET 2
#define MXGEFW_LINK_UNKNOWN 3
  uint32_t link_up;
  uint32_t dropped_link_overflow;
  uint32_t dropped_link_error_or_filtered;
  uint32_t dropped_runt;
  uint32_t dropped_overrun;
  uint32_t dropped_no_small_buffer;
  uint32_t dropped_no_big_buffer;
  uint32_t rdma_tags_available;

  uint8_t tx_stopped;
  uint8_t link_down;
  uint8_t stats_updated;
  uint8_t valid;
} mxgefw_intr_data_t;

/* 
   flags used in mcp_kreq_ether_send_t:

   The SMALL flag is only needed in the first segment. It is raised
   for packets that are total less or equal 512 bytes.

   The CKSUM flag must be set in all segments.

   The PADDED flags is set if the packet needs to be padded, and it
   must be set for all segments.

   The  MXGEFW_FLAGS_ALIGN_ODD must be set if the cumulative
   length of all previous segments was odd.
*/
#define MXGEFW_FLAGS_SMALL      0x1
#define MXGEFW_FLAGS_TSO_HDR    0x1
#define MXGEFW_FLAGS_FIRST      0x2
#define MXGEFW_FLAGS_ALIGN_ODD  0x4
#define MXGEFW_FLAGS_CKSUM      0x8
#define MXGEFW_FLAGS_TSO_LAST   0x8
#define MXGEFW_FLAGS_NO_TSO     0x10
#define MXGEFW_FLAGS_TSO_CHOP   0x10
#define MXGEFW_FLAGS_TSO_PLD    0x20

#define	MXGEFW_MAX_SEND_DESC 12
#define MXGEFW_PAD	    2

/* 16 Bytes */
typedef struct mxgefw_tx_desc {
  uint32_t addr_high;
  uint32_t addr_low;
  uint16_t pseudo_hdr_offset;
  uint16_t length;
  uint8_t  pad;
  uint8_t  rdma_count;
  uint8_t  cksum_offset; 	/* where to start computing cksum */
  uint8_t  flags;	       	/* as defined above */
} mxgefw_tx_desc_t;


#endif /* _myri_mxgefw_h_ */
