/*************************************************************************
 * The contents of this file are subject to the MYRICOM SNIFFER10G
 * LICENSE (the "License"); User may not use this file except in
 * compliance with the License.  The full text of the License can found
 * in LICENSE.TXT
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * Copyright 2008 - 2010 by Myricom, Inc.  All rights reserved.
 ***********************************************************************/

#ifndef _myri_snf_io_h
#define _myri_snf_io_h

#ifdef MYRI_PHOENIX
#error "Do not include this file for Phoenix build"
#endif

#include "mal_io.h"
#include "mcp_types.h"
#include "mcp_config.h"
#include "mcp_requests.h"
#include "mal_piocopy.h"
#include "mal_byteswap.h"

/*
 * Tunables
 */
#define MYRI_SNF_MAX_RINGS  32

#define SNF_RSS_FLAGS_DEFAULT         (SNF_RSS_IP|SNF_RSS_SRC_PORT|SNF_RSS_DST_PORT)
#define SNF_DATARING_SZ_DEFAULT       (256ULL<<20)
#define SNF_MTUWRAP_SIZE              65536

/* 
 * Non-tunables
 */
#define SNF_VPAGE_SIZE 4096
#define SNF_VPAGE_SHIFT 12
#define SNF_DESC_SHIFT 3

#define SNF_PORTNUM_MAX 16
#define SNF_RPD_ALIGN 16
#define SNF_RPD_MASK (SNF_RPD_ALIGN-1)
#define SNF_RPD_MAXLENGTH 16384

#define SNF_RSS_FLAGS_CUSTOM_FUNC 0x80000000

#define SNF_RING_UNUSED -1

#define SNF_ENDPOINT_RX -3
#define SNF_ENDPOINT_RX_RING -4

#define MYRI_SNF_F_RX_DUP_MA 0x80
#define MYRI_SNF_F_RX_COPY  0x100 /* Same as SNF_F_RX_PRIVATE */
#define MYRI_SNF_F_RX_DUP   0x200
#define MYRI_SNF_F_RX_SHARED 0x400
/* SNF_F_RX_DUPLICATE at API level is a mask of RX_COPY|RX_DUP */

#define MYRI_SNF_RXMAP_PQ_DESC 0x1
#define MYRI_SNF_RXMAP_PQ_DATA 0x2
#define MYRI_SNF_RXMAP_VQ_DESC 0x4
#define MYRI_SNF_RXMAP_VQ_DATA 0x8
#define MYRI_SNF_RXMAP_SHARED  0x10

#define MYRI_OFFSET_UNMAPPED -1

#if !MAL_OS_WINDOWS
#define SNF__CACHEALIGN_BEGIN
#define SNF__PIOALIGN_BEGIN
#define SNF__CACHEALIGN_END __attribute__((aligned(64)))
#define SNF__PIOALIGN_END   __attribute__((aligned(8)))
#else
#define SNF__CACHEALIGN_BEGIN	__declspec(align(64))
#define SNF__PIOALIGN_BEGIN	__declspec(align(8))
#define SNF__CACHEALIGN_END
#define SNF__PIOALIGN_END
#endif

#define SNF_ALIGNDOWN(p,P) (((uint64_t)(uintptr_t)(p))&~((uint64_t)(uintptr_t)((P)-1)))
#define SNF_ALIGNUP(p,P)   (SNF_ALIGNDOWN((uint64_t)(uintptr_t)(p)+((uint64_t)(uintptr_t)((P)-1)),P))

#define snf_assert mal_assert
#define snf_always_assert mal_always_assert

#ifndef MAL_KERNEL
#undef mal_assertion_failed
#define mal_assertion_failed snf__assertion_failed
void snf__assertion_failed(const char *assertion, int line, const char *file);
#endif

#ifndef _WIN32
#define SNF__CTA(x)						\
  do								\
    {								\
      char (*a)[(__builtin_constant_p(x) && (x)) ? 1 : -1] = 0;	\
      (void) a;							\
    }								\
  while (0)
#else
#define SNF__CTA(x)						\
  do								\
    {								\
      char (*a)[(x) ? 1 : -1] = 0;				\
      (void) a;							\
    }								\
  while (0)
#endif

#define SNF_IS_POWER_OF_TWO(x) (((x) & ((x) - 1)) == 0)

#define SNF_DEBUGM_ERR0R    0x0
#define SNF_DEBUGM_WARN     0x1
#define SNF_DEBUGM_PARAM    0x2
#define SNF_DEBUGM_QSTATS   0x4
#define SNF_DEBUGM_TIMESYNC 0x8
#define SNF_DEBUGM_IOCTL    0x10
#define SNF_DEBUGM_QEVENTS  0x20
#define SNF_DEBUGM_ARISTA   0x40

#define SNF_DEBUGM_LABELS { "WARN", "PARAM", "QSTATS", "TIMESYNC", "IOCTL", "QEVENTS", "ARISTA" }

#define SNF_DEBUGM_PREFIX_WARN ' '
#define SNF_DEBUGM_PREFIX_ERR0R 'E'
#define SNF_DEBUGM_PREFIX_PARAM 'P'
#define SNF_DEBUGM_PREFIX_QSTATS 'S'
#define SNF_DEBUGM_PREFIX_TIMESYNC 'T'
#define SNF_DEBUGM_PREFIX_IOCTL 'I'
#define SNF_DEBUGM_PREFIX_QEVENTS 'Q'
#define SNF_DEBUGM_PREFIX_ARISTA 'A'

#ifdef MAL_KERNEL
#define _SNF_DPRINTF(p,level,format,...)  do {                  \
        if (((p)->debug_mask & SNF_DEBUGM_ ## level) || SNF_DEBUGM_ERR0R == SNF_DEBUGM_ ## level) \
            MX_PRINT(("snf.%d.%d %c " format,   \
                    (p)->portnum, (p)->ring_id, SNF_DEBUGM_PREFIX_ ## level, \
                    __VA_ARGS__));                              \
        } while (0)
#else
extern int snf__getpid(void);
#define _SNF_DPRINTF(p,level,format,...)  do {                  \
        if (((p)->debug_mask & SNF_DEBUGM_ ## level) || SNF_DEBUGM_ERR0R == SNF_DEBUGM_ ## level) \
            fprintf((p)->debug_fp ? (p)->debug_fp : stderr, "%5u snf.%c.%-2d %c " format,      \
		    snf__getpid(), \
                    (p)->portnum >= 0 && (p)->portnum < SNF_PORTNUM_MAX ? \
                            '0' + (p)->portnum : 'X',                   \
                    (p)->ring_id, SNF_DEBUGM_PREFIX_ ## level, \
                    __VA_ARGS__);                               \
	    if ((p)->debug_fp) fflush((p)->debug_fp);		\
        } while (0)
#endif

#define SNF_DPRINTF(params,level,format,...) \
            _SNF_DPRINTF(params,level,format,__VA_ARGS__)


struct snf_recv_req;
typedef int (*snf_rss_fn_t)(struct snf_recv_req *, void *, uint32_t *);

#ifndef MAL_KERNEL
#include <stdio.h> /* FILE */
#endif
struct snf__params {
  uint32_t     debug_mask;
  uint32_t     portnum;
  int32_t      ring_id;
  snf_rss_fn_t rss_hash_fn;
  void        *rss_context;
  int32_t      port_ring_id; /* DUP_MA */
#ifndef MAL_KERNEL
  const char  *debug_filename;
  FILE        *debug_fp;
  char         debug_file_buf[128];
#endif
};

struct snf__flow_state {
  uint32_t flow_desc;
  uint32_t flow_data;
};

typedef struct { volatile uint32_t slock; } snf__spinlock_t;

#define SNF__RX_STATE_SIZE (SNF_ALIGNUP(sizeof(struct snf__rx_state) + 256, SNF_VPAGE_SIZE))
#define SNF__RX_STATE(rx_s_base,ring_id) \
  ((struct snf__rx_state *)(uintptr_t)((uintptr_t)rx_s_base + ring_id * SNF__RX_STATE_SIZE))
#define SNF_VPAGE_IDX(rq_off) (((uint64_t) rq_off) >> SNF_VPAGE_SHIFT)
#define SNF_FLOW_DATA(data_cnt)  ((uint32_t)((((uint64_t)data_cnt)) >> SNF_VPAGE_SHIFT))
#define SNF_DESC_VPAGES(cnt) (uint32_t)(((uint64_t) cnt) >> SNF_DESC_SHIFT)
#define SNF_DATA_VPAGES(rq_off) (uint32_t)SNF_VPAGE_IDX(rq_off)

struct snf__rx_sync {
    uint64_t rs_desc_cnt;  /* Desc queue for current batch */
    uint64_t rs_data_head;
    uint64_t rs_data_tail;
};

struct snf__pq_state {
  /* The user updates these via kernel */
  struct snf__pq_state_rw  {
    uint32_t evq_prehash_seqnum;
    uint32_t evq_next_prehash_idx;
    uint64_t evq_desc_cnt;
    uint64_t rq_prehash_data_off; /* Offset in receive queue */
    uint64_t evq_ts_last;
  } rw;
  /* Set once by the kernel, immutable afterwards */
  uint32_t evq_last_prehash_idx;
};

struct snf__rx_state {
  struct snf__eventq_state  {
    uint32_t evq_seqnum;
    uint32_t evq_next_idx;
    uint32_t evq_last_idx;
    int32_t  evq_evinfo_idx;
    uint64_t evq_desc_cnt;
    uint64_t evq_data_head;
  } evq;

  struct snf__recvq_state {
    uint64_t rq_data_off; /* Offset in receive queue */

    /* if head == tail, the queue is empty */
    uint64_t rq_data_head; /* Where next packet will be written */
    uint64_t rq_data_tail; /* End of the last packet that was returned */
    uint64_t rq_data_tail_last; /* Last update */
    uint64_t rq_data_size;

    struct snf__recvq_stats {
      uint64_t rq_pkt_cnt;
      uint64_t rq_pkt_length;
      uint64_t rq_pkt_cnt_skip_rss;
      uint64_t rq_pkt_cnt_drop;
      uint64_t rq_neg_adj_event;
    } stats;

    uint64_t rq_is_reattach;
    uint64_t pad;
    /* Pad so ts_last starts at a cache line.. */
    uint64_t rq_ev_ts_last;

    struct snf__eventq_evinfo {
      uint64_t ev_timestamp;
      uint32_t ev_length;
      uint32_t ev_length_time;
      uint32_t ev_pkt_cnt;
      uint32_t ev_pkt_off;
      union {
        uint16_t ev_pkt_lens[SNF_RX_DESC_PKT_MAX]; /* non-prehash mode */
        struct {
          uint32_t ev_ring_id;
          uint32_t pad;
          struct {
            uint16_t len;
            uint16_t skip_off;
          } ev_pkt[SNF_RX_DESC_PKT_MAX];
        } prehash_v1;
      } u;
    } ev_info;
  } recvq;
};

/* Structure is shared between user and kernel page.
 * It must be 32/64-bit clean in terms of sizes and alignments
 */
struct snf__tx_state {
  SNF__CACHEALIGN_BEGIN myri_endpt_flow_t ep_flow SNF__CACHEALIGN_END;
  SNF__PIOALIGN_BEGIN   snf__spinlock_t tx_lock   SNF__PIOALIGN_END;

  uint32_t dq_avail;
  uint32_t dq_head_idx;
  int32_t  dq_inflight;
  uint32_t dq_lengths[SNF_TX_DESC_MAX];

  uint64_t tx_pkts_cnt;
  uint64_t tx_flush_cnt;
  uint64_t tx_flush_intr_cnt;

  struct snf__tx_desc_state {
    uint32_t d_cnt;
    uint32_t d_length;
    SNF__PIOALIGN_BEGIN uint16_t d_pkt_lens_be16[SNF_TX_DESC_PKT_MAX] SNF__PIOALIGN_END;
  } txd;
};

struct snf__tx_shared_params {
  uint32_t  txd_cnt;
  uint32_t *txd_dbaddr_be32;
  uint8_t  *txd_compl_update;
  uintptr_t txd_pio_base;
  struct snf__tx_state *tx_s;
};

typedef struct {
  myri_set_endpt_t set_endpt;
  uint32_t num_rings;
  uint32_t open_flags;
  uint32_t rss_flags;
  uint32_t aggr_portmask; /* ports that belong to the aggregation group */
  int32_t  app_id; /* id of the calling process (DUP_MA mode) */
  uint32_t rx_map_flags;
  uint64_t rx_data_preamble_size;
  uint64_t rx_data_ring_size; /* without preamble */
  uint64_t rx_desc_ring_size;
} myri_snf_rx_params_t;

typedef struct {
  myri_set_endpt_t set_endpt;
  uint32_t epid;
  uint32_t pad;
  uint64_t tx_data_ring_size;
  uint64_t tx_data_ring_offset;
  uint64_t tx_desc_ring_size;
  uint64_t tx_desc_ring_offset;

  uint64_t tx_doorbell_size;
  uint64_t tx_doorbell_offset;
  uint64_t tx_flow_size;
  uint64_t tx_flow_offset;
  uint64_t tx_compl_size;
  uint64_t tx_compl_offset;
} myri_snf_tx_params_t;

typedef struct {
  myri_set_endpt_t      set_endpt;
  int32_t               ring_id;  /* in-out */
  int32_t               port_ring_id;  /* DUP_MA mode, to aid debugging */
  myri_snf_rx_params_t  params;
  myri_clksync_nticks_t nticks;

  struct snf__flow_state flow;

  uint64_t rx_data_ring_offset;
  uint64_t rx_desc_ring_offset;

  uint64_t vstate_all_size;
  uint64_t vstate_all_offset;
  uint64_t kernel_window_size;
  uint64_t kernel_window_offset;
  uint64_t rx_doorbell_size;
  uint64_t rx_doorbell_offset;
  uint64_t rx_recv_info_size;
  uint64_t rx_recv_info_offset;
  uint64_t rx_recv_info_pg_offset;
} myri_snf_rx_attach_t;

typedef struct {
  myri_set_endpt_t      set_endpt;
  myri_snf_rx_params_t  params;
  uint64_t rx_data_ring_offset;
  uint64_t rx_desc_ring_offset;
  uint64_t rx_doorbell_size;
  uint64_t rx_doorbell_offset;
  uint64_t pq_state_size;
  uint64_t pq_state_offset;
} myri_snf_pq_attach_t;

typedef struct {
  uint64_t snf_pkt_recv;
  uint64_t snf_pkt_overflow;
  uint64_t nic_pkt_overflow;
  uint64_t nic_pkt_bad;
  uint64_t nic_bytes_recv;
  uint64_t nic_pkt_send;
  uint64_t nic_bytes_send;
  uint64_t nic_pkt_dropped;
} myri_snf_stats_t;

typedef struct {
  uint32_t  port_id;
  uint32_t  ring_id;
  mcp_snf_rx_desc_t *evq_base; /* used for prehash + aggregation */
} myri_snf_port_group_t;

typedef struct {
  union {
    uint32_t portmask;
    uint32_t portnum; /* locked port */
  } u;
  int timeout_ms;
} myri_snf_pq_group_lock_t;

static inline
void
snf__rx_flow_update(uint64_t *rx_flow_be64, const struct snf__flow_state *flowh)
{
  struct {
    uint32_t flow_desc_be32;
    uint32_t flow_data_be32;
  } flow;
  flow.flow_desc_be32 = htonl(flowh->flow_desc);
  flow.flow_data_be32 = htonl(flowh->flow_data);
  *rx_flow_be64 = *((uint64_t *)&flow);
  MAL_STBAR();
  return;
}

static inline
int snf__ffs(int x)
{
#if !MAL_OS_WINDOWS
  int r;
  __asm__("bsfl %1,%0\n\t"
	  "cmovzl %2,%0" 
	  : "=r" (r) : "rm" (x), "r" (-1));
  return r+1;
#else
  int r = 0;
  return _BitScanForward(&r, x) ? r+1 : 0;
#endif
}

/* First call must be with idx_prev==-1.
 * Use case really intended to be through snf__ffs_foreach */
static inline
int snf__ffs_next(int mask, int idx_prev)
{
  int idx = idx_prev+1;
  mask &= ~( (1<<idx)-1 );
  if (mask == 0 || idx >= 32)
    return -1;
  return snf__ffs(mask) - 1;
}

#define SNF_FFS_FOREACH(mask, i) for (i = -1; (i = snf__ffs_next(mask, i)) >= 0; )

static inline
uint32_t snf__cmpswap_u32(uint32_t *p, uint32_t oldv, uint32_t newv)
{
#if !MAL_OS_WINDOWS
  __asm__("lock\n\t"
          "cmpxchgl %1,%2"
          : "=a" (oldv) : "q" (newv), "m" (*p), "0" (oldv) : "memory");
  return oldv;
#else
  return _InterlockedCompareExchange(p, newv, oldv);
#endif
}

static inline
int snf__spin_trylock(snf__spinlock_t *lock)
{
#if !MAL_OS_WINDOWS
  int oldv;

  __asm__ __volatile__("xchgl %0, %1"
                       :"=q" (oldv), "=m" (lock->slock)
                       :"0" (0) : "memory");
  return oldv > 0;
#else
  return _InterlockedExchange(&lock->slock, 0) > 0;
#endif
}

static inline
void snf__spin_init(snf__spinlock_t *lock)
{
  lock->slock = 1;
}

static inline
void snf__spin_lock(snf__spinlock_t *lock)
{
  while (!snf__spin_trylock(lock))
#if !MAL_OS_WINDOWS
       __asm__ __volatile__("rep;nop": : :"memory");
#else
       ; /* Not sure what to replace this with. */
#endif
}

static inline
void snf__spin_unlock(snf__spinlock_t *lock)
{
#if !MAL_OS_WINDOWS
  __asm__ __volatile__("movl $1, %0"
                       :"=m" (lock->slock)
                       : : "memory");
#else
  _InterlockedExchange(&lock->slock, 1);
#endif
}

static inline
void
snf__tx_pioflush(struct snf__tx_shared_params *p)
{
  struct snf__tx_state *ss = p->tx_s;
  size_t piolen;
  uint32_t fdata, d_length;
  uintptr_t pioaddr;
  uint8_t idx;

  mal_assert(ss->txd.d_cnt > 0);
  mal_assert(ss->txd.d_length > 0);

  idx = ss->dq_head_idx & (p->txd_cnt - 1);
  d_length = ss->txd.d_length;
  mal_assert(d_length != 0);
  ss->dq_head_idx++;
  ss->dq_lengths[idx] = d_length;

  piolen = (ss->txd.d_cnt * sizeof(uint16_t) + 7) & ~7;
  pioaddr = p->txd_pio_base + idx * SNF_TX_DESC_PKT_MAX * sizeof(uint16_t);
  mal_piocopy64_inline((uint64_t *) pioaddr,
                       (uint64_t *) ss->txd.d_pkt_lens_be16, (unsigned)piolen, 0);

  fdata = ((uint32_t) ss->txd.d_length << 16) | ss->txd.d_cnt;
  ss->tx_pkts_cnt += ss->txd.d_cnt;
  ss->txd.d_length = 0;
  ss->txd.d_cnt = 0;
  ss->dq_inflight++;
  ss->dq_avail -= (SNF_ALIGNUP(d_length, SNF_VPAGE_SIZE) - d_length);
  ss->tx_flush_cnt++;
  *p->txd_dbaddr_be32 = htonl(fdata);
  MAL_STBAR();
}

static inline
int
snf__tx_update_completion(struct snf__tx_shared_params *tx_sp)
{
  struct snf__tx_state *ss = tx_sp->tx_s;
  uint8_t tx_compl_host, tx_compl_nic, idx;
  uint32_t d_length;
  int done = 0;

  /* reclaim all requests that are complete right now */
  tx_compl_nic = *tx_sp->txd_compl_update;
  tx_compl_host = (uint8_t) ss->ep_flow.tx_flow.flow_desc;
  if (tx_compl_nic == tx_compl_host)
    return 0; /* nothing */

  while (tx_compl_host != tx_compl_nic) {
    ss->dq_inflight--;
    mal_assert(ss->dq_inflight >= 0);
    idx = tx_compl_host & (uint8_t)(tx_sp->txd_cnt-1);
    d_length = ss->dq_lengths[idx];
    mal_assert(d_length != 0);
    ss->dq_avail += SNF_ALIGNUP(d_length, SNF_VPAGE_SIZE);
    ss->dq_lengths[idx] = 0;
    tx_compl_host++;
    done++;
  }

  /* Write updated value on host page */
  ss->ep_flow.tx_flow.flow_desc = tx_compl_host;
  return done;
}

#endif /* _myri_snf_io_h */
