#ifndef PHX_FIRMWARE_VERSIONS_H
#define PHX_FIRMWARE_VERSIONS_H
/* 
 * Phoenix firmware version is not stored, but has known SHA1 checksum.
 * This table allows working backwards, future revs will say unknown. 
 */ 
typedef struct version_entry {
        uint32_t fw_sha1;
        char *fw_name;
} ver_log_t;

/* The ordering of this in the tools is chronological.
 * In the kernel it is feature based.
 */
const ver_log_t  known_versions[] = {
/*  1 */  { 0xc23d7e4d, "E1-prerelease" },   
/*  2 */  { 0xa92cb372, "E7-prerelease" },
/*  3 */  { 0xb41d740a, "0.5.2" },
/*  4 */  { 0xbf61be63, "0.5.3" },
/*  5 */  { 0x4467aa03, "0.5.4" },
/*  6 */  { 0xcec5f239, "0.6.1" },
/*  7 */  { 0x72aa379e, "0.7.2" },
/*  8 */  { 0xad83ee84, "0.5.5" },
/*  9 */  { 0x77d63ec9, "0.8.1" },
/* 10 */  { 0x4245b5ca, "0.8.2" },
/* 11 */  { 0x384f41af, "0.8.3" },
/* 12 */  { 0x3e54665f, "1.0.0" },
/* 13 */  { 0xdb0cf112, "1.0.1" },
/* 14 */  { 0xd13d8579, "1.0.2" },
/* 15 */  { 0x46206eb2, "1.0.3" },
/* 16 */  { 0x31152d0f, "1.0.4" },
/* 17 */  { 0x4a6a299e, "1.0.5" },
/* 18 */  { 0x4865ac4b, "1.0.6" },
/* 19 */  { 0xd81517b6, "1.0.7" },
/* 20 */  { 0xb7ad6175, "1.0.8" },
/* 21 */  { 0xa941b4ab, "PKTGEN" },
/* 22 */  { 0x71b25343, "1.1.0" },
/* 23 */  { 0x58307af3, "1.1.0" },
/* 24 */  { 0x45ee3c6c, "1.2.0" },
/* 25 */  { 0x5fe2b310, "1.2.1" },
/* 26 */  { 0x341951bf, "1.3.0" },
/* 27 */  { 0xaf20e0f4, "1.2.2" },
/* 28 */  { 0xb50e0a6f, "1.4.0" },
/* 29 */  { 0xf912a98a, "1.4.0" },
/* 30 */  { 0xcfbb6836, "1.4.1" },
/* 31 */  { 0x01d14a04, "1.4.1" },
/* 32 */  { 0x6790572f, "1.5.0" },
/* 33 */  { 0x23486542, "2.0.g0" },
/* 34 */  { 0x0000adde, "2.0.g2" }, // The K35 version Hex values
/* 35 */  { 0xe28d461d, "2.0.0" },
/* 36 */  { 0xb2abcb44, "2.0.1" },
/* 37 */  { 0x30821cc7, "2.0.2" },
/* 38 */  { 0x51bf6cd2, "2.0.3" },
/* 39 */  { 0xe9222777, "2.0.4" },
/* 40 */  { 0x36e93785, "2.0.5" },
/* 41 */  { 0x057dc79c, "2.1.0_rc0" },  /* k35 sniffer 4 port */
/* 42 */  { 0x9c28437d, "2.1.0_rc1" },  /* k35 sniffer 4 port */
/* 43 */  { 0xab6bf1d1, "2.1.0_rc2" },  /* k35 sniffer 4 port */
/* 44 */  { 0x634a3dbd, "2.1.1_rc2" },
/* 45 */  { 0xb5a06904, "2.1.1_rc2" },
/* 46 */  { 0x7c5a2785, "2.1.1_rc3" },
/* 47 */  { 0x3e254a05, "2.0.6_rc1" },
/* 48 */  { 0xce8b1ce0, "2.0.6_rc2" },
/* 49 */  { 0xc2bb1930, "2.0.6" },
/* 50 */  { 0x550ae95a, "2.1.1_rc4" },  /* 4-port */
/* 51 */  { 0x81af7960, "2.1.1_rc4" },  /* 2-port */
/* 52 */  { 0x7cabf1ca, "2.1.1_rc5" },
/* 53 */  { 0xf25043a0, "2.1.1" },
/* 54 */  { 0x8b1b760b, "2.1.1" },   /* 2x10G, 4x10G */
/* 55 */  { 0xae6d09a6, "2.1.1" },   /* 2x1G/2x10G, 4x1G */
/* 56 */  { 0x787bb0c3, "2.1.2" },   /* 4x1G, 4x10G */
/* 57 */  { 0x7ebb6a94, "2.1.2" },   /* 2x1G/2x10G, 2x1G, 2x1G */
/* 58 */  { 0xcb4a2552, "2.1.3" },   /* k35 snf all */
/* 59 */  { 0x1d3c9c37, "2.1.4" },   /* k35 snf all */
/* 60 */  { 0x2d13f73a, "2.1.5" },   /* k35 snf all */
/* 61 */  { 0x8bf80520, "2.1.6" },   /* k35 snf all */
/* 0  */  { 0x0, "unknown" }
};

/* List above starts at  1 */
#define MAX_FIRMWARE_VERSIONS    62

#endif
