/*************************************************************************
 * The contents of this file are subject to the MYRICOM SNIFFER10G
 * LICENSE (the "License"); User may not use this file except in
 * compliance with the License.  The full text of the License can found
 * in LICENSE.TXT
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * Copyright 2008 - 2010 by Myricom, Inc.  All rights reserved.
 * Copyright 2015 - 2016 by CSP, Inc.  All rights reserved.
 ***********************************************************************/

#ifndef _phx_snf_io_h
#define _phx_snf_io_h

#include "mal_io.h"
#include "mcp_config.h"
#include "mal_piocopy.h"
#include "mal_byteswap.h"
#include "phx_hw.h"

/***************************************/

#define PHX_SNF_MAX_COALESCE_PKT  8

/*
 * Tunables
 */
#define MYRI_SNF_MAX_RINGS  32

#define SNF_RSS_FLAGS_DEFAULT         (SNF_RSS_IP|SNF_RSS_SRC_PORT|SNF_RSS_DST_PORT)
#define SNF_DATARING_SZ_DEFAULT       (16ULL<<20)
#define SNF_MTUWRAP_SIZE              65536
/* K35SNF */
#define SNF_TX_DATARING_SZ_DEFAULT    (4<<20)

/* 
 * Non-tunables
 */
#define SNF_VPAGE_SIZE 4096
#define SNF_VPAGE_SHIFT 12
#define SNF_DESC_SHIFT 3 /* 512B descriptors, so 2^3 = 8 of them in one page */

#define SNF_PORTNUM_MAX 16
#define SNF_RPD_ALIGN 16
#define SNF_RPD_MASK (SNF_RPD_ALIGN-1)
#define SNF_RPD_MAXLENGTH 16384

#define SNF_RSS_FLAGS_CUSTOM_FUNC 0x80000000
#define SNF_RSS_FLAGS_HW 0x40000000 /* Indicates ring is using HW hash */
/* The app may set this flag to not auto-enable HW hash. That is, the driver
 * does not enable HW hash, and the app will use software hash. This flag is
 * used for testing only and is not exposed to the user via snf.h
 */
#define SNF_RSS_FLAGS_NO_HW 0x10000
/* These values must match snf_rss_mode_flags in snf.h. The kernel driver
 * cannot include snf.h directly, so uses these.
 */
#define SNF_RSS_FLAGS_IP 0x01
#define SNF_RSS_FLAGS_SRC_PORT 0x10
#define SNF_RSS_FLAGS_DST_PORT 0x20
#define SNF_RSS_FLAGS_GTP 0x40
#define SNF_RSS_FLAGS_GRE 0x80

/* Supported port merging modes */
#define SNF_PORT_MERGE_MODE_NONE      0x00
#define SNF_PORT_MERGE_MODE_NEXT_RING 0x01

#define SNF_RING_UNUSED -1

#define SNF_ENDPOINT_RX -3
#define SNF_ENDPOINT_RX_RING -4

/* These must match SNF_F flags in snf.h */
#define MYRI_SNF_F_PSHARED 0x1 /* Same as SNF_F_PSHARED */
#define MYRI_SNF_F_AGGREGATE_PORTMASK 0x2 /* Same as SNF_F_AGGREGATE_PORTMASK */
#define MYRI_SNF_F_RX_DUP_MA 0x80
#define MYRI_SNF_F_RX_COPY  0x100 /* Same as SNF_F_RX_PRIVATE */
#define MYRI_SNF_F_RX_DUP   0x200
/* SNF_F_RX_DUPLICATE at API level is a mask of RX_COPY|RX_DUP */
/* Tell the driver not to enable/allow next_ring merging */
#define MYRI_SNF_F_NO_NEXT_RING_MERGE 0x10000

#define MYRI_SNF_RXMAP_PQ_DESC 0x1
#define MYRI_SNF_RXMAP_PQ_DATA 0x2
#define MYRI_SNF_RXMAP_VQ_DESC 0x4
#define MYRI_SNF_RXMAP_VQ_DATA 0x8
#define MYRI_SNF_RXMAP_SHARED  0x10

#define MYRI_OFFSET_UNMAPPED -1

#define SNF_BOARD_TYPE_DBL4P   0
#define SNF_BOARD_TYPE_K35SNF  1

#if !MAL_OS_WINDOWS
#define SNF__CACHEALIGN_BEGIN
#define SNF__PIOALIGN_BEGIN
#define SNF__CACHEALIGN_END __attribute__((aligned(64)))
#define SNF__PIOALIGN_END   __attribute__((aligned(8)))
#else
#define SNF__CACHEALIGN_BEGIN	__declspec(align(64))
#define SNF__PIOALIGN_BEGIN	__declspec(align(8))
#define SNF__CACHEALIGN_END
#define SNF__PIOALIGN_END
#endif

#define SNF_ALIGNDOWN(p,P) (((uint64_t)(uintptr_t)(p))&~((uint64_t)(uintptr_t)((P)-1)))
#define SNF_ALIGNUP(p,P)   (SNF_ALIGNDOWN((uint64_t)(uintptr_t)(p)+((uint64_t)(uintptr_t)((P)-1)),P))

#define snf_assert mal_assert
#define snf_always_assert mal_always_assert

#ifndef MAL_KERNEL
#undef mal_assertion_failed
#define mal_assertion_failed snf__assertion_failed
void snf__assertion_failed(const char *assertion, int line, const char *file);
#endif

#ifndef _WIN32
#define SNF__CTA(x)						\
  do								\
    {								\
      char (*a)[(__builtin_constant_p(x) && (x)) ? 1 : -1] = 0;	\
      (void) a;							\
    }								\
  while (0)
#else
#define SNF__CTA(x)						\
  do								\
    {								\
      char (*a)[(x) ? 1 : -1] = 0;				\
      (void) a;							\
    }								\
  while (0)
#endif

#define SNF_IS_POWER_OF_TWO(x) (((x) & ((x) - 1)) == 0)

#define SNF_DEBUGM_ERR0R    0x0
#define SNF_DEBUGM_WARN     0x1
#define SNF_DEBUGM_PARAM    0x2
#define SNF_DEBUGM_QSTATS   0x4
#define SNF_DEBUGM_TIMESYNC 0x8
#define SNF_DEBUGM_IOCTL    0x10
#define SNF_DEBUGM_QEVENTS  0x20
#define SNF_DEBUGM_ARISTA   0x40

#define SNF_DEBUGM_LABELS { "WARN", "PARAM", "QSTATS", "TIMESYNC", "IOCTL", "QEVENTS", "ARISTA" }

#define SNF_DEBUGM_PREFIX_WARN ' '
#define SNF_DEBUGM_PREFIX_ERR0R 'E'
#define SNF_DEBUGM_PREFIX_PARAM 'P'
#define SNF_DEBUGM_PREFIX_QSTATS 'S'
#define SNF_DEBUGM_PREFIX_TIMESYNC 'T'
#define SNF_DEBUGM_PREFIX_IOCTL 'I'
#define SNF_DEBUGM_PREFIX_QEVENTS 'Q'
#define SNF_DEBUGM_PREFIX_ARISTA 'A'

#ifdef MAL_KERNEL
#define _SNF_DPRINTF(p,level,format,...)  do {                  \
        if (((p)->debug_mask & SNF_DEBUGM_ ## level) || SNF_DEBUGM_ERR0R == SNF_DEBUGM_ ## level) \
            MX_PRINT(("snf.%d.%d %c " format,   \
                    (p)->portnum, (p)->ring_id, SNF_DEBUGM_PREFIX_ ## level, \
                    __VA_ARGS__));                              \
        } while (0)
#else
extern int snf__getpid(void);
#define _SNF_DPRINTF(p,level,format,...)  do {                  \
        if (((p)->debug_mask & SNF_DEBUGM_ ## level) || SNF_DEBUGM_ERR0R == SNF_DEBUGM_ ## level) \
            fprintf((p)->debug_fp ? (p)->debug_fp : stderr, "%5u snf.%c.%-2d %c " format,      \
		    snf__getpid(), \
                    (p)->portnum >= 0 && (p)->portnum < SNF_PORTNUM_MAX ? \
                            '0' + (p)->portnum : 'X',                   \
                    (p)->ring_id, SNF_DEBUGM_PREFIX_ ## level, \
                    __VA_ARGS__);                               \
	    if ((p)->debug_fp) fflush((p)->debug_fp);		\
        } while (0)
#endif

#define SNF_DPRINTF(params,level,format,...) \
            _SNF_DPRINTF(params,level,format,__VA_ARGS__)


struct snf_recv_req;
typedef int (*snf_rss_fn_t)(struct snf_recv_req *, void *, uint32_t *);

#ifndef MAL_KERNEL
#include <stdio.h> /* FILE */
#endif
struct snf__params {
  uint32_t     debug_mask;
  uint32_t     portnum;
  int32_t      ring_id;
  snf_rss_fn_t rss_hash_fn;
  void        *rss_context;
  int32_t      port_ring_id; /* DUP_MA */
#ifndef MAL_KERNEL
  const char  *debug_filename;
  FILE        *debug_fp;
  char         debug_file_buf[128];
#endif
};

struct snf__flow_state {
  uint32_t flow_desc;
  uint32_t flow_data;
};

typedef struct { volatile uint32_t slock; } snf__spinlock_t;

/* Hardware will have interrupts.  Will use them to self clock transmits.
 * - When opening the send endpoint, enable the interrupt.
 * - Flush when there is nothing (i.e. first packet).
 * - Flush when there are enough buffered packets.
 */
/* These three options are mutually exclusive */
#define K35SNF_FLUSH_EVERY_PACKET 0  /* Use this to see how bad PIOs can be */
#define K35SNF_FLUSH_CHUNK        1  /* Use this to see max throughput */
#define K35SNF_FLUSH_INTERRUPT    2  /* Interrupt driven, same as SNF v2/3 */

#define K35SNF_FLUSH_STRATEGY K35SNF_FLUSH_INTERRUPT

#if K35SNF_FLUSH_STRATEGY == K35SNF_FLUSH_INTERRUPT
/* Shared between the user and the driver to coordinate flushing of buffered
 * packets to the NIC.  Both read and update the fields.  See SNF v2 and v3's
 * interrupt based flushing.
 */
struct k35snf__tx_shared_state {
  SNF__CACHEALIGN_BEGIN myri_endpt_flow_t ep_flow SNF__CACHEALIGN_END;
  SNF__PIOALIGN_BEGIN   snf__spinlock_t tx_lock   SNF__PIOALIGN_END;
  uint32_t queued_desc;
  uint32_t queued_length;
  uint32_t inflight;
  uint32_t data_avail;
  int acked; /* wrote to COUNT_ACK */
  uint64_t stat_pkt;
  uint64_t stat_byte;
};

/* The user library and the driver each has this structure to process
 * completions.  The pointer fields have different values depending on
 * user/kernel.  The others have the identical values for both.
 * See k35snf_tx_update_completion.
 */
struct k35snf__tx_shared_params {
  struct k35snf__tx_shared_state *ss;
  uint32_t compl_threshold;
  volatile uint32_t *prod_doorbell; /* K35SNF_TX_DESC_PRODW_ADVANCE */
  volatile uint32_t *compl_ack; /* K35SNF_TX_COMPL_COUNT_ACK */
  volatile k35snf_tx_completion_t *compl_page;
};
#endif

#define SNF__RX_STATE_SIZE (SNF_ALIGNUP(sizeof(struct snf__rx_state) + 256, SNF_VPAGE_SIZE))
#define SNF__RX_STATE(rx_s_base,ring_id) \
  ((struct snf__rx_state *)(uintptr_t)((uintptr_t)rx_s_base + ring_id * SNF__RX_STATE_SIZE))
#define SNF_VPAGE_IDX(rq_off) (((uint64_t) rq_off) >> SNF_VPAGE_SHIFT)
#define SNF_FLOW_DATA(data_cnt)  ((uint32_t)((((uint64_t)data_cnt)) >> SNF_VPAGE_SHIFT))
#define SNF_DESC_VPAGES(cnt) (((uint64_t) cnt) >> SNF_DESC_SHIFT)
#define SNF_DATA_VPAGES(rq_off) SNF_VPAGE_IDX(rq_off)
/* K35SNF has a smaller 16B descriptor */
#define SNF_K35SNF_DESC_SHIFT 8 /* 256 descriptors per page */
#define SNF_K35SNF_DESC_VPAGES(cnt) (((uint64_t) cnt) >> SNF_K35SNF_DESC_SHIFT)

#if MYRI_SNF_PREHASH
#define PHX_SNF_PREHASH_DESC_PAD_V1   (512 - 64 - 4 - (PHX_RX_DESC_PKT_MAX * 4) - 32)
#define PHX_SNF_PREHASH_DESC_PAD_V2   (512 - 64 - 4 - (PHX_RX_DESC_PKT_MAX * 2))

/* 512B intermediate descriptor that we produce from phx_rx_desc_t. */
typedef struct {
  uint8_t seqnum;
  uint8_t pad0[63];
  uint32_t total_length;
  union {
    struct {
      uint8_t start_idx[32];
      struct {
        uint16_t off;
        uint8_t pad;
        uint8_t next;
      } pkt[PHX_RX_DESC_PKT_MAX];
      uint8_t pad[PHX_SNF_PREHASH_DESC_PAD_V1];
    } v1;
    struct {
      uint16_t hash_val[PHX_RX_DESC_PKT_MAX];
      uint8_t pad[PHX_SNF_PREHASH_DESC_PAD_V2];
    } v2;
  } u;
} phx_snf_prehash_desc_t;

/* 16B.  It is intentionally the same size as k35snf_rx_desc_t.
 * FIXME: Really need 16B?  hash_val is the only thing prehash code produces.
 * But then if we copy all the information like this, ring threads can avoid
 * touching the descriptor ring.
 */
typedef struct {
  uint64_t timestamp;
  uint16_t pktlen;
  uint8_t pad0;
  uint8_t next_ring;
  uint16_t hash_val;
  uint8_t pad1;
  uint8_t seqnum;
} k35snf_prehash_desc_t;
#endif

/* Flow control variables.  These are all 64 bits and never wrap. */
struct snf__rx_sync {
  uint64_t rs_desc_cnt;  /* Desc queue for current batch */
  uint64_t rs_data_head; /* Byte count */
  uint64_t rs_data_tail; /* Byte count */
};

struct snf__pq_state {
  /* The user updates these via kernel */
  struct snf__pq_state_rw  {
    uint32_t evq_prehash_seqnum;
    uint32_t evq_next_prehash_idx;
    uint64_t evq_desc_cnt;
    uint64_t rq_prehash_data_off; /* Offset in receive queue */
    uint64_t evq_ts_last;
  } rw;
  /* Set once by the kernel, immutable afterwards */
  uint32_t evq_last_prehash_idx;
};

struct snf__rx_state {
  struct snf__eventq_state  {
    uint32_t evq_seqnum;
    uint32_t evq_next_idx;
    uint32_t evq_last_idx;
    int32_t  evq_evinfo_idx;
    uint64_t evq_desc_cnt;
    uint64_t evq_data_head;
  } evq;

  struct snf__recvq_state {
    uint64_t rq_data_off; /* Offset in receive queue */

    /* if head == tail, the queue is empty */
    uint64_t rq_data_head; /* Where next packet will be written */
    uint64_t rq_data_tail; /* End of the last packet that was returned */
    uint64_t rq_data_tail_last; /* Last update */
    uint64_t rq_data_size;

    struct snf__recvq_stats {
      uint64_t rq_pkt_cnt;
      uint64_t rq_pkt_length;
      uint64_t rq_pkt_cnt_skip_rss;
      uint64_t rq_pkt_cnt_drop;
      uint64_t rq_neg_adj_event;
      uint64_t rq_coal[PHX_SNF_MAX_COALESCE_PKT];
    } stats;

    uint64_t rq_is_reattach;
    uint64_t pad;
    /* Pad so ts_last starts at a cache line.. */
    uint64_t rq_ev_ts_last;

    struct snf__eventq_evinfo {
      /* The entire buffer length occupied by the descriptor.  This includes
       * all pad bytes (packet alignments).
       */
      uint32_t ev_length;
      uint32_t ev_pkt_cnt;
      uint32_t ev_pkt_off;
      union {
        struct {
          struct {
            uint16_t len;
          } ev_pkt[PHX_RX_DESC_PKT_MAX];
        } allhash; /* non-prehash mode */
        struct {
          uint32_t ev_ring_id;
          uint32_t pad;
          struct {
            uint16_t len;
            uint16_t skip_off;
          } ev_pkt[PHX_RX_DESC_PKT_MAX];
        } prehash_v1;
      } u;
      uint64_t nanoseconds[PHX_RX_DESC_PKT_MAX];
    } ev_info;
  } recvq;
};

typedef struct {
  myri_set_endpt_t set_endpt;
  uint32_t num_rings;
  uint32_t open_flags;
  uint32_t rss_flags;
  uint32_t aggr_portmask; /* ports that belong to the aggregation group */
  int32_t  app_id; /* id of the calling process (DUP_MA mode) */
  uint32_t rx_map_flags;
  uint64_t rx_data_preamble_size;
  uint64_t rx_data_ring_size; /* without preamble */
  uint64_t rx_desc_ring_size;
  uint32_t board_type; /* SNF_BOARD_TYPE_ */
  uint32_t port_merge_flags;
} myri_snf_rx_params_t;

typedef struct {
  myri_set_endpt_t set_endpt;
  uint32_t phx_endpt;
  uint32_t phx_port;
  uint32_t dmasend_threshold;
  uint32_t nic_mem_tx_offset;
  uint32_t nic_mem_tx_len;
  uint32_t nic_mem_tx_pio_offset;
  uint32_t nic_mem_tx_pio_len;
  uint32_t flow_window_offset;
  uint32_t flow_window_len;
  uint32_t send_compcnt_offset;
  uint32_t send_compcnt_len;
  uint32_t tx_dmabuf_offset;
  uint32_t tx_dmabuf_len;
  uint32_t board_type; /* SNF_BOARD_TYPE_ */
  /* K35SNF.  The app specifies the ring sizes. */
  uint32_t tx_data_ring_offset; /* out */
  uint32_t tx_data_ring_len; /* in */
  uint32_t tx_desc_ring_offset; /* out */
  uint32_t tx_desc_ring_len; /* in */
  uint32_t tx_scale_factor;
} myri_snf_tx_params_t;

typedef struct {
  myri_set_endpt_t      set_endpt;
  int32_t               ring_id;  /* in-out */
  int32_t               port_ring_id;  /* DUP_MA mode, to aid debugging */
  myri_snf_rx_params_t  params;
  myri_clksync_nticks_t nticks;

  struct snf__flow_state flow;

  /* Kernel window, same as DBL */
  uint32_t kernel_window_offset;
  uint32_t kernel_window_len;

  /* No send_compcnt because this endpoint is receive only */

  /* Sniffer specific meta data (e.g. software hashing) */
  uint64_t vstate_all_offset;
  uint64_t vstate_all_len;

  /* Huge sniffer rings:
   * The data ring has the preamble, which is not part of the DBL ring.
   * The descriptor ring has no preamble, and the whole thing is the
   * DBL ring.
   */
  uint64_t rx_desc_ring_offset;
  uint64_t rx_desc_ring_len;
  uint64_t rx_data_ring_offset;
  uint64_t rx_data_ring_len;
} myri_snf_rx_attach_t;

typedef struct {
  myri_set_endpt_t      set_endpt;
  myri_snf_rx_params_t  params;

  /* No mapping the board */
  /* No kernel window */
  /* No flow control */

  /* PQ specific meta data */
  uint64_t pq_state_offset;
  uint64_t pq_state_len;

  /* Prehash descriptor ring, same size as the descriptor ring */
  uint64_t prehash_desc_ring_offset;
  uint64_t prehash_desc_ring_len;

  /* Huge sniffer rings */
  uint64_t rx_desc_ring_offset;
  uint64_t rx_desc_ring_len;
  uint64_t rx_data_ring_offset;
  uint64_t rx_data_ring_len;
} myri_snf_pq_attach_t;

typedef struct {
  uint64_t nic_pkt_bad;
  uint64_t nic_pkt_recv;
  uint64_t nic_bytes_recv;
  uint64_t nic_pkt_no_buffer;
  uint64_t nic_pkt_dropped;
} phx_snf_stats_t;

typedef struct {
  uint32_t  port_id;
  uint32_t  ring_id;
  void     *evq_base; /* used for prehash + aggregation */
} myri_snf_port_group_t;

typedef struct {
  union {
    uint32_t portmask;
    uint32_t portnum; /* locked port */
  } u;
  int timeout_ms;
} myri_snf_pq_group_lock_t;

static inline
int snf__ffs(int x)
{
#if !MAL_OS_WINDOWS
  int r;
  __asm__("bsfl %1,%0\n\t"
	  "cmovzl %2,%0" 
	  : "=r" (r) : "rm" (x), "r" (-1));
  return r+1;
#else
  int r = 0;
  return _BitScanForward(&r, x) ? r+1 : 0;
#endif
}

/* First call must be with idx_prev==-1.
 * Use case really intended to be through snf__ffs_foreach */
static inline
int snf__ffs_next(int mask, int idx_prev)
{
  int idx = idx_prev+1;
  mask &= ~( (1<<idx)-1 );
  if (mask == 0 || idx >= 32)
    return -1;
  return snf__ffs(mask) - 1;
}

#define SNF_FFS_FOREACH(mask, i) for (i = -1; (i = snf__ffs_next(mask, i)) >= 0; )

static inline
uint32_t snf__cmpswap_u32(uint32_t *p, uint32_t oldv, uint32_t newv)
{
#if !MAL_OS_WINDOWS
  __asm__("lock\n\t"
          "cmpxchgl %1,%2"
          : "=a" (oldv) : "q" (newv), "m" (*p), "0" (oldv) : "memory");
  return oldv;
#else
  return _InterlockedCompareExchange(p, newv, oldv);
#endif
}

static inline
int snf__spin_trylock(snf__spinlock_t *lock)
{
#if !MAL_OS_WINDOWS
  int oldv;

  __asm__ __volatile__("xchgl %0, %1"
                       :"=q" (oldv), "=m" (lock->slock)
                       :"0" (0) : "memory");
  return oldv > 0;
#else
  return _InterlockedExchange(&lock->slock, 0) > 0;
#endif
}

static inline
void snf__spin_init(snf__spinlock_t *lock)
{
  lock->slock = 1;
}

static inline
void snf__spin_lock(snf__spinlock_t *lock)
{
  while (!snf__spin_trylock(lock))
#if !MAL_OS_WINDOWS
       __asm__ __volatile__("rep;nop": : :"memory");
#else
       ; /* Not sure what to replace this with. */
#endif
}

static inline
void snf__spin_unlock(snf__spinlock_t *lock)
{
#if !MAL_OS_WINDOWS
  __asm__ __volatile__("movl $1, %0"
                       :"=m" (lock->slock)
                       : : "memory");
#else
  _InterlockedExchange(&lock->slock, 1);
#endif
}

#if K35SNF_FLUSH_STRATEGY == K35SNF_FLUSH_INTERRUPT
static inline
void
k35snf_tx_write_compl_ack(struct k35snf__tx_shared_params *sp)
{
  struct k35snf__tx_shared_state *ss = sp->ss;
  uint32_t ack;

  if (!ss->acked && (ack = ss->inflight)) {
    /* Request an update when min(inflight, threshold) packets are sent */
    if (ack > sp->compl_threshold)
      ack = sp->compl_threshold;
    *sp->compl_ack = ack;
    ss->acked = 1;
    MAL_STBAR();
  }
}

static inline
int
k35snf_tx_update_completion(struct k35snf__tx_shared_params *sp)
{
  struct k35snf__tx_shared_state *ss;
  uint32_t compl_packets;
  int done = 0;

  compl_packets = sp->compl_page->packets;
  if (!compl_packets)
    return 0; /* No new completions */
  sp->compl_page->packets = 0;
  ss = sp->ss;
  ss->inflight -= compl_packets;
  ss->data_avail += sp->compl_page->bytes;
  ss->acked = 0;
  k35snf_tx_write_compl_ack(sp);
  return done;
}

static inline
void
k35snf_tx_flush(struct k35snf__tx_shared_params *sp)
{
  struct k35snf__tx_shared_state *ss = sp->ss;

  MAL_STBAR();
  *sp->prod_doorbell = ss->queued_desc;
  MAL_STBAR();
  ss->inflight += ss->queued_desc;
  ss->stat_pkt += ss->queued_desc;
  ss->stat_byte += ss->queued_length;
  ss->queued_desc = 0;
  ss->queued_length = 0;
  k35snf_tx_write_compl_ack(sp);
}
#endif /* K35SNF_FLUSH_STRATEGY == K35SNF_FLUSH_INTERRUPT */

#endif /* _phx_snf_io_h */
