
/*************************************************************************
 * The contents of this file are subject to the MYRICOM ABSTRACTION      *
 * LAYER (MAL) SOFTWARE AND DOCUMENTATION LICENSE (the "License").       *
 * User may not use this file except in compliance with the License.     *
 * The full text of the License can found in LICENSE.TXT                 *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2012 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

/*
 * These are mal_ utility definitions that do not need tons of other
 * includes like mal.h does.
 */

#ifdef __WORDSIZE
#define MAL_WORDSIZE __WORDSIZE
#elif defined(__x86_64) || defined(_M_IA64) || defined(_M_AMD64)
#define MAL_WORDSIZE 64
#else
#define MAL_WORDSIZE 32
#endif

#if MAL_WORDSIZE == 64
#define MAL_PIOWORDSIZE 8
typedef uint64_t mal_pioword_t;
#else
#define MAL_PIOWORDSIZE 4
typedef uint32_t mal_pioword_t;
#endif

typedef uintptr_t mal_wcsum;

uint32_t mal_piocopy_csum(uintptr_t *piobuf, const uint8_t *src,
                          uint32_t len, uint32_t *offp);
MAL_FUNC(int) mal_get_macaddr(const char *ifname, uint8_t mac_addr[6]);
MAL_FUNC(uint16_t) mal_get_vlan_tag(const char *ifname);

#if !MAL_OS_WINDOWS
#define add_carry32(sum, v)  __asm__("addl %1, %0 ;"                   \
                                     "adcl $0, %0 ;"                   \
                                     : "=r" (sum)                      \
                                     : "g" ((uint32_t) v), "0" (sum))

#define add_carry64(sum, v)  __asm__("addq %1, %0 ;"                   \
                                     "adcq $0, %0 ;"                   \
                                     : "=r" (sum)                      \
                                     : "g" ((uint64_t) v), "0" (sum))
#else
#define add_carry32(sum, v) add__carry32(&sum, v)
#define add_carry64(sum, v) add__carry64(&sum, v)

static inline
void add__carry32(uint32_t *sum, uint32_t v)
{
  *sum += v;
  if (*sum < v) {
    (*sum)++;
  }
}

static inline
void add__carry64(uint64_t *sum, uint64_t v)
{
  *sum += v;
  if (*sum < v) {
    (*sum)++;
  }
}
#endif

static inline
uint32_t
do_add_carry32(uint32_t csum, uint32_t val)
{
  add_carry32(csum, val);
  return csum;
}


static
mal_wcsum
wcsum_copy_odd(uint8_t *dst, const uint8_t *src, int len, mal_wcsum wcsum)
{
  mal_assert(len > 0);
  mal_assert(len < sizeof(mal_wcsum));
  mal_assert(((uintptr_t) dst & (sizeof(uintptr_t)-1)) == 
            ((uintptr_t) src & (sizeof(uintptr_t)-1)));

  if (1 & (uintptr_t) src) {
    *dst = *src;
    wcsum += (*src << 8);
    dst++;
    src++;
    len--;
  }
  if (len >= 2) { /* at least 2 bytes */
    if (2 & (uintptr_t) src) {
      wcsum += *(uint16_t *)src;
      *(uint16_t *)dst = *(uint16_t *)src;
      dst += 2;
      src += 2;
      len -= 2;
    }
    if (len >= 4) { /* at least 4 bytes */
        mal_assert(((uintptr_t) src & 0x3) == 0);
        wcsum += *(uint32_t *)src;
        *(uint32_t *)dst = *(uint32_t *) src;
        dst += 4;
        src += 4;
        len -= 4;
    }
    if (len & 2) {
      wcsum += *(uint16_t *) src;
      *(uint16_t *)dst = *(uint16_t *)src;
      dst += 2;
      src += 2;
      len -= 2;
    }
  }
  if (len) {
    wcsum += *src;
    *dst = *src;
  }

  return wcsum;
}

uint32_t
mal_piocopy_csum(uintptr_t *piobuf, const uint8_t *src, uint32_t len, uint32_t *offp)
{
  mal_wcsum wcsum = 0;
  uint32_t csum;
  uintptr_t tmp = 0, *bufp = (uintptr_t *) src;
  uint32_t ubytes = 0;
  uint32_t off = (uint32_t)((uintptr_t) src & (MAL_PIOWORDSIZE-1));

  mal_assert(len > 0 && (((uintptr_t) piobuf & (MAL_PIOWORDSIZE-1)) == 0));

  if (unlikely(off)) {
    ubytes = MAL_PIOWORDSIZE - off;
    if (len < ubytes)
      ubytes = len;
    wcsum = wcsum_copy_odd((uint8_t *)&tmp + off, src, ubytes, wcsum);
    bufp = (uintptr_t *)(src + ubytes);
    len -= ubytes;
    *piobuf++ = tmp;
  }

  while (len > (MAL_PIOWORDSIZE-1)) {
    tmp = *bufp++;
#if MAL_PIOWORDSIZE == 8
    add_carry64(wcsum, tmp);
#else
    add_carry32(wcsum, tmp);
#endif
    *piobuf++ = tmp;
    len -= MAL_PIOWORDSIZE;
  }

  if (unlikely(len)) {
    uint8_t *pdst = (uint8_t *)&tmp;
    uint8_t *psrc = (uint8_t *)bufp;
    if (len & 4) {
#if MAL_PIOWORDSIZE == 8
      add_carry64(wcsum, *(uint32_t *)psrc);
#else
      add_carry32(wcsum, *(uint32_t *)psrc);
#endif
      *(uint32_t *)pdst = *(uint32_t *)psrc;
      pdst += 4; psrc += 4;
    }
    if (len & 2) {
#if MAL_PIOWORDSIZE == 8
      add_carry64(wcsum, *(uint16_t *)psrc);
#else
      add_carry32(wcsum, *(uint16_t *)psrc);
#endif
      *(uint16_t *)pdst = *(uint16_t *)psrc;
      pdst += 2; psrc += 2;
    }
    if (len & 1) {
#if MAL_PIOWORDSIZE == 8
      add_carry64(wcsum, *(uint8_t *)psrc);
#else
      add_carry32(wcsum, *(uint8_t *)psrc);
#endif
      *(uint8_t *)pdst = *(uint8_t *)psrc;
    }
    *piobuf = tmp;
  }

  MAL_STBAR();

#if MAL_PIOWORDSIZE == 8
  csum = do_add_carry32((wcsum>>32), wcsum & 0xffffffff);
#else
  csum = wcsum;
#endif

  /* When 1-byte mis-aligned, the 16-bit checksum needs to be reversed */
  if (unlikely(off & 1)) {
    csum = (csum >> 16) + (csum & 0xffff);
    csum += (csum >> 16);
    csum = ((csum >> 8) & 0xff) | ((csum & 0xff) << 8);
  }

  *offp = off;
  return csum;
}


#if 0
/*
 * toolkit subroutine to calculate tcp checksum from scratch
 * for validation purposes
 */
static inline
uint32_t
do_sum(void *buf, int len)
{
  uint32_t sum;
  uint16_t *p;
  uint16_t *end;

  p = buf;
  end = (uint16_t *)((char *)p + len);
  sum = 0;
  while (p < end) {
    sum += *p++;
  }

  return sum;
}


static inline
void
tcp_csum(
  char *pkt, int len)
{
  struct iphdr *ip;
  struct tcphdr *tcp;
  uint32_t sum;
  uint32_t check;
  int t;

  if (len & 1) {
    pkt[len] = 0;		/* add a 0 in case pad needed */
  }


  ip = (struct iphdr *)((char *)pkt + 14);
  tcp = (struct tcphdr *)((char *)ip + sizeof(*ip));

  check = tcp->check;
  tcp->check = 0;
  sum = do_sum((char *)tcp-8, len-(sizeof(*ip)+14)+8);
  t = 6 + len - (sizeof(*ip)+14);
  sum += htons(t);
  sum = (sum & 0xFFFF) + (sum >> 16);
  sum = (sum & 0xFFFF) + (sum >> 16);
  sum = sum ^ 0xffff;

  printk("len=%d, check=0x%x  sum=0x%x\n", len, check, sum);
  tcp->check = check;
}
#endif
