/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2012 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "mx_misc.h"
#include "mx_pio.h"
#include "mcp_config.h"
#include "mal_stbar.h"


int
mx_mcp_command(mx_instance_state_t *is, enum mcp_cmd_type type, uint32_t data0, 
	       uint32_t data1, uint32_t data2, uint32_t *result)
{
  struct mcp_cmd *cmd;
  struct mcp_cmd_done *done;
  uint64_t buf[64/sizeof(uint64_t)];
  int sleep, status = 0;

  done = (struct mcp_cmd_done *)(is->sysbox.addr + MCP_SYSBOX_OFF_CMD);

  /* FIXME: remove the spinlock and check that the callers are serialized */
  mx_spin_lock(&is->cmd_spinlock);

  /* ensure cmd is aligned */
  cmd = (struct mcp_cmd *)buf;

  cmd->type = htonl((uint32_t)type);
  cmd->data0 = htonl(data0);
  cmd->data1 = htonl(data1);
  cmd->data2 = htonl(data2);
  cmd->done_addr.high = htonl(is->sysbox.pin.dma.high);
  cmd->done_addr.low = htonl(is->sysbox.pin.dma.low + MCP_SYSBOX_OFF_CMD);

  done->status = -1;
  MAL_STBAR();
  mx_pio_memcpy((char *)is->lanai.sram + MCP_CMD_ADDR, 
		buf, sizeof(struct mcp_cmd), MX_PIO_FLUSH);

  
  /* wait for completion (commands should complete quickly, not worth 
     sleeping over it) */
  sleep = MCP_COMMAND_TIMEOUT * 1000;
  while ((done->status == -1) && (sleep > 0)) {
    sleep -= 10;
    mx_spin(10);
    MAL_READBAR();
  }

  if (done->status == -1) {
    /* command timed out */
    if (mx_is_dead(is)) {
      status = -EIO;
      goto abort_with_lock;
    }
    MX_WARN(("command %d timed out, marking instance %d dead\n", type, is->id));
    mx_mark_board_dead(is, MX_DEAD_COMMAND_TIMEOUT, type);
  }
  if (mx_is_dead(is)) {
    status = -EIO;
    goto abort_with_lock;
  }
  
  if (result)
    *result = ntohl(done->data);
  
  switch (ntohl(done->status)) {
  case MCP_CMD_OK:
    status = 0;
    break;
    
  case MCP_CMD_ERROR_RANGE:
    status = -ERANGE;
    break;
    
  case MCP_CMD_ERROR_BUSY:
    status = -EBUSY;
    break;
    
  case MCP_CMD_ERROR_CLOSED:
    status = -EBADF;
    break;

  case MCP_CMD_UNKNOWN:
    status = -ENOSYS;
    break;
    
  case MCP_CMD_ERROR_LICENSE:
    status = -EACCES;
    break;
    
  default:
    status = -ENXIO;
    break;
  }

 abort_with_lock:
  mx_spin_unlock(&is->cmd_spinlock);

  return status;
}

