#ifndef _MYRI_SNF_FWD_H
#define _MYRI_SNF_FWD_H

#ifdef MYRI_PHOENIX
#error "Do not include this file for Phoenix build"
#endif

#define SNF_PORTGROUP_MAX_MEMBERS 8

extern myri_endpt_ops_t myri_ops_snf;

void myri_snf_init_ops(void);

enum snf_rx_state { SNF_RX_S_FREE = 0,
                    SNF_RX_S_INIT = 1,
                    SNF_RX_S_STARTED = 2,
                    SNF_RX_S_STOPPED = 3 };

struct myri_snf_waiter;
STAILQ_HEAD(snf_waitq, myri_snf_waiter);

struct myri_snf_rx_state;
struct myri_snf_waiter {
  TAILQ_ENTRY(myri_snf_waiter) link;    /* waitq link */
  TAILQ_ENTRY(myri_snf_waiter) pg_link; /* portgroup link */
  TAILQ_ENTRY(myri_snf_waiter) wake_link; /* wakeup link */
  int                intr_requested;
  int                intr_wake_pending;
  struct myri_snf_rx_state *snf;
  mx_sync_t           *wait_sync;
};

typedef
union snf_endpt_state {
  struct {
    int rx_is_attached;
    struct snf_dup_ma_app_state *app_info; /* DUP_MA mode. */
  } es_rx_h;

  struct {
    struct snf__rx_state  *rx_s;
    struct snf__pq_state  *pq_s;
    struct myri_snf_waiter rx_waitsync;
    struct myri_snf_waiter vq_waitsync;
    int                    rx_ring_id;
    mx_endpt_state_t      *rx_pg; /* If in a portgroup */
    struct snf_dup_ma_ring_state *app_ring_info; /* DUP_MA mode. */
  } es_rx;

  struct {
    mx_spinlock_t   tx_lock;
    int             tx_intr_requested;
    struct snf__tx_shared_params
                    tx_sp;
  } es_tx;

  struct snf_endpt_portgroup {
    /* One ring per board, max */
    struct myri_snf_waiter       pg_waiters[SNF_PORTGROUP_MAX_MEMBERS];
    TAILQ_HEAD(,myri_snf_waiter) pg_waitq;
    /* prehash + aggregation mode */
    struct myri_snf_rx_state    *pq_locked_port;
    struct myri_snf_waiter       vq_waitsync;
    uint32_t                     port_mask;
    /* The endpoint's user va. Used to check if the ring is empty.
     * The index is the port number.
     */
    mcp_snf_rx_desc_t           *evq_base[SNF_PORTGROUP_MAX_MEMBERS];
  } es_pg;
}
snf_endpt_state_t;

struct snf__rx;

typedef struct myri_snf_rx_state {
  mx_instance_state_t   *is;
  mx_sync_t              sync;
  uint32_t               owner_pid;
  uint32_t               num_rings_attached;
  uint32_t		 num_rings_requested;
  int                    refcount;
  enum snf_rx_state      state;
  myri_snf_rx_params_t   params;

  struct snf__rx_state  *rx_s;
  unsigned               rx_s_length;
  mx_spinlock_t          pq_lock;
  struct mx_endpt_state *pq_locked_es;
  struct snf__pq_state  *pq_s;
  unsigned               pq_s_length;
  int                    pq_refcount;
  int                    pq_group_pq_waiter;
  uint32_t		 rx_desc_ring_count;
  uint64_t              *rx_doorbell_be64;

  mx_huge_copyblock_t desc_cb;
  mx_huge_copyblock_t data_cb;

  /* State for synchronizing with NIC */
  struct snf_sync_state {
    mx_spinlock_t             rs_lock;
    uint32_t		      attach_mask;
    uint32_t		      detach_mask;
    struct snf__rx_sync       rs;
    struct snf__rx_sync	      r_rs[MYRI_SNF_MAX_RINGS];
  } sync_s;

  mx_endpt_state_t *ring_map_ep[MYRI_SNF_MAX_RINGS];

  /* RX interrupts and virtual queue waiters */
  TAILQ_HEAD(myri_snf_waitq, myri_snf_waiter)
                  rx_intr_waitq, vq_waitq;
  int             rx_intr_requested;
  mx_spinlock_t   rx_intr_lock;

  /* DUP_MA mode.  Per-process/application ring information. */
  struct snf_dup_ma_app_state {
    int app_id;
    int refcount;
    uint32_t num_rings_requested;
    uint32_t rss_flags;
    uint32_t num_rings_attached;
    uint32_t attach_mask;
  } app_info[MYRI_SNF_MAX_RINGS];
  struct snf_dup_ma_ring_state {
    int app_info_idx;      /* index into app_info */
    uint32_t port_ring_id; /* port-wide id used in the driver */
    uint32_t hash_ring_id; /* per-application id, used for hashing only */
  } app_ring_info[MYRI_SNF_MAX_RINGS];
} myri_snf_rx_state_t;

void     myri_snf_rx_teardown(mx_endpt_state_t *es);
uint32_t myri_snf_intr_req(mx_endpt_state_t *es, uint32_t intr_flags);
int      myri_snf_intr_tx_try_flush(mx_endpt_state_t *es);
int      myri_snf_find_node_id(mx_endpt_state_t *es);

/* OS-specific */
void myri_snf_handle_tx_intr(mx_endpt_state_t *es);

#endif /* _MYRI_SNF_FWD_H */

