/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2011 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_instance.h"

#ifdef MYRI_PHOENIX
#error "Do not include this file for Phoenix build"
#endif

uint32_t myri_snf_rings = 0;
uint32_t myri_snf_flags = -1;

void myri_dump_snf_state(myri_snf_rx_state_t *snf);
int myri_ether_soft_rx(mx_instance_state_t *is, myri_soft_rx_t *r);

static void snf_rx_teardown(mx_endpt_state_t *es);
static int snf_rx_open_handle(mx_endpt_state_t *es, myri_snf_rx_params_t *x);
static int snf_rx_attach(mx_endpt_state_t *es, myri_snf_rx_attach_t *x);
static int snf_pq_attach(mx_endpt_state_t *es, myri_snf_pq_attach_t *x);
static int snf_tx_endpt_enable(mx_endpt_state_t *es, myri_snf_tx_params_t *x);
static int snf_rx_portgroup_del(mx_endpt_state_t *es, mx_endpt_state_t *es_rx_ring);
static void snf_handle_tx_intr(mx_endpt_state_t *es);
static void snf_intr_tx_flush_work(
#if defined(NDIS50_MINIPORT) || defined(NDIS51_MINIPORT)
				   NDIS_HANDLE wi,
#endif
				   void *arg
#if defined(NDIS60_MINIPORT)
				   , NDIS_HANDLE wi
#endif
);

static
int
snf_eop_alloc(mx_endpt_state_t *es, mx_instance_state_t *is, 
              myri_set_endpt_t *set_endpt)
{
  int min_endpt, max_endpt;
  enum myri_endpt_type es_type;
  int endpoint;
  union snf_endpt_state *ess = &es->snf;

  es_type = set_endpt->endpt_type;

  if (es_type == MYRI_ES_SNF_TX) {
    endpoint = set_endpt->endpoint;
    if (is->es[endpoint] != 0)
      return EBUSY;

    es->session_id = 1;

    ess->es_tx.tx_intr_requested = 0;
    mx_spin_lock_init(&ess->es_tx.tx_lock, is, es->endpt, "es_tx.tx_lock");
    MX_LXX_INIT_DELAYED_WORK(&es->arch.snf.tx_flush_work,
                             (void *) snf_intr_tx_flush_work, es);
    is->es[endpoint] = es;
    es->flags |= MX_ES_EARLY_ENABLE;
  }
  else {
    if (es_type == MYRI_ES_SNF_RX) {
      min_endpt = myri_max_endpoints;
      max_endpt = 2*myri_max_endpoints;
      es->flags |= MX_ES_EARLY_ENABLE;
      es->snf.es_rx.rx_ring_id = SNF_RING_UNUSED;
    }
    else if (es_type == MYRI_ES_SNF_RX_RING) {
      min_endpt = 2*myri_max_endpoints;
      max_endpt = 3*myri_max_endpoints;
      es->flags |= MX_ES_EARLY_ENABLE;
      es->snf.es_rx.rx_ring_id = SNF_RING_UNUSED;
    }
    else if (es_type == MYRI_ES_SNF_RX_GROUP) {
      min_endpt = 3*myri_max_endpoints;
      max_endpt = 4*myri_max_endpoints;
      TAILQ_INIT(&es->snf.es_pg.pg_waitq);
    }
    else if (es_type == MYRI_ES_SNF_HANDLE) {
      min_endpt = 4*myri_max_endpoints;
      max_endpt = 5*myri_max_endpoints;
    }
    else if (es_type == MYRI_ES_SNF_RX_PQ) {
      min_endpt = 5*myri_max_endpoints;
      max_endpt = 6*myri_max_endpoints;
      es->flags |= MX_ES_EARLY_ENABLE;
    }
    else
      return EINVAL;

    for (endpoint = min_endpt; endpoint < max_endpt; endpoint++) {
      if (is->es[endpoint] == 0)
          break;
    }
    if (endpoint == max_endpt)
      return EBUSY;
    es->flags |= MX_ES_SOFT;
    is->es[endpoint] = es;
  }

  es->endpt = endpoint;
  es->is = is;

  return 0;
}

/* XXX 
 * Sniffer requires early enable, can't support splitting opening
 * from enabling... yet.
 */
static
int
snf_eop_enable(mx_endpt_state_t *es, uaddr_t in)
{
  mx_instance_state_t *is = es->is;
  uaddr_t out = in;
  int status;

  mx_mutex_enter(&es->sync);

  switch (es->es_type) {

  case MYRI_ES_SNF_TX:
  {
    myri_snf_tx_params_t x;
#if !(MAL_OS_LINUX || MAL_OS_FREEBSD || MAL_OS_WINDOWS)
    status = EINVAL;
    break;
#endif

    if ((status = myri_endpt_enable_common_copyblocks(es))) {
      MX_WARN(("%s: endpt %d can't allocate common copyblocks\n", is->is_name, es->endpt));
      break;
    }
    if ((status = myri_endpt_enable_mcp(es))) {
      MX_WARN(("%s: endpt %d can't enable endpoint\n", is->is_name, es->endpt));
      break;
    }

    if ((status = mx_copyin(in, &x, sizeof(x), es->is_kernel)))
      break;
    status = snf_tx_endpt_enable(es, &x);
    if (status)
      break;
    status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
  }
  break;

  case MYRI_ES_SNF_RX_RING:
  {
    myri_snf_rx_attach_t x;
    if ((status = mx_copyin(in, &x, sizeof(x), es->is_kernel)))
      break;
    if ((status = snf_rx_attach(es, &x)))
      break;
    status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
  }
  break;

  case MYRI_ES_SNF_RX:
  {
    myri_snf_rx_params_t x;
    if ((status = mx_copyin(in, &x, sizeof(x), es->is_kernel)))
      break;
    if ((status = snf_rx_open_handle(es, &x)))
      break;
    status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
  }
  break;

  case MYRI_ES_SNF_RX_PQ:
  {
    myri_snf_pq_attach_t x;
    if ((status = mx_copyin(in, &x, sizeof(x), es->is_kernel)))
      break;
    if ((status = snf_pq_attach(es, &x)))
      break;
    status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
  }
  break;

  case MYRI_ES_SNF_HANDLE:
    status = 0;
    break;

  default:
    status = EINVAL;
  }

  mx_mutex_exit(&es->sync);

  return status;
}

static
void
snf_eop_intr_is(mx_instance_state_t *is, void *context)
{
  myri_snf_rx_state_t *snf = &is->snf;
  struct myri_snf_waitq waitq;
  struct myri_snf_waiter *waiter;
  int rx_intr_requested;

  if (snf->state == SNF_RX_S_FREE)
    return;

  /* Step 1: If interrupt was requested, save each wakeup for further
   * processing w/o lock held */
  mx_spin_lock(&snf->rx_intr_lock);
  rx_intr_requested = snf->rx_intr_requested;
  snf->rx_intr_requested = 0;
  TAILQ_INIT(&waitq);
  if (rx_intr_requested && !TAILQ_EMPTY(&snf->rx_intr_waitq)) {
    while ((waiter = TAILQ_FIRST(&snf->rx_intr_waitq)) != NULL) {
      TAILQ_REMOVE(&snf->rx_intr_waitq, waiter, link);
      waiter->intr_requested = 0;
      waiter->intr_wake_pending = 1;
      TAILQ_INSERT_TAIL(&waitq, waiter, wake_link);
    }
  }
  mx_spin_unlock(&snf->rx_intr_lock);

  /* Step 2: Do all the (potentially slow) wakes w/o lock */
  TAILQ_FOREACH(waiter, &waitq, wake_link) {
    if (waiter->wait_sync)
      mx_wake(waiter->wait_sync);
    waiter->intr_wake_pending = 0;
  }
}

static
int
snf_eop_intr_endpt(mx_endpt_state_t *es, int flags)
{
  union snf_endpt_state *ess = &es->snf;
  int status = 0;

  if (es->flags & MX_ES_CLOSING) /* stop trying to flush */
    return 0;

  if (es->es_type == MYRI_ES_SNF_TX) {
    if (flags & MCP_INTR_ENDPT_WAKE_TX) {
      int intr_requested;
      mx_spin_lock(&ess->es_tx.tx_lock);
      intr_requested = ess->es_tx.tx_intr_requested;
      ess->es_tx.tx_intr_requested = 0;
      mx_spin_unlock(&ess->es_tx.tx_lock);
      snf_handle_tx_intr(es);
      if (intr_requested)
        mx_wake(&es->wait_sync);
    }
    if (flags & MCP_INTR_ENDPT_HELPER) {
      /* caused by refcount going to 0 in send queue */
      snf_handle_tx_intr(es);
    }
  }
  else
    status = EINVAL;

  return status;
}

static
void
snf_eop_close(mx_endpt_state_t *es)
{
  mx_instance_state_t *is = es->is;

  mx_mutex_enter(&is->sync);
  mal_always_assert(es->endpt >= 0 && es->endpt < is->es_count);
  is->es[es->endpt] = 0;
  mx_mutex_exit(&is->sync);

  if (es->es_type == MYRI_ES_SNF_RX || es->es_type == MYRI_ES_SNF_RX_RING || es->es_type == MYRI_ES_SNF_RX_PQ)
    snf_rx_teardown(es);
  else if (es->es_type == MYRI_ES_SNF_TX) {
    myri_cancel_delayed_work(&es->arch.snf.tx_flush_work);
    mx_spin_lock_destroy(&es->snf.es_tx.tx_lock);
    mx_endpt_free_copyblock(es, &es->sendq);
    mx_endpt_free_huge_copyblock(es, &es->recvq);
    mx_endpt_free_copyblock(es, &es->eventq);
  }
  else if (es->es_type == MYRI_ES_SNF_RX_GROUP)
    snf_rx_portgroup_del(es, NULL);
}

static int
snf_tx_endpt_enable(mx_endpt_state_t *es, myri_snf_tx_params_t *x)
{
  int status;
  mx_get_copyblock_t desc;
  union snf_endpt_state *ess = &es->snf;
  struct snf__tx_shared_params *tx_sp = &ess->es_tx.tx_sp;

  if (es->es_type != MYRI_ES_SNF_TX)
    return EINVAL;

  status = myri_endpt_get_common_copyblocks(es, &desc, NULL);
  if (status)
    return status;

  /* Convert the common copyblock naming to the naming we have under SNF */
  x->tx_data_ring_size = desc.sendq_len;
  x->tx_data_ring_offset = desc.sendq_offset;
  x->tx_desc_ring_size = desc.user_mmapped_sram_len;
  x->tx_desc_ring_offset = desc.user_mmapped_sram_offset;
  x->tx_doorbell_size = desc.user_mmapped_zreq_len;
  x->tx_doorbell_offset = desc.user_mmapped_zreq_offset;
  x->tx_flow_size = desc.flow_window_len;
  x->tx_flow_offset  = desc.flow_window_offset;
  x->tx_compl_size = desc.send_compcnt_len;
  x->tx_compl_offset = desc.send_compcnt_offset;

  tx_sp->txd_cnt = MX_VPAGE_SIZE / (SNF_TX_DESC_PKT_MAX * 2);
  tx_sp->txd_dbaddr_be32 = (uint32_t *) es->user_mmapped_zreq.addr;
  tx_sp->txd_pio_base = (uintptr_t) es->user_mmapped_sram.addr;
  tx_sp->tx_s = (struct snf__tx_state *) es->flow_window;
  tx_sp->txd_compl_update = (uint8_t *) es->is->tx_done.addr + es->endpt;

  sprintf(es->opener.user_info, "tx ring %d", es->endpt);
  return 0;
}

void 
myri_dump_snf_state(myri_snf_rx_state_t *snf)
{
  int nrings;
  MX_WARN(("attach_mask=%#x, refcount=%d, state=%d, rx_s=%p\n",
        snf->sync_s.attach_mask, snf->refcount, snf->state, snf->rx_s));
  MX_WARN(("desc_cb [kva=%p, size=%lu, pins=%p, pinned_pages=%lu\n",
        (void *) snf->desc_cb.kva, (unsigned long) snf->desc_cb.size,
        snf->desc_cb.pins, snf->desc_cb.pinned_pages));
  MX_WARN(("data_cb [kva=%p, size=%lu, pins=%p, pinned_pages=%lu\n",
        (void *) snf->data_cb.kva, (unsigned long) snf->data_cb.size,
        snf->data_cb.pins, snf->data_cb.pinned_pages));
  nrings = snf->num_rings_requested;
  if (!nrings)
    nrings = MYRI_SNF_MAX_RINGS;
  MX_WARN(("rings_attached=%d\n", snf->num_rings_attached));
}

static
int
snf_eop_poll_events_pending(mx_endpt_state_t *es)
{
  int revents = 0;
  uint8_t *tx_done_vec;

  if (es->flow_window && es->es_type == MYRI_ES_SNF_TX) {
    tx_done_vec = (uint8_t *)(uintptr_t) es->is->tx_done.pin.va;
    if ((uint8_t) es->flow_window->tx_flow.flow_desc != tx_done_vec[es->endpt])
      revents |= MYRI_WAIT_TX_EVENT;
  }
  return revents;
}

static
uint32_t
snf_rx_flow_desc_latest(myri_snf_rx_state_t *snf)
{
  struct snf_sync_state *ss = &snf->sync_s;
  struct snf__rx_state *rx_s;
  int i, first = 1;
  uint32_t desc_cnt, desc_cnt_latest, desc_size;

  desc_cnt_latest = 0;

  mx_spin_lock(&ss->rs_lock);

  desc_size = snf->rx_desc_ring_count;

  SNF_FFS_FOREACH(ss->attach_mask, i) {
    rx_s = SNF__RX_STATE(snf->rx_s, i);
    desc_cnt = (uint32_t)rx_s->evq.evq_desc_cnt;
    if (first) {
	desc_cnt_latest = desc_cnt;
	first = 0;
    }
    else if (desc_cnt - desc_cnt_latest < desc_size)
	desc_cnt_latest = desc_cnt;
  }

  mx_spin_unlock(&ss->rs_lock);

  return desc_cnt_latest;
}

static
uint32_t
snf_intr_rx_group_req(mx_endpt_state_t *es)
{
  struct snf_endpt_portgroup *ep_pg = &es->snf.es_pg;
  struct myri_snf_waiter *waiter;
  uint32_t intr_flags_set = 0;
  uint32_t desc_cnt;
  unsigned long flags;
  myri_snf_rx_state_t *snf;

  flags = 0; /* -Wunused on non-linux */

  if (es->es_type != MYRI_ES_SNF_RX_GROUP) 
    return 0;

  TAILQ_FOREACH(waiter, &ep_pg->pg_waitq, pg_link) {
    snf = waiter->snf;
    mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);
    if (snf->rx_intr_requested == 0) {
      desc_cnt = snf_rx_flow_desc_latest(snf);
      MX_PIO_WRITE((uint32_t *)(snf->is->lanai.sram + 0x7c0000),
		   htonl(desc_cnt));
      intr_flags_set |= MYRI_WAIT_RX_EVENT;
      snf->rx_intr_requested = 1;
    }
    if (waiter->intr_requested == 0) {
      waiter->intr_requested = 1;
      TAILQ_INSERT_TAIL(&snf->rx_intr_waitq, waiter, link);
    }
    mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);
  }

  return intr_flags_set;
}

/* FIXME.  Duplicate code.  snf_eop_poll_wait and snf_intr_rx_group_req. */
static
int
snf_pq_group_wait(mx_endpt_state_t *es, int timeout_ms)
{
  struct snf_endpt_portgroup *ep_pg = &es->snf.es_pg;
  struct myri_snf_waiter *waiter;
  uint32_t desc_cnt;
  unsigned long flags;
  myri_snf_rx_state_t *snf;
  int status;

  if (timeout_ms < 0)
    timeout_ms = MX_MAX_WAIT;

  flags = 0;
  TAILQ_FOREACH(waiter, &ep_pg->pg_waitq, pg_link) {
    snf = waiter->snf;
    mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);
    if (snf->rx_intr_requested == 0) {
      /* This might be stale since we run this code after releasing pq locks.
       * So be it.
       */
      desc_cnt = (uint32_t)snf->pq_s->rw.evq_desc_cnt;
      MX_PIO_WRITE((uint32_t *)(snf->is->lanai.sram + 0x7c0000),
        htonl(desc_cnt));
      snf->rx_intr_requested = 1;
    }
    if (waiter->intr_requested == 0) {
      waiter->intr_requested = 1;
      TAILQ_INSERT_TAIL(&snf->rx_intr_waitq, waiter, link);
    }
    mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);
  }
  status = mx_sleep(&es->wait_sync, timeout_ms, MX_SLEEP_INTR);
  return status;
}

static
int
snf_vqwait_req(mx_endpt_state_t *es, int timeout_ms)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  union snf_endpt_state *ess = &es->snf;
  struct myri_snf_waiter *vq_wait;
  int status;

  if (timeout_ms < 0)
    timeout_ms = MX_MAX_WAIT;

  if (es->es_type == MYRI_ES_SNF_RX_GROUP)
    vq_wait = &ess->es_pg.vq_waitsync;
  else
    vq_wait = &ess->es_rx.vq_waitsync;

  /* FIXME.  Use a lighter spinlock instead of heavy mutex. */
  mx_mutex_enter(&snf->sync);
  if (vq_wait->intr_requested == 0) {
    vq_wait->intr_requested = 1;
    TAILQ_INSERT_TAIL(&snf->vq_waitq, vq_wait, link);
  }

  mx_mutex_exit(&snf->sync);
  status = mx_sleep(&es->vq_sync, timeout_ms, MX_SLEEP_INTR);
  return status;
}

static
void
snf_vqwake_req(myri_snf_rx_state_t *snf, int locked)
{
  struct myri_snf_waiter *vq_waiter;

  if (!locked)
   mx_mutex_enter(&snf->sync);

  while ((vq_waiter = TAILQ_FIRST(&snf->vq_waitq)) != NULL) {
    vq_waiter->intr_requested = 0;
    if (vq_waiter->wait_sync)
      mx_wake(vq_waiter->wait_sync);
    TAILQ_REMOVE(&snf->vq_waitq, vq_waiter, link);
  }

  if (!locked)
    mx_mutex_exit(&snf->sync);
}

static
uint32_t
snf_intr_req(mx_endpt_state_t *es, uint32_t intr_flags)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  union snf_endpt_state *ess = &es->snf;
  unsigned long flags;
  uint32_t intr_flags_set = 0;
  uint32_t evq_desc_cnt_be;

  flags = 0; /* -Wunused on non-linux */

  /* Ensure only one interrupt request is in flight */
  if (intr_flags & MYRI_WAIT_RX_EVENT) {

    mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);

    if (es->es_type == MYRI_ES_SNF_RX_PQ)
      evq_desc_cnt_be = htonl((uint32_t)ess->es_rx.pq_s->rw.evq_desc_cnt);
    else
      evq_desc_cnt_be = htonl((uint32_t)ess->es_rx.rx_s->evq.evq_desc_cnt);

    if (snf->rx_intr_requested == 0) {
      MX_PIO_WRITE((uint32_t *)(es->is->lanai.sram + 0x7c0000), evq_desc_cnt_be);
      intr_flags_set |= MYRI_WAIT_RX_EVENT;
      TAILQ_INIT(&snf->rx_intr_waitq);
      snf->rx_intr_requested = 1;
    }
    if (ess->es_rx.rx_waitsync.intr_requested == 0) {
      ess->es_rx.rx_waitsync.intr_requested = 1;
      TAILQ_INSERT_TAIL(&snf->rx_intr_waitq, &ess->es_rx.rx_waitsync, link);
    }
    mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);
  }
  if ((intr_flags & MYRI_WAIT_TX_EVENT) && es->flow_window) {
    mal_assert(es->endpt < myri_max_endpoints); /* real mcp endpoints only */
    mx_spin_lock_irqsave(&ess->es_tx.tx_lock, flags);
    if (ess->es_tx.tx_intr_requested == 0) {
      MX_PIO_WRITE((uint32_t *)(es->is->lanai.sram + 0x7c0004 + (es->endpt << 4)),
                   htonl(es->flow_window->tx_flow.flow_desc));
      ess->es_tx.tx_intr_requested = 1;
      intr_flags_set |= MYRI_WAIT_TX_EVENT;
    }
    mx_spin_unlock_irqrestore(&ess->es_tx.tx_lock, flags);
  }
  if (intr_flags_set) { /* Requested at least one interrupt, flush doorbell write */
    MAL_STBAR();
  }
  return intr_flags_set;
}

static
void
snf_eop_poll_prep_wait(mx_endpt_state_t *es)
{
  /* Only TX interrupt for now */
  snf_intr_req(es, MYRI_WAIT_TX_EVENT);
}


static
int
snf_intr_tx_try_flush(mx_endpt_state_t *es)
{
  struct snf__tx_state *ss = es->snf.es_tx.tx_sp.tx_s;

  if (es->flags & MX_ES_CLOSING) /* stop trying to flush */
    return 1;

  if (!snf__spin_trylock(&ss->tx_lock))
    return 0;

  snf__tx_update_completion(&es->snf.es_tx.tx_sp);
  if (ss->txd.d_cnt && ss->dq_inflight == 0) {
    snf__tx_pioflush(&es->snf.es_tx.tx_sp);
    ss->tx_flush_intr_cnt++;
  }
  snf__spin_unlock(&ss->tx_lock);
  return 1;
}

static
int
snf_eop_poll_wait(mx_endpt_state_t *es, int timeout_ms, int events)
{
  int status;

  if (timeout_ms < 0)
    timeout_ms = MX_MAX_WAIT;

  if (timeout_ms == 0) {
    /* We're not going to block below in sleep, but claim the interrupt if it
     * is available.  This is mostly useful when poll_wait is called with
     * timeout_ms==0 when the endpoint is used with poll() */
    status = mx_sleep(&es->wait_sync, 0, MX_SLEEP_INTR);
    if (status != EAGAIN) {
      MX_WARN(("unexpected return from poll_wait: %d\n", status));
      return status;
    }
  }
  else {
    if (events) {
      if (es->es_type == MYRI_ES_SNF_RX_GROUP) 
        snf_intr_rx_group_req(es);
      else
        snf_intr_req(es, events);
    }
    status = mx_sleep(&es->wait_sync, timeout_ms, MX_SLEEP_INTR);

    return status;
  }

  return status;
}

static
int
snf_eop_mmap(mx_endpt_state_t *es, unsigned long req, 
             unsigned long *off, void **kva,
	     int *mem_type, int *readonly, mx_page_pin_t **pin)
{
  void *tmpkva;
  mx_instance_state_t *is = es->is;
  int index;

  if (es->es_type == MYRI_ES_SNF_TX)
    return ENOENT;

  if (es->es_type == MYRI_ES_SNF_RX_PQ) {
    /* Sniffer-only mappings start at the end of the mcp-common endpoints */
    if (is->snf.pq_s) {        /* S.1 PQ */
      unsigned long rel_off = req - *off;
      tmpkva = (void *)((char *)is->snf.pq_s + rel_off);
      *off += is->snf.pq_s_length;
      if (req < *off) {
        *kva = tmpkva;
        *mem_type = MX_MEM_HOSTMEM;
        *readonly = 1;
        return 0;
      }
    }
  } else { 
    /* mapping S.1: snf rx doorbell page */
    if (is->snf.rx_doorbell_be64 && is->snf.rx_s) {
      tmpkva = (void *)((char *)is->snf.rx_doorbell_be64 + (req - *off));
      *off += MX_PAGE_SIZE;
      if (req < *off) {
        *kva = tmpkva;
        *mem_type = MX_MEM_SRAM;
        *readonly = 0;
        return 0;
      }
    }
  
    /* mapping S.2: snf rx state */
    if (is->snf.rx_s) {
      unsigned long rel_off = req - *off;
      int ring_req = rel_off / SNF__RX_STATE_SIZE;
      tmpkva = (void *)((char *)is->snf.rx_s + rel_off);
      *off += is->snf.rx_s_length;
      if (req < *off) {
        *kva = tmpkva;
        *mem_type = MX_MEM_HOSTMEM;
        /* Only the actual ring endpoint can map its state read+write */
        if (es->es_type == MYRI_ES_SNF_RX_RING && es->snf.es_rx.rx_ring_id == ring_req)
          *readonly = 0;
        else
          *readonly = 1;
        return 0;
      }
    }
  }

  /* mapping S.3 or S.2 PQ: snf desc addrs */
  if (is->snf.desc_cb.size != 0) {
    index = (req - *off) / PAGE_SIZE;
    *off += (unsigned long)is->snf.desc_cb.size;
    if (req < *off) {
      *mem_type = MX_MEM_HOSTMEM;
      if (is->snf.desc_cb.kva != NULL)
        *kva = (void *)((uintptr_t)is->snf.desc_cb.kva + PAGE_SIZE * index);
      else
	*kva = (void *)(uintptr_t)is->snf.desc_cb.pins[index].va;
      *pin = &is->snf.desc_cb.pins[index];
      if (es->es_type == MYRI_ES_SNF_RX_PQ) {
        *readonly = 0;
      }
      else {
        *readonly = 1;
      }
      return 0;
    }
  }

  /* mapping S.4 or S.3 PQ: snf data addrs */
  if (is->snf.data_cb.size != 0) {
    index = (req - *off) / PAGE_SIZE;
    *off += (unsigned long)is->snf.data_cb.size;
    if (req < *off) {
      *mem_type = MX_MEM_HOSTMEM;
      if (is->snf.data_cb.kva != NULL)
        *kva = (void *)((uintptr_t)is->snf.data_cb.kva + PAGE_SIZE * index);
      else
	*kva = (void *)(uintptr_t)is->snf.data_cb.pins[index].va;
      *pin = &is->snf.data_cb.pins[index];
      *readonly = 0;
      return 0;
    }
  }
  return ENOENT;
}

static int
myri_snf_start_stop(mx_endpt_state_t *es, int start)
{
  mx_instance_state_t *is = es->is;
  uint32_t dummy;
  int status;
  status = mx_mcp_command(is, MCP_CMD_SNF_RECV, start, 0, 0, &dummy);
  return status;
}

static
int
myri_snf_rx_detach(mx_endpt_state_t *es, int from_teardown)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  union snf_endpt_state *ess = &es->snf;
  mx_endpt_state_t *es_pg;
  int ring_id = ess->es_rx.rx_ring_id;
  unsigned long flags;
  int wake_pending;
  uint32_t ring_id_mask = (1u << ring_id);
  int wakeup_waiters = 0;

  flags = 0;  /* useless initialization to pacify -Wunused */

  if (es->es_type != MYRI_ES_SNF_RX_RING)
    return EINVAL;

  mal_assert(ring_id >= 0);

  /* HACK.  Remove the vq waiter if pending.  See snf_vqwait_req.
   * The caller already locked sync.
   */
  if (ess->es_rx.vq_waitsync.intr_requested) {
    /* MX_INFO(("snf_rx_detach: remove vq_waitsync\n")); */
    ess->es_rx.vq_waitsync.intr_requested = 0;
    TAILQ_REMOVE(&snf->vq_waitq, &ess->es_rx.vq_waitsync, link);
    if (ess->es_rx.vq_waitsync.wait_sync)
      mx_wake(ess->es_rx.vq_waitsync.wait_sync);
  }

  /* Fault tolerant pq lock.  If this endpoint is the owner, then release it. */
  mx_spin_lock(&snf->pq_lock);
  if (snf->pq_locked_es == es) {
    snf->pq_locked_es = NULL;
    wakeup_waiters = 1;
  }
  mx_spin_unlock(&snf->pq_lock);

  /* Then wake up all waiters in case any one is sleeping forever. */
  if (wakeup_waiters) {
    /* MX_INFO(("Snf ring closed without unlocking pq. Release it and wake up waiters.\n")); */
    snf_vqwake_req(snf, 1);
  }

  mx_spin_lock(&snf->sync_s.rs_lock);
  if (snf->sync_s.attach_mask & ring_id_mask && 
      es->snf.es_rx.rx_ring_id != SNF_RING_UNUSED) {
    /* Ensure future updates don't consider this ring but only if the user
     * cleanly requested to be detached.  If the detach comes as part of the
     * normal close path, the ring may or may not have been detached */
    if (!from_teardown)
      snf->sync_s.detach_mask |= ring_id_mask;
    snf->sync_s.attach_mask &= ~ring_id_mask;
    es->is->kernel_window->attach_mask = snf->sync_s.attach_mask;
    snf->num_rings_attached--;
    mx_spin_unlock(&snf->sync_s.rs_lock);

    /* DUP_MA state */
    if (es->snf.es_rx.app_ring_info) {
      struct snf_dup_ma_ring_state *rs = es->snf.es_rx.app_ring_info;

      snf->app_info[rs->app_info_idx].num_rings_attached--;
      snf->app_info[rs->app_info_idx].attach_mask &= ~(1 << rs->hash_ring_id);
      rs->app_info_idx = -1;
      rs->port_ring_id = 0;
      rs->hash_ring_id = 0;
      es->snf.es_rx.app_ring_info = NULL;
    }

    /* Prevent further enqueues */
    snf->refcount--;
    snf->ring_map_ep[ring_id] = NULL;

    mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);
    if (ess->es_rx.rx_waitsync.intr_requested) {
      TAILQ_REMOVE(&snf->rx_intr_waitq, &ess->es_rx.rx_waitsync, link);
      ess->es_rx.rx_waitsync.intr_requested = 0;
      if (TAILQ_EMPTY(&snf->rx_intr_waitq))
        snf->rx_intr_requested = 0;
    }
    wake_pending = ess->es_rx.rx_waitsync.intr_wake_pending;
    mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);

    /* If this thread was waiting on an interrupt along with others, it's
     * possible that it is in the process of being awakened.  If this is the
     * case, we give the kernel a chance to complete its wakeups */
    if (wake_pending) {
      wake_pending = 2;
      do {
        mx_sleep(&es->sync, MX_SMALL_WAIT, MX_SLEEP_NOINTR);
      } while (ess->es_rx.rx_waitsync.intr_wake_pending && --wake_pending >= 0);
      if (wake_pending == 0)
        MX_WARN(("%s: ring %d unable to clear wake_pending\n",
                  snf->is->is_name, ring_id));
    }
    es_pg = es->snf.es_rx.rx_pg;
    if (es_pg != NULL) {
      es->snf.es_rx.rx_pg = NULL;
      snf_rx_portgroup_del(es_pg, es);
    }
  }
  else
    mx_spin_unlock(&snf->sync_s.rs_lock);
  es->snf.es_rx.rx_ring_id = SNF_RING_UNUSED;
  return 0;
}

static void
snf_intr_tx_flush_work(
#if defined(NDIS50_MINIPORT) || defined(NDIS51_MINIPORT)
		       PNDIS_WORK_ITEM wi,
#endif
		       void *arg
#if defined(NDIS60_MINIPORT)
		       , NDIS_HANDLE wi
#endif
)
{
  mx_endpt_state_t *es = MX_WORK_STRUCT_ARG_CONTAINER_OF(arg,
                          mx_endpt_state_t, arch.snf.tx_flush_work);
  struct snf__tx_state *ss = es->snf.es_tx.tx_sp.tx_s;

#if defined(NDIS50_MINIPORT) || defined(NDIS51_MINIPORT)
  NdisFreeMemory(wi, sizeof (*wi), 0);
#elif defined(NDIS60_MINIPORT)
  NdisFreeIoWorkItem(wi);
#endif

  if (es->arch.snf.tx_flush_cnt == ss->tx_flush_cnt) {
    /* No new flushes, try again and if unsuccessful, reschedule work queue */
    if (!snf_intr_tx_try_flush(es)) {
#ifdef _WIN32 /* Any NDIS */
      InterlockedDecrement(&es->arch.snf.work_cnt);
#endif
      myri_schedule_delayed_work(&es->arch.snf.tx_flush_work, 1);
      return;
    }
  }
#ifdef _WIN32 /* Any NDIS */
  KeSetEvent(&es->arch.snf.no_work, 0, FALSE);
  InterlockedDecrement(&es->arch.snf.work_cnt);
#endif
}

static
void
snf_handle_tx_intr(mx_endpt_state_t *es)
{
  /* Try to flush any outstanding batched packets, If unsuccessful,
   * reschedule the flush */
  if (!snf_intr_tx_try_flush(es)) {
    struct snf__tx_state *ss = es->snf.es_tx.tx_sp.tx_s;

    es->arch.snf.tx_flush_cnt = ss->tx_flush_cnt;
    myri_schedule_delayed_work(&es->arch.snf.tx_flush_work, 0);
  }
}

static int
myri_snf_rx_get_kva_maps(const mx_endpt_state_t *es, myri_snf_rx_attach_t *x)
{
  int status;
  mx_get_copyblock_t desc;
  uint32_t offc;
  uint64_t off;
  const myri_snf_rx_state_t *snf = &es->is->snf;

  status = myri_endpt_get_common_copyblocks(es, &desc, &offc);
  if (status)
    return status;

  x->kernel_window_offset = desc.kernel_window_offset;
  x->kernel_window_size = desc.kernel_window_len;

  /* Sniffer-only mappings start at the end of the mcp-common endpoints */
  off = (uint64_t) offc;
  x->rx_doorbell_offset = off;          /* S.1 */
  x->rx_doorbell_size = MX_PAGE_SIZE;
  off += MX_PAGE_SIZE;
 
  x->vstate_all_offset = off;           /* S.2 */
  x->vstate_all_size = snf->rx_s_length;
  off += snf->rx_s_length;
  
  x->rx_desc_ring_offset = off;         /* S.3 */
  off += snf->desc_cb.size;

  x->rx_data_ring_offset = off;         /* S.4 */
  off += snf->data_cb.size;

  x->rx_recv_info_size = desc.send_compcnt_len; /* A.5 */
  x->rx_recv_info_offset = desc.send_compcnt_offset;
  x->rx_recv_info_pg_offset = MCP_SNF_RECVINFO_OFF;

  return 0;
}

static int
myri_snf_rx_pq_get_kva_maps(const mx_endpt_state_t *es, myri_snf_pq_attach_t *x)
{
  int status;
  mx_get_copyblock_t desc;
  uint32_t offc;
  uint64_t off;
  const myri_snf_rx_state_t *snf = &es->is->snf;

  status = myri_endpt_get_common_copyblocks(es, &desc, &offc);
  if (status)
    return status;

  /* Sniffer-only mappings start at the end of the mcp-common endpoints */
  off = (uint64_t) offc;                /* S.1 PQ */
  x->pq_state_offset = off;
  x->pq_state_size = snf->pq_s_length;
  off += snf->pq_s_length;
  
  x->rx_desc_ring_offset = off;         /* S.2 PQ */
  off += snf->desc_cb.size;

  x->rx_data_ring_offset = off;         /* S.3 PQ */
  off += snf->data_cb.size;

  return 0;
}

static
int
snf_rx_attach_endpt(myri_snf_rx_state_t *snf, mx_endpt_state_t *es, int ring_id, int app_id)
{
  union snf_endpt_state *ess = &es->snf;
  struct snf_sync_state *ss = &snf->sync_s;
  struct snf__rx_state *rx_s;
  uint32_t ring_id_mask;
  int rc = 0;
  int hash_ring_id;
  struct snf_dup_ma_ring_state *rs = NULL;

  mal_assert(es->es_type == MYRI_ES_SNF_RX_RING);

  mx_spin_lock(&snf->sync_s.rs_lock);

  /* DUP_MA mode.  ring_id is the user request (i.e. per-application id).
   * We always allocate per-port id regardless of ring_id >= 0.  So save it
   * for later use.  Ugly...
   */
  hash_ring_id = ring_id;
  if ((snf->params.open_flags & MYRI_SNF_F_RX_DUP_MA) && ring_id >= 0)
    ring_id = -1; /* force to find a free slot */

  if (ring_id < 0) { /* Look for next free ring id */
    for (ring_id = 0; ring_id < snf->num_rings_requested; ring_id++) {
      ring_id_mask = (1u << ring_id);
      if ((ss->attach_mask & ring_id_mask) == 0)
	break;
    }
    if (ring_id == snf->num_rings_requested) {
      rc = EBUSY;
      goto bail;
    }
  }
  /* FIXME.  What if ring_id >= 0 and that id is already taken? */

  /* DUP_MA mode.  Now do per-application ring id allocation. */
  if ((snf->params.open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    struct snf_dup_ma_app_state *as = NULL;
    int i, pid = mx_kgettgpid(); /* tgid, not pid */

    /* Find the application */
    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      if (snf->app_info[i].app_id == app_id) {
        as = &snf->app_info[i];
        break;
      }
    }
    /* This is a programming error on the user's part.  Should have opened
     * the port by now.
     */
    if (as == NULL) {
      MX_WARN(("SNF rx attach: cannot find the application/process."
        " app_id=%d pid=%d\n", app_id, pid));
      rc = EINVAL;
      goto bail;
    }
    /* Do we have free slots? */
    if (as->num_rings_attached >= as->num_rings_requested) {
      MX_WARN(("SNF rx attach: out of free rx rings. app_id=%d pid=%d rings_attached=%d"
        " rings_requested=%d\n",
        app_id, pid, as->num_rings_attached, as->num_rings_requested));
      rc = EBUSY;
      goto bail;
    }
    /* If requested a specific id, see if it is already taken. */
    if (hash_ring_id >= 0) {
      if (hash_ring_id >= as->num_rings_requested) {
        MX_WARN(("SNF rx attach: invalid ring id. app_id=%d pid=%d ring_id=%d"
          " rings_requested=%d\n",
          app_id, pid, hash_ring_id, as->num_rings_requested));
        rc = EINVAL;
        goto bail;
      }
      if ((as->attach_mask & (1 << hash_ring_id))) {
        MX_WARN(("SNF rx attach: the requested ring id is already taken."
          " app_id=%d pid=%d ring_id=%d rings_attached=%d rings_requested=%d\n",
          app_id, pid, hash_ring_id, as->num_rings_attached, as->num_rings_requested));
        rc = EBUSY;
        goto bail;
      }
    }
    else {
      for (hash_ring_id = 0; hash_ring_id < as->num_rings_requested;
           hash_ring_id++) {
        if ((as->attach_mask & (1 << hash_ring_id)) == 0)
          break;
      }
      if (hash_ring_id == as->num_rings_requested) {
        MX_WARN(("SNF rx attach: out of free rx rings. app_id=%d pid=%d rings_attached=%d"
          " rings_requested=%d\n",
          app_id, pid, as->num_rings_attached, as->num_rings_requested));
        rc = EBUSY;
        goto bail;
      }
    }

    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      if (snf->app_ring_info[i].app_info_idx == -1) {
        rs = &snf->app_ring_info[i];
        rs->app_info_idx = as - snf->app_info;
        rs->port_ring_id = ring_id;
        rs->hash_ring_id = hash_ring_id;
        as->num_rings_attached++;
        as->attach_mask |= (1 << hash_ring_id);
        break;
      }
    }
    if (rs == NULL) {
      /* This should not happen */
      MX_WARN(("SNF rx attach: out of free application ring slots."
        " app_id=%d pid=%d rings_attached=%d rings_requested=%d\n",
        app_id, pid, as->num_rings_attached, as->num_rings_requested));
      rc = EBUSY;
      goto bail;
    }
  }

  ring_id_mask = (1u << ring_id);
  rx_s = SNF__RX_STATE(snf->rx_s, ring_id);
  ss->attach_mask |= ring_id_mask;
  /* Provide a copy of the mask for userspace apps, for convenience */
  es->is->kernel_window->attach_mask = ss->attach_mask;
  snf->num_rings_attached++;
  snf->refcount++;
  snf->ring_map_ep[ring_id] = es;

  ess->es_rx.app_ring_info = rs;
  ess->es_rx.rx_ring_id = ring_id;
  ess->es_rx.rx_s = rx_s;
  ess->es_rx.rx_waitsync.wait_sync = &es->wait_sync;
  ess->es_rx.rx_waitsync.intr_requested = 0;
  ess->es_rx.rx_waitsync.snf = snf;

  ess->es_rx.vq_waitsync.wait_sync = &es->vq_sync;
  ess->es_rx.vq_waitsync.snf = snf;
  ess->es_rx.vq_waitsync.intr_requested = 0;

  sprintf(es->opener.user_info, "rx ring %d", ring_id);

  bzero(&rx_s->recvq.stats, sizeof(struct snf__recvq_stats));

  /* If the id_mask is non-zero, we're reattching to a ring that had valid live
   * state. The ring is no longer orphaned and we want to prevent the sync code
   * from reclaiming (and clearing) the state of a ring that is no longer
   * detached with valid live state. */
  if ((snf->sync_s.detach_mask & ring_id_mask) != 0) {
    snf->sync_s.detach_mask &= ~ring_id_mask;
    rx_s->recvq.rq_is_reattach = 1;
  }
  else {
    uint32_t descq_elems;
    struct snf__rx_sync *rs = &ss->rs;

    bzero(rx_s, sizeof(struct snf__rx_state));

    ss->r_rs[ring_id] = ss->rs; /* deep copy */

    descq_elems = snf->rx_desc_ring_count;

    rx_s->evq.evq_seqnum  = (rs->rs_desc_cnt % SNF_RX_DESC_SEQNUM_MAX) + 1;
    rx_s->evq.evq_next_idx = rs->rs_desc_cnt % descq_elems;
    rx_s->evq.evq_last_idx = descq_elems - 1;
    rx_s->evq.evq_evinfo_idx = -1;
    rx_s->evq.evq_desc_cnt = rs->rs_desc_cnt;
    rx_s->evq.evq_data_head = rs->rs_data_head;

    rx_s->recvq.rq_data_off = rs->rs_data_head % snf->params.rx_data_ring_size;
    rx_s->recvq.rq_data_head = rs->rs_data_head;
    rx_s->recvq.rq_data_tail = rs->rs_data_head;
    rx_s->recvq.rq_data_tail_last = rs->rs_data_head;
    rx_s->recvq.rq_data_size = snf->params.rx_data_ring_size;
    rx_s->recvq.rq_ev_ts_last = 1;
  }
  memset(&rx_s->recvq.stats, 0, sizeof(struct snf__recvq_stats));

bail:
  mx_spin_unlock(&snf->sync_s.rs_lock);
  return rc;
}

static
int
snf_pq_attach_endpt(myri_snf_rx_state_t *snf, mx_endpt_state_t *es)
{
  union snf_endpt_state *ess = &es->snf;
  struct snf_sync_state *ss = &snf->sync_s;
  struct snf__pq_state *pq_s;
  int rc = 0;

  mal_assert(es->es_type == MYRI_ES_SNF_RX_PQ);

  ess->es_rx.pq_s = snf->pq_s;
  ess->es_rx.rx_s = snf->rx_s;

  ess->es_rx.rx_waitsync.snf = snf;
  ess->es_rx.rx_waitsync.wait_sync = &es->wait_sync;
  ess->es_rx.rx_waitsync.intr_requested = 0;

  sprintf(es->opener.user_info, "pq processor");

  if (snf->pq_refcount > 1) {
    /* MX_INFO(("pq_attach_endpt. re-attach.\n")); */
    return 0;
  }
  
  mx_spin_lock(&snf->sync_s.rs_lock);
  
  pq_s = snf->pq_s;

  {
    uint32_t descq_elems;
    struct snf__rx_sync *rs = &ss->rs;

    bzero(pq_s, sizeof(struct snf__pq_state));

    descq_elems = snf->rx_desc_ring_count;

    pq_s->rw.evq_prehash_seqnum  = (rs->rs_desc_cnt % SNF_RX_DESC_SEQNUM_MAX) + 1;
    pq_s->rw.rq_prehash_data_off = rs->rs_data_head % snf->params.rx_data_ring_size;
    pq_s->rw.evq_next_prehash_idx = rs->rs_desc_cnt % descq_elems;
    pq_s->rw.evq_desc_cnt = rs->rs_desc_cnt;
    pq_s->rw.evq_ts_last = 1; /* FIXME.  This might be wrong? */
    pq_s->evq_last_prehash_idx = descq_elems - 1;

    /* FIXME.  We cannot get these from snf__rx_sync.
     * We should not re-initialize every time like this. */
  }

  mx_spin_unlock(&snf->sync_s.rs_lock);

  /* MX_INFO(("pq_attach_endpt\n")); */
  return rc;
}

static int
snf_rx_attach(mx_endpt_state_t *es, myri_snf_rx_attach_t *x)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  int ring_id = x->ring_id;
  int rc = 0;
  if (es->es_type != MYRI_ES_SNF_RX_RING)
    return EINVAL;

  mx_mutex_enter(&snf->sync);
  /* Get the easy ones out of the way first */
  if (ring_id >= (int) snf->num_rings_requested) {
    rc = EINVAL;
    goto bail;
  }
  if (snf->num_rings_attached == snf->num_rings_requested) {
    rc = EBUSY;
    goto bail;
  }

  rc = myri_snf_rx_get_kva_maps(es, x);
  if (rc) {
    MX_WARN(("Failed to obtain kva mappings: %d\n", rc));
    goto bail;
  }

  rc = snf_rx_attach_endpt(snf, es, ring_id, x->params.app_id);
  if (rc) {
    MX_WARN(("Failed to attach to ring %d with err=%d\n", ring_id, rc));
    goto bail;
  }

  x->params = snf->params; /* Copy back out to attaching user */
  x->ring_id = es->snf.es_rx.rx_ring_id;
  x->port_ring_id = x->ring_id; /* By default, hash ring id == port ring id */
  myri_clksync_copy_nticks_params(&es->is->kernel_window->clksync, &x->nticks);

  /* DUP_MA mode.  Override parameters with per-application ones */
  if ((snf->params.open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    struct snf_dup_ma_ring_state *rs = es->snf.es_rx.app_ring_info;
    struct snf_dup_ma_app_state *as = &snf->app_info[rs->app_info_idx];

    x->ring_id = rs->hash_ring_id; /* The application sees per-app ring id */
    x->params.num_rings = as->num_rings_requested;
    x->params.rss_flags = as->rss_flags;
    if (MAL_DEBUG) {
      MX_INFO(("rx_attach: copy per-app parameters. app_info_idx=%d app_id=%d num_rings=%d"
        " rss_flags=%d ring_id=%d port_ring_id=%d\n",
        rs->app_info_idx, as->app_id, x->params.num_rings, x->params.rss_flags,
        x->ring_id, x->port_ring_id));
    }
  }

bail:
  mx_mutex_exit(&snf->sync);
  return rc;
}

static int
snf_pq_attach(mx_endpt_state_t *es, myri_snf_pq_attach_t *x)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  int rc = 0;
  if (es->es_type != MYRI_ES_SNF_RX_PQ)
    return EINVAL;

  mx_mutex_enter(&snf->sync);

  rc = myri_snf_rx_pq_get_kva_maps(es, x);
  if (rc) {
    MX_WARN(("Failed to obtain kva mappings: %d\n", rc));
    goto bail;
  }

  rc = snf_pq_attach_endpt(snf, es);
  if (rc) {
    MX_WARN(("Failed to attach PQ, err=%d\n", rc));
    goto bail;
  }

  x->params = snf->params; /* Copy back out to attaching user */
  snf->pq_refcount++;
  snf->refcount++;

bail:
  mx_mutex_exit(&snf->sync);
  return rc;
}

/* HACK.  Clean up interrupt request.  Much of this code is from myri_snf_rx_detach.
 * This function stays until we remove the concept of PQ endpoints.
 */
static
void
snf_pq_detach(mx_endpt_state_t *es)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  union snf_endpt_state *ess = &es->snf;
  unsigned long flags;
  int wake_pending;

  flags = 0;  /* useless initialization to pacify -Wunused */

  /* MX_INFO(("snf_pq_detach\n")); */
  if (es->es_type != MYRI_ES_SNF_RX_PQ) {
    MX_WARN(("PQ detach called with non RX_PQ endpoint\n"));
    return;
  }
  /* MX_INFO(("intr_requested=%d ring_id=%d\n", ess->es_rx.rx_waitsync.intr_requested, ess->es_rx.rx_ring_id)); */

  mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);
  if (ess->es_rx.rx_waitsync.intr_requested) {
    TAILQ_REMOVE(&snf->rx_intr_waitq, &ess->es_rx.rx_waitsync, link);
    ess->es_rx.rx_waitsync.intr_requested = 0;
    if (TAILQ_EMPTY(&snf->rx_intr_waitq))
      snf->rx_intr_requested = 0;
  }
  wake_pending = ess->es_rx.rx_waitsync.intr_wake_pending;
  mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);

  /* If this thread was waiting on an interrupt along with others, it's
   * possible that it is in the process of being awakened.  If this is the
   * case, we give the kernel a chance to complete its wakeups */
  if (wake_pending) {
    wake_pending = 2;
    do {
      mx_sleep(&es->sync, MX_SMALL_WAIT, MX_SLEEP_NOINTR);
    } while (ess->es_rx.rx_waitsync.intr_wake_pending && --wake_pending >= 0);
    if (wake_pending == 0)
      MX_WARN(("%s: pq unable to clear wake_pending\n",
                snf->is->is_name));
  }

  snf->pq_refcount--;
  snf->refcount--;
}

static
void
snf_rx_teardown(mx_endpt_state_t *es)
{
  myri_snf_rx_state_t *snf = &es->is->snf;
  mx_instance_state_t *is = es->is;
  uint32_t dummy;
  int status;

  mx_mutex_enter(&snf->sync);
  if (es->es_type == MYRI_ES_SNF_RX_RING)
    myri_snf_rx_detach(es, 1);
  if (es->es_type == MYRI_ES_SNF_RX_PQ) {
    mal_always_assert(snf->pq_refcount > 0);
    snf_pq_detach(es);
    if (MAL_DEBUG) {
      MX_INFO(("rx_teardown. pq. seq=%d ev_idx=%d\n",
             snf->pq_s->rw.evq_prehash_seqnum, snf->pq_s->rw.evq_next_prehash_idx));
    }
  } else if (es->es_type == MYRI_ES_SNF_RX && es->snf.es_rx_h.rx_is_attached) {
    struct snf_dup_ma_app_state *app;

    es->snf.es_rx_h.rx_is_attached = 0;
    snf->refcount--;
    /* DUP_MA mode.  If all handles to this app are closed, clear the app
     * information too.
     */
    if ((app = es->snf.es_rx_h.app_info)) {
      es->snf.es_rx_h.app_info = NULL;
      app->refcount--;
      if (app->refcount == 0)
        app->app_id = -1;
    }
  }

  if (snf->refcount == 0) {
    int i;

    snf->num_rings_attached = 0;
    snf->num_rings_requested = 0;
    snf->sync_s.detach_mask = 0;
    snf->sync_s.attach_mask = 0;
    snf->owner_pid = -1;
    if (snf->state == SNF_RX_S_STARTED)
      myri_snf_start_stop(es, 0);
    if (snf->state != SNF_RX_S_FREE) {
      mx_spin_lock_destroy(&snf->sync_s.rs_lock);
      snf->state = SNF_RX_S_FREE;
      status = mx_mcp_command(is, MCP_CMD_SNF_RING_FREE, 0, 0, 0, &dummy);
      if (status)
        MX_WARN(("could not unmap snf mcp ring: %d\n", status));
    }
    if (snf->rx_s)
      mx_kfree(snf->rx_s);
    if (snf->pq_s)
      mx_kfree(snf->pq_s);
    snf->pq_s = NULL;
    snf->rx_s = NULL;
    snf->rx_s_length = 0;
    mx_free_huge_copyblock(is, &snf->data_cb);
    snf->data_cb.size = 0;
    mx_free_huge_copyblock(is, &snf->desc_cb);

    /* Reset DUP_MA state as well */
    memset(snf->app_info, 0, sizeof(snf->app_info));
    memset(snf->app_ring_info, 0, sizeof(snf->app_ring_info));
    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      snf->app_info[i].app_id = -1;
      snf->app_info[i].refcount = 0;
      snf->app_ring_info[i].app_info_idx = -1;
    }
  }
  mx_mutex_exit(&snf->sync);
}

static int
myri_snf_kmap_copyblocks(mx_endpt_state_t *es)
{
  if (!es->is_kernel)
    return 0;

#if MAL_OS_LINUX || MAL_OS_FREEBSD
  /* For kernel-based endpoints, make sure that the copyblocks are virtually
   * mapped */
  if (es->is_kernel) {
    myri_snf_rx_state_t *snf = &es->is->snf;
    if (snf->desc_cb.kva == NULL) {
      if (!mx_map_huge_copyblock(&snf->desc_cb)) {
        MX_WARN(("Failed to kmap snf desc copyblock\n"));
        return ENXIO;
      }
    }
    if (snf->data_cb.kva == NULL) {
      if (!mx_map_huge_copyblock(&snf->data_cb)) {
        MX_WARN(("Failed to kmap snf data copyblock, es=%p\n", es));
        return ENXIO;
      }
    }
  }
#endif
  return 0;
}

static int
myri_snf_rx_open_get_params(mx_endpt_state_t *es, myri_snf_rx_params_t *x)
{
  mx_instance_state_t *is = es->is;
  myri_snf_rx_state_t *snf = &is->snf;
  unsigned long meta_size;

  if (snf->state != SNF_RX_S_FREE) {
    if (!es->is_kernel && mx_kgetpid() != snf->owner_pid && 
        !(snf->params.open_flags & 0x1)) {
      MX_WARN(("SNF can't share rings between processes unless SNF_F_PSHARED "
               "mode is used (owner is PID %d, request from PID %d)\n", 
               snf->owner_pid, mx_kgetpid()));
      return EACCES;
    }
    /* Deny mixed aggregation and non-aggregation modes.  Also check
     * the portmask so we the port does not belong to two different
     * aggregation/port groups.
     * 0x2 is SNF_F_AGGREGATE_PORTMASK in snf.h.  FIXME
     */
    if (((snf->params.open_flags ^ x->open_flags) & 0x2)) {
      MX_WARN(("SNF cannot open the port in both aggregation and"
          " non-aggregation mode at the same time.\n"));
      return EINVAL;
    }
    if ((snf->params.open_flags & 0x2) &&
      snf->params.aggr_portmask != x->aggr_portmask) {
      MX_WARN(("SNF port cannot belong to different port aggregation"
          " groups.\n"));
      return EINVAL;
    }
    *x = snf->params;
    return 0;
  }

  if (myri_snf_rings != 0)
    x->num_rings = myri_snf_rings;

  if (myri_snf_flags != -1)
    x->open_flags = myri_snf_flags;

  /* If there's only one ring, ignore all dup/copy requests */
  if (x->num_rings == 1)
    x->open_flags &= ~(MYRI_SNF_F_RX_COPY|MYRI_SNF_F_RX_DUP);

  x->rx_map_flags = 0;

  if (x->num_rings > 1)
    x->rx_map_flags |= MYRI_SNF_RXMAP_SHARED;

  /* DUP_MA mode.  Implicitly carries SHARED mode so multiple processes
   * can share the port.
   */
  if ((x->open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    x->open_flags |= 0x1;
    x->rx_map_flags |= MYRI_SNF_RXMAP_SHARED;
  }

  /* Validate that sizes are a multiple of 2 MB */
  meta_size = (unsigned long)
    (MX_VPAGE_SIZE * MX_VPAGE_SIZE / sizeof(mcp_dma_addr_t));

  if ((unsigned long)x->rx_data_ring_size & (meta_size - 1) ||
      (unsigned long)x->rx_desc_ring_size & (meta_size - 1)) {
    MX_WARN(("Desc or Data ring size not a multiple of 2MB\n"));
    return EINVAL;
  }
  if (x->num_rings > MYRI_SNF_MAX_RINGS) {
    MX_WARN(("Unsupported number of rings: %d\n", x->num_rings));
    return EINVAL;
  }

  x->rx_data_preamble_size = SNF_ALIGNUP(SNF_MTUWRAP_SIZE, PAGE_SIZE);

  return 0;
}

/* DUP_MA mode.  When the application/process first opens the port handle, we
 * call this function to save its app_id, number of rings and such.
 */
static int
myri_snf_app_join(mx_endpt_state_t *es, myri_snf_rx_params_t *x,
  myri_snf_rx_params_t *x_orig, struct snf_dup_ma_app_state **app)
{
  mx_instance_state_t *is = es->is;
  myri_snf_rx_state_t *snf = &is->snf;
  int status = 0;

  *app = NULL;
  if (MAL_DEBUG) {
    MX_INFO(("SNF DUP_MA mode application/process joins\n"));
  }

  /* All processes must also be using DUP_MA mode. */
  if (!(x->open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    MX_WARN(("SNF port is in DUP_MA mode. The joining process must also use"
        " DUP_MA mode.\n"));
    status = EINVAL;
  }
  else {
    /* Remember this process */
    struct snf_dup_ma_app_state *s = NULL;
    int app_id, i;

    app_id = x_orig->app_id;
    if (MAL_DEBUG) {
      int pid = mx_kgettgpid();
      MX_INFO(("DUP_MA process joining the port. app_id=%d pid=%d\n", app_id, pid));
    }
    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      if (snf->app_info[i].app_id == app_id) {
        /* The process has opened the port before.  It should not open
         * again, but is harmless.
         */
        s = &snf->app_info[i];
        *app = s;

        /* Pass the application's original parameters */
        x->num_rings = s->num_rings_requested;
        x->rss_flags = s->rss_flags;
        goto exit;
      }
    }

    /* A new application/process.  Save its params. */
    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      if (snf->app_info[i].app_id == -1) {
        /* A free slot. Only need to save num_rings and rss_flags as these may
         * be different from the port's.  Each application may have its own
         * num_rings and rss_flags too.  All the other parameters are inherited
         * from the port's.
         */
        s = &snf->app_info[i];
        s->app_id = app_id;
        s->num_rings_requested = x_orig->num_rings; /* Not the port's param */
        s->rss_flags = x_orig->rss_flags; /* Not the port's param */
        s->num_rings_attached = 0;
        s->attach_mask = 0;
        *app = s;
        break;
      }
    }
    if (s == NULL) {
      int pid = mx_kgettgpid();

      /* MCP_ENDPOINTS_CNT is conviniently 32 (common/mcp_config.h).  So this
       * should not happen.
       */
      status = EBUSY;
      MX_WARN(("SNF app join: out of free application slots. app_id=%d pid=%d\n", app_id, pid));
    }
  }

exit:
  return status;
}

static int
myri_snf_rx_join(mx_endpt_state_t *es)
{
  int status = 0;

  if (es->is_kernel && ((status = myri_snf_kmap_copyblocks(es)) != 0))
    return status;

  return status;
}

static int
myri_snf_rx_enable(mx_endpt_state_t *es, const myri_snf_rx_params_t *x)
{
  mx_instance_state_t *is = es->is;
  myri_snf_rx_state_t *snf = &is->snf;
  uint32_t offset;
  unsigned long flags;
  int status = 0;
  unsigned long meta_size;
  int node_id;
  uint32_t dummy;
  struct snf__flow_state flow_h;
  int rc;

  flags = 0;  /* useless initialization to pacify -Wunused */

  if (x->num_rings > MYRI_SNF_MAX_RINGS) {
    MX_WARN(("Exceeded maximum number of %d rings\n", MYRI_SNF_MAX_RINGS));
    return ENXIO;
  }

  snf->data_cb.size = 0;
  TAILQ_INIT(&snf->rx_intr_waitq);
  TAILQ_INIT(&snf->vq_waitq);
  snf->rx_intr_requested = 0;

  meta_size = (unsigned long)
    (MX_VPAGE_SIZE * MX_VPAGE_SIZE / sizeof(mcp_dma_addr_t));
  /* allocate rings in NIC */
  status = mx_mcp_command(is, MCP_CMD_SNF_RING_ALLOC, 0,
	        (uint32_t)(x->rx_data_ring_size / (meta_size / sizeof(mcp_dma_addr_t))), 
		(uint32_t)(x->rx_desc_ring_size / (meta_size / sizeof(mcp_dma_addr_t))), 
		&offset);
  if (status) {
    MX_WARN(("could not get snf ring offset\n"));
    return status;
  }

  /* In the preamble, we stuff enough for data premables */
  snf->data_cb.size = x->rx_data_ring_size + x->rx_data_preamble_size;

  node_id = myri_snf_find_node_id(es);

  /* allocate the huge copyblocks */
  status = mx_alloc_huge_copyblock(is, &snf->data_cb, 
                         (unsigned long)x->rx_data_preamble_size, node_id);
  if (status) {
    MX_WARN(("Failed to allocate snf data copyblock\n"));
    rc = ENOMEM;
    goto err_with_mcp_ring_mapped;
  }
  status = mx_dma_map_copyblock(es, &snf->data_cb.cb, offset);
  if (status) {
    MX_WARN(("Failed to map snf data copyblock\n"));
    rc = ENXIO;
    goto err_with_mcp_ring_mapped;
  }

  snf->desc_cb.size = x->rx_desc_ring_size;
  status = mx_alloc_huge_copyblock(is, &snf->desc_cb, 0, node_id);
  if (status) {
    MX_WARN(("Failed to allocate snf desc copyblock\n"));
    rc = ENOMEM;
    goto err_with_mcp_ring_mapped;
  }
  offset += (uint32_t)(x->rx_data_ring_size >> (2*(MX_VPAGE_SHIFT-3)));
  status = mx_dma_map_copyblock(es, &snf->desc_cb.cb, offset);
  if (status) {
    MX_WARN(("Failed to map snf desc copyblock\n"));
    rc = ENXIO;
    goto err_with_mcp_ring_mapped;
  }

  snf->rx_s_length = 
    SNF_ALIGNUP(MYRI_SNF_MAX_RINGS * SNF__RX_STATE_SIZE, PAGE_SIZE);
  snf->rx_s = mx_kmalloc(snf->rx_s_length, MX_WAITOK|MX_MZERO);
  if (snf->rx_s == NULL) {
    MX_WARN(("Failed to alloc rx state page\n"));
    rc = ENOMEM;
    goto err_with_mcp_ring_mapped;
  }

  snf->pq_s_length = SNF_ALIGNUP(sizeof(struct snf__pq_state), PAGE_SIZE*2);
  snf->pq_s = mx_kmalloc(snf->pq_s_length, MX_WAITOK|MX_MZERO);
  if (snf->rx_s == NULL) {
    MX_WARN(("Failed to alloc pq state page\n"));
    rc = ENOMEM;
    goto err_with_mcp_ring_mapped;
  }

  if (es->is_kernel && ((status = myri_snf_kmap_copyblocks(es)) != 0)) {
    rc = status;
    goto err_with_mcp_ring_mapped;
  }

  status = mx_mcp_command(is, MCP_CMD_SNF_RING_INIT, 0, 0, 0, &dummy);
  if (status) {
    MX_WARN (("Failed to initialize snf ring with mcp\n"));
    rc = status;
    goto err_with_mcp_ring_mapped;
  }

  /* Used to implement the fault tolerant pq lock */
  mx_spin_lock_init(&snf->pq_lock, snf->is, 0, "snf pq lock");
  snf->pq_locked_es = NULL;

  memset(&snf->sync_s, 0, sizeof(struct snf_sync_state));
  mx_spin_lock_init(&snf->sync_s.rs_lock, snf->is, 0, "snf flow lock");
  flow_h.flow_desc = (uint32_t)SNF_VPAGE_IDX(x->rx_desc_ring_size);
  flow_h.flow_data = (uint32_t)SNF_VPAGE_IDX(x->rx_data_ring_size);
  snf__rx_flow_update((uint64_t *) snf->rx_doorbell_be64, &flow_h);

  snf->sync_s.detach_mask = 0;
  snf->owner_pid = mx_kgettgpid();
  snf->params = *x; /* Save params, these are the default now */
  snf->num_rings_requested = snf->params.num_rings;
  snf->rx_desc_ring_count = (uint32_t)(snf->params.rx_desc_ring_size / sizeof(mcp_snf_rx_desc_t));
  snf->num_rings_attached = 0;
  snf->state = SNF_RX_S_INIT;

  /* Initialize DUP_MA state */
  if ((x->open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    int i;

    if (MAL_DEBUG) {
      MX_INFO(("Enabling snf port using MYRI_SNF_F_RX_DUP_MA. pid=%d app_id=%d rings=%d\n",
        snf->owner_pid, x->app_id, x->num_rings));
    }

    /* DUP_MA mode.  The very first open of the port sets the flag.
     * We do not know how many applications/processes are going to attach
     * to the port.  So, override and set the number of rings to max.
     */
    snf->num_rings_requested = MYRI_SNF_MAX_RINGS;

    memset(snf->app_info, 0, sizeof(snf->app_info));
    memset(snf->app_ring_info, 0, sizeof(snf->app_ring_info));
    for (i = 0; i < MYRI_SNF_MAX_RINGS; i++) {
      snf->app_info[i].app_id = -1;
      snf->app_ring_info[i].app_info_idx = -1;
    }
  }

  return 0;

err_with_mcp_ring_mapped:
  status = mx_mcp_command(is, MCP_CMD_SNF_RING_FREE, 0, 0, 0, &dummy);
  if (status)
    MX_WARN(("could not unmap snf mcp ring: %d\n", status));
  return rc;
}

static int
snf_rx_sync(myri_snf_rx_state_t *snf, int ring_id, struct snf__rx_sync *rs_update)
{
  struct snf_sync_state *ss = &snf->sync_s;
  int idx_n = -1, i;
  int status = 0;
  uint64_t da = 0;
  const uint64_t da_min_update = (16ull << SNF_VPAGE_SHIFT);
  const uint64_t da_max = snf->params.rx_data_ring_size;
  const uint64_t de_max = snf->rx_desc_ring_count;

  mx_spin_lock(&ss->rs_lock);

  /*
   * The function can be used to force a synchronization (rs_update == NULL)
   * but if not, make sure the ring is valid before we go on using its index.
   *
   * As means of precaution, we also validate the input data from rs_update.
   *
   * We set 'min_update' as the minimum amount of bytes we want to increment
   * before looking through tail pointers and eventually updating the true
   * desc/data heads.
   */
  if (rs_update) {
    uint64_t da = rs_update->rs_data_tail - ss->rs.rs_data_tail;
    if (ring_id < 0 || ring_id >= MYRI_SNF_MAX_RINGS || 
	((ss->attach_mask & (1u<<ring_id)) == 0)) {
      status = EINVAL;
      goto done;
    }
    if (da > da_max || (rs_update->rs_desc_cnt - ss->rs.rs_desc_cnt > de_max)) {
      status = EINVAL;
      goto done;
    }
    ss->r_rs[ring_id] = *rs_update; /* deep copy */
    if (da < da_min_update)
      goto done;
  }

  /*
   * Find the next smallest increment that moves the data tail
   */
  idx_n = -1;
  {
    uint64_t da_min = da_max;

    SNF_FFS_FOREACH(ss->attach_mask, i) {
      da = ss->r_rs[i].rs_data_tail - ss->rs.rs_data_tail;
      if (da > da_max) {
	idx_n = -1;
	break;
      }
      else if (da <= da_min) {
	idx_n = i;
	da_min = da;
      }
    }
  }

  /*
   * If a valid increment is found and it warrants at least "da_min_update" bytes,
   * update the new tail pointer.
   */
  if (idx_n >= 0 && da >= da_min_update) {
    struct snf__flow_state flow_h;
    uint64_t de, de_min = de_max;

    ss->rs.rs_data_tail = ss->r_rs[idx_n].rs_data_tail;

    /* 
     * For the data ring, recently detached rings may refer to data that is now
     * stale.  If that is the case, make sure they are not soon reopened to
     * point to data entries that will soon be released back to the NIC. */ 
    if (ss->detach_mask) {
      SNF_FFS_FOREACH(ss->detach_mask, i) {
	da = ss->r_rs[i].rs_data_tail - ss->rs.rs_data_tail;
	if (da > da_max)
	  ss->detach_mask &= ~(1u<<i);
      }
    }

    /*
     * Since we'll be updating the data ring, go ahead and update the latest
     * descriptor ring value while we're there 
     */
    idx_n = -1;
    de_min = de_max;
    SNF_FFS_FOREACH(ss->attach_mask, i) {
      de = ss->r_rs[i].rs_desc_cnt - ss->rs.rs_desc_cnt;
      if (de > de_max) {
	idx_n = -1;
	break;
      }
      else if (de <= de_min) {
	idx_n = i;
	de_min = de;
      }
    }
    
    if (idx_n >= 0) {
      /* 
       * The descriptor ring can now be advanced and we also save where the
       * data header ring happened to be for that descriptor count.  This is
       * useful when a new ring needs to join and consume from the same ring 
       */
      ss->rs.rs_desc_cnt = ss->r_rs[idx_n].rs_desc_cnt;
      ss->rs.rs_data_head = ss->r_rs[idx_n].rs_data_head;

      /* Ditto for descriptors and detach_mask (see above) */
      if (ss->detach_mask) {
	SNF_FFS_FOREACH(ss->detach_mask, i) {
	  de = ss->r_rs[i].rs_desc_cnt - ss->rs.rs_desc_cnt;
	  if (de > de_max) {
	    ss->detach_mask &= ~(1u<<i);
	  }
	}
      }
    }

    /*
     * At least data flow warrants an update and we take whatever desc cnt we
     * happened to figure out above as a result of advancing at least
     * da_min_update bytes in the data queue.
     */
    flow_h.flow_desc = SNF_DESC_VPAGES(ss->rs.rs_desc_cnt) + 
	(uint32_t)SNF_VPAGE_IDX(snf->params.rx_desc_ring_size);
    flow_h.flow_data = SNF_DATA_VPAGES(ss->rs.rs_data_tail) + 
	(uint32_t)SNF_VPAGE_IDX(snf->params.rx_data_ring_size);
    snf__rx_flow_update(snf->rx_doorbell_be64, &flow_h);
  }

done:
  mx_spin_unlock(&ss->rs_lock);
  return status;
}

static int
snf_rx_open_handle(mx_endpt_state_t *es, myri_snf_rx_params_t *x)
{
  mx_instance_state_t *is = es->is;
  myri_snf_rx_state_t *snf = &is->snf;
  int status = 0;
  myri_snf_rx_params_t x_orig;

  if (es->es_type != MYRI_ES_SNF_RX)
    return EINVAL;
  x_orig = *x; /* Save the original parameters for DUP_MA mode. */

  mx_mutex_enter(&snf->sync);
  /* First do all parameter checking where failures won't consume any SNF
   * resources */
  if ((status = myri_snf_rx_open_get_params(es, x))) {
    mx_mutex_exit(&snf->sync);
    return status;
  }

  if (snf->state == SNF_RX_S_FREE) { /* First Sniffer opener */
    if ((status = myri_snf_rx_enable(es, x)))
      goto fail;
  } else {
    if ((status = myri_snf_rx_join(es)))
      goto fail;
  }

  /* DUP_MA mode.  Save the application/process info. */
  if ((snf->params.open_flags & MYRI_SNF_F_RX_DUP_MA)) {
    struct snf_dup_ma_app_state *app;

    if ((status = myri_snf_app_join(es, x, &x_orig, &app)))
      goto fail;
    es->snf.es_rx_h.app_info = app;
    app->refcount++;

    /* A one process may open the handle multiple times.  On the second open
     * and later, it passes blank parameters and expects to retrieve
     * parameters via the kernel.  Probably an overkill.  This usage is
     * highly unlikely.  The user must pass in the correct app_id in any case.
     */
    x->num_rings = app->num_rings_requested;
    x->app_id = x_orig.app_id; /* open_get_params clobbers this... */
    sprintf(es->opener.user_info, "app_id=%d rx handle (%d rings)",
        x->app_id, x->num_rings);
  }
  else {
    sprintf(es->opener.user_info, "rx handle (%d %s rings)", 
        snf->num_rings_requested,
        (snf->params.open_flags & 0x1) ? "shared" : "private");
  }
  snf->refcount++;
  es->snf.es_rx_h.rx_is_attached = 1;

 fail:
  mx_mutex_exit(&snf->sync);
  if (status)
    snf_rx_teardown(es);
  return status;
}

static
int
snf_rx_portgroup_add(mx_endpt_state_t *es_pg, mx_endpt_state_t *es,
  myri_snf_port_group_t *param)
{
  struct snf_endpt_portgroup *ep_pg = &es_pg->snf.es_pg;
  int status;
  int port_id;
  struct myri_snf_waiter *waiter;

  mx_mutex_enter(&es_pg->sync);
  port_id = es->is->id;
  if (es_pg->es_type == MYRI_ES_SNF_RX_GROUP) {
    /* FIXME  We use pg's vq wait to wait on the port...
     * This is probably not the right place to initialize waitsync.
     */
    if (ep_pg->vq_waitsync.wait_sync == NULL) {
      ep_pg->vq_waitsync.wait_sync = &es_pg->vq_sync;
      ep_pg->vq_waitsync.snf = &es_pg->is->snf;
      ep_pg->vq_waitsync.intr_requested = 0;
    }

    waiter = &ep_pg->pg_waiters[port_id];
    if (waiter->wait_sync != NULL) {
      status = EBUSY;
    }
    else {
      waiter->wait_sync = &es_pg->wait_sync;
      waiter->intr_requested = 0;
      waiter->snf = &es->is->snf;
      TAILQ_INSERT_TAIL(&ep_pg->pg_waitq, waiter, pg_link);
      es->snf.es_rx.rx_pg = es_pg;
      ep_pg->evq_base[port_id] = param->evq_base;
      status = 0;
    }
  }
  else
    status = EINVAL;
  mx_mutex_exit(&es_pg->sync);

  return status;
}

/* 
 * For removal, if es_rx_ring == NULL, remove all refs.
 */
static
int
snf_rx_portgroup_del(mx_endpt_state_t *es, mx_endpt_state_t *es_rx_ring)
{
  struct snf_endpt_portgroup *ep_pg = &es->snf.es_pg;
  struct myri_snf_waiter *waiter;
  myri_snf_rx_state_t *snf;
  int i;
  unsigned long flags;
  int wake_pending;

  flags = 0;  /* useless initialization to pacify -Wunused */

  if (es->es_type != MYRI_ES_SNF_RX_GROUP)
    return EINVAL;

  if (es_rx_ring != NULL) {
    i = es_rx_ring->is->id;
    waiter = &ep_pg->pg_waiters[i];
    if (waiter->wait_sync == NULL)
      return EINVAL;
  }
  else {
    waiter = TAILQ_FIRST(&ep_pg->pg_waitq);
  }

  mx_mutex_enter(&es->sync);

  while (waiter != NULL) {
    if (waiter->wait_sync != NULL) {
      snf = waiter->snf;
      ep_pg->port_mask &= ~(1 << snf->is->id);
      mx_spin_lock_irqsave(&snf->rx_intr_lock, flags);
      if (waiter->intr_requested) {
        TAILQ_REMOVE(&snf->rx_intr_waitq, waiter, link);
        waiter->intr_requested = 0;
      }
      wake_pending = waiter->intr_wake_pending;
      mx_spin_unlock_irqrestore(&snf->rx_intr_lock, flags);
      if (wake_pending) {
        wake_pending = 2;
        do {
          mx_sleep(&es->sync, MX_SMALL_WAIT, MX_SLEEP_NOINTR);
        } while (waiter->intr_wake_pending && --wake_pending >= 0);
        if (wake_pending == 0)
          MX_WARN(("%s: unable to clear wake_pending in portgroup\n",
                  snf->is->is_name));
      }
      waiter->wait_sync = NULL;
    }
    if (es_rx_ring != NULL)
      break;
    else
      waiter = TAILQ_NEXT(waiter, pg_link);
  }

  /* Wake up vq waiters on the group.  FIXME */
  if (es_rx_ring == NULL) {
    union snf_endpt_state *ess = &es->snf;
    if (ess->es_pg.vq_waitsync.intr_requested) {
      ess->es_pg.vq_waitsync.intr_requested = 0;
      snf = &es->is->snf;
      TAILQ_REMOVE(&snf->vq_waitq, &ess->es_pg.vq_waitsync, link);
      if (ess->es_pg.vq_waitsync.wait_sync)
        mx_wake(ess->es_pg.vq_waitsync.wait_sync);
    }
    ep_pg->port_mask = 0;
  }

  mx_mutex_exit(&es->sync);
  return 0;
}

static
int
snf_rx_portgroup_join(mx_endpt_state_t *es_pg, myri_snf_port_group_t *param)
{
  mx_instance_state_t *is;
  mx_endpt_state_t *es_rx_ring;
  int status;
  uint32_t port_id, ring_id;

  if (es_pg->es_type != MYRI_ES_SNF_RX_GROUP)
    return EINVAL;

  port_id = param->port_id;
  ring_id = param->ring_id;
  if (ring_id >= MYRI_SNF_MAX_RINGS || port_id >= SNF_PORTGROUP_MAX_MEMBERS)
    return EINVAL;

  if (es_pg->is->id != port_id) {
    is = mx_get_instance(port_id);
    if (is == 0)
      return ENODEV;
  }
  else
    is = es_pg->is;

  es_rx_ring = is->snf.ring_map_ep[ring_id];
  if (es_rx_ring == NULL) {
    if (es_pg->is->id != port_id)
      mx_release_instance(is);
    return ENODEV;
  }
  (void) mx_get_endpoint(is, es_rx_ring->endpt);

#ifdef _LINUX_UIDGID_H
  if (!uid_eq(es_rx_ring->euid,es_pg->euid)) status = EPERM;
#else
  if (es_rx_ring->euid != es_pg->euid) status = EPERM;
#endif
  else if (es_rx_ring->snf.es_rx.rx_pg != NULL)
    status = EBUSY; /* already in a port group */
  else
    status = snf_rx_portgroup_add(es_pg, es_rx_ring, param);

  mx_put_endpoint(es_rx_ring);
  if (es_pg->is->id != port_id)
    mx_release_instance(is);

  return status;
}

static
int
snf_eop_ioctl(mx_endpt_state_t *es, uint32_t cmd, const uaddr_t in)
{
  uaddr_t out = in;
  int status = ENOTTY;
  myri_snf_rx_state_t *snf = &es->is->snf;

  if (mx_is_dead(es->is)) {
    /* if the board is dead, return an error to the application...
     * except in case of parity since the recovery is notified to the
     * application through the wait_status or get_board_status
     * (other ioctls can still be processed safely since they do not
     * touch the board for anything that won't be redone a afterwards)
     */
    if (!(es->is->flags & MX_PARITY_RECOVERY)) {
      MX_WARN(("firmware dead on board %d, ignoring ioctl\n", es->is->id));
      return EIO;
    }
  }

  /* Lock the pg's port, instead of all ports in the pg. */
#define LOCK_PG 1

  if (cmd == MYRI_SNF_PQ_GROUP_LOCK) {
    myri_snf_pq_group_lock_t param;
    struct myri_snf_waiter *waiter;
    myri_snf_rx_state_t *pg_snf;
    struct snf_endpt_portgroup *ep_pg;
    int locked, wait;
    unsigned long flags;

    flags = 0; /* -Wunused on non-linux */

    if ((status = mx_copyin(in, &param, sizeof(param), es->is_kernel)))
      return status;

    /* Endpoint is the ring group. */
    ep_pg = &es->snf.es_pg;

    /* The first port stores group wide variables... */
    pg_snf = snf;

  pq_group_lock_tryagain:
    locked = 0; /* 1 = locked a port, 2 = locked by another */
    wait = 0; /* 1 = wait on pq, 2 = wait on vq */

#if LOCK_PG
    mx_spin_lock_irqsave(&pg_snf->pq_lock, flags);
#else
    /* Acquire all spin locks */
    TAILQ_FOREACH(waiter, &ep_pg->pg_waitq, pg_link) {
      snf = waiter->snf;
      mx_spin_lock(&snf->pq_lock);
    }
#endif

    /* Check each port.  FIXME.  Make a faster scan. */
    TAILQ_FOREACH(waiter, &ep_pg->pg_waitq, pg_link) {
      int portnum;

      snf = waiter->snf;
      portnum = snf->is->id;
      /* Skip if we are not supposed to check this port */
      if ((param.u.portmask & (1 << portnum)) == 0)
        continue;

      if (snf->pq_locked_es == NULL) {
        /* Now check if the ring is non-empty */
        mcp_snf_rx_desc_t  *evq_base;
        mcp_snf_rx_desc_t *event;
        uint32_t seqnum;

        evq_base = ep_pg->evq_base[portnum];
        event = evq_base + snf->pq_s->rw.evq_next_prehash_idx;
        seqnum = event->seqnum;
        if (seqnum == snf->pq_s->rw.evq_prehash_seqnum) {
          /* Non-empty.  Lock the port and we are done. */
          snf->pq_locked_es = es;
          param.u.portnum = portnum;
          /* Remember the port number so we can unlock it later */
          ep_pg->pq_locked_port = snf;
          locked = 1;
          break;
        }
        else {
          /* Empty.  Continue to the next port. */
        }
      }
      else {
        /* Locked by another thread.  Skip */
        locked = 2;
      }
    }

    /* See if we want to wait on pq or vq */
    if (locked != 1 && param.timeout_ms != 0) {
      if (locked == 2 || pg_snf->pq_group_pq_waiter) {
        /* Another thread is hashing or is already waiting on pq.
         * Wait on vq.
         */
        wait = 2;
      }
      else {
        /* Wait on pq */
        wait = 1;
        pg_snf->pq_group_pq_waiter = 1;
      }
    }

#if LOCK_PG
    mx_spin_unlock_irqrestore(&pg_snf->pq_lock, flags);
#else
    /* Release all spin locks */
    TAILQ_FOREACH(waiter, &ep_pg->pg_waitq, pg_link) {
      snf = waiter->snf;
      mx_spin_unlock(&snf->pq_lock);
    }
#endif

    /* Did we lock any port? */
    if (locked == 1) {
      /* Tell the user which port is locked */
      status = mx_copyout(&param, out, sizeof(param), es->is_kernel);
      if (status) {
        /* FIXME. Not sure what we should do */
      }
      status = 0;
    }
    else if (wait != 0) {
      if (wait == 1) {
        /* PQ */
        status = snf_pq_group_wait(es, param.timeout_ms);
        /* Do not bother locking.  Worst, all other threads wait on vq
         * without anyone to wake them up.  The current thread would likely hash
         * the pq again and wake up vq waiters.  Or, if it closes unexpectedly,
         * we unconditionally wake up all waiters.
         */
        pg_snf->pq_group_pq_waiter = 0;
        if (status == 0) {
          /* If we wake up early, try locking a port again to avoid coming into
           * the kernel right afterwards.  But this time, use timeout=0 so
           * we do not sleep repeatedly.
           */
          param.timeout_ms = 0;
          goto pq_group_lock_tryagain;
        }
      }
      else {
        /* VQ.  Wait on the first port for now.  es is the ring group
         * and IS on the first port.
         */
        status = snf_vqwait_req(es, param.timeout_ms);
      }
      if (status == EINTR) {
        /* timeout = -1.  EINTR is the expected return code. */
        /* status = EAGAIN; */
      }
      else if (status != EAGAIN)
        status = EBUSY;
    }
    else {
      /* Could not lock any port and did not go to sleep either. */
      status = EAGAIN;
    }
    return status;
  }
  else if (cmd == MYRI_SNF_PQ_GROUP_UNLOCK) {
    struct snf_endpt_portgroup *ep_pg;
    struct snf__pq_state_rw pq_rw;
    myri_snf_rx_state_t *pg_snf;
    unsigned long flags;

    flags = 0; /* -Wunused on non-linux */

    if (in != 0) {
      if ((status = mx_copyin(in, &pq_rw, sizeof(pq_rw), es->is_kernel)))
        return status;
    }

    pg_snf = snf; /* Group's first port */
    ep_pg = &es->snf.es_pg;
    snf = ep_pg->pq_locked_port;

#if LOCK_PG
    mx_spin_lock_irqsave(&pg_snf->pq_lock, flags);
#else
    /* Only need to acquire the port's spinlock */
    mx_spin_lock(&snf->pq_lock);
#endif

    /* Update the shared pq_s. */
    if (in != 0)
      memcpy(&snf->pq_s->rw, &pq_rw, sizeof(pq_rw));
    snf->pq_locked_es = NULL;
    status = 0;

#if LOCK_PG
    mx_spin_unlock_irqrestore(&pg_snf->pq_lock, flags);
#else
    mx_spin_unlock(&snf->pq_lock);
#endif

    /* Wake up vq waiters.  VQ waiters are on the first port. */
    snf_vqwake_req(pg_snf, 0);
  }

  if (cmd >= MYRI_SNF_PQ_TRYLOCK && cmd <= MYRI_SNF_PQ_UNLOCK_WAKE) {
    struct snf__pq_state_rw pq_rw;
    unsigned long flags;

    flags = 0; /* -Wunused on non-linux */

    if (cmd == MYRI_SNF_PQ_UNLOCK_WAKE && in != 0) {
      if ((status = mx_copyin(in, &pq_rw, sizeof(pq_rw), es->is_kernel)))
        return status;
    }

    mx_spin_lock_irqsave(&snf->pq_lock, flags);
    if (cmd == MYRI_SNF_PQ_TRYLOCK || cmd == MYRI_SNF_PQ_TRYLOCK_WAIT) {
      if (!snf->pq_locked_es) {
        snf->pq_locked_es = es;
        status = 0;
      }
      else
        status = EBUSY;
    }
    else {
      snf->pq_locked_es = NULL;
      status = 0;
      /* Update the shared pq_s */
      if (cmd == MYRI_SNF_PQ_UNLOCK_WAKE && in != 0)
        memcpy(&es->is->snf.pq_s->rw, &pq_rw, sizeof(pq_rw));
    }
    mx_spin_unlock_irqrestore(&snf->pq_lock, flags);

    if (cmd == MYRI_SNF_PQ_TRYLOCK_WAIT && status == EBUSY) {
      /* Trylock and then wait for vq in one syscall */
      uint32_t x;
      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (!status) {
        if (x == 0)
          status = EAGAIN;
        else {
          status = snf_vqwait_req(es, x);
          if (status == EINTR) {
            /* timeout = -1.  EINTR is the expected return code. */
            /* status = EAGAIN; */
          }
          else if (status != EAGAIN)
            status = EBUSY;
        }
      }
      /* If copyin fails? */      
    }
    else if (cmd == MYRI_SNF_PQ_UNLOCK_WAKE) {
      /* Unlock and wake vq waiters in one syscall */
      snf_vqwake_req(&es->is->snf, 0);
    }
    return status;
  }

  switch (cmd) {

  case MYRI_SNF_RX_START:
  case MYRI_SNF_RX_STOP:
    if (es->es_type != MYRI_ES_SNF_RX) {
      status = EINVAL;
      break;
    }
    mx_mutex_enter(&snf->sync);
    if (cmd == MYRI_SNF_RX_START) {
      if (snf->state != SNF_RX_S_STARTED) {
        status = myri_snf_start_stop(es, 1);
        snf->state = SNF_RX_S_STARTED;
      } else {
        status = 0;
      }
    }
    else {
      if (snf->state == SNF_RX_S_STARTED) {
        status = myri_snf_start_stop(es, 0);
        snf->state = SNF_RX_S_STOPPED;
      } else {
        status = 0;
      }
    }
    mx_mutex_exit(&snf->sync);
    break;

  case MYRI_SNF_WAIT:
    {
      uint32_t x;
      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
        break;

      if (es->es_type == MYRI_ES_SNF_TX)
        status = snf_eop_poll_wait(es, x, MYRI_WAIT_TX_EVENT);
      else if (es->es_type == MYRI_ES_SNF_RX_RING ||
               es->es_type == MYRI_ES_SNF_RX_GROUP ||
               es->es_type == MYRI_ES_SNF_RX_PQ)
        status = snf_eop_poll_wait(es, x, MYRI_WAIT_RX_EVENT);
      else
        status = EINVAL;
    }
    break;

  case MYRI_SNF_STATS:
    {
      myri_snf_stats_t x;
      status = myri_counters64_get(es->is, (uint64_t *) &x, 
                sizeof(x)/sizeof(uint64_t), MYRI_COUNTERS_GET_SNF);
      if (status == 0) {
        /* MCP counters handles these as kbytes */
        x.nic_bytes_recv *= 1024;
        x.nic_bytes_send *= 1024;
        status = mx_copyout(&x, out, sizeof(x), es->is_kernel);
      }
    }
    break;

  case MYRI_SNF_RX_SYNC:
    {
      struct snf__rx_sync x;

      if (es->es_type != MYRI_ES_SNF_RX_RING) {
        status = EINVAL;
        break;
      }
      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
	break;
      status = snf_rx_sync(snf, es->snf.es_rx.rx_ring_id, &x);
    }
    break;


  case MYRI_SNF_RX_GROUP_JOIN:
    {
      myri_snf_port_group_t x;

      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status == 0)
        status = snf_rx_portgroup_join(es, &x);
    }
    break;

  case MYRI_SNF_RX_DETACH:
    mx_mutex_enter(&snf->sync);
    status = myri_snf_rx_detach(es, 0);
    mx_mutex_exit(&snf->sync);
    break;

#if MAL_OS_LINUX || MAL_OS_FREEBSD || MAL_OS_WINDOWS
  case MYRI_SNF_SOFT_RX:
    {
      myri_soft_rx_t x;
      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
       break;
      status = myri_ether_soft_rx(es->is, &x);
    }
    break;
#endif /* MAL_OS_LINUX */
  }

  return status;
}

myri_endpt_ops_t myri_ops_snf;

void
myri_snf_init_ops(void)
{
  bzero(&myri_ops_snf, sizeof(myri_ops_snf));
  myri_ops_snf.flags = MYRI_EOP_F_POLL;
  myri_ops_snf.alloc = snf_eop_alloc;
  myri_ops_snf.enable = snf_eop_enable;
  myri_ops_snf.ioctl = snf_eop_ioctl;
  myri_ops_snf.mmap  = snf_eop_mmap;
  myri_ops_snf.close = snf_eop_close;
  myri_ops_snf.interrupt_is = snf_eop_intr_is;
  myri_ops_snf.interrupt_ep = snf_eop_intr_endpt;
  myri_ops_snf.poll_events_pending = snf_eop_poll_events_pending;
  myri_ops_snf.poll_prep_wait = snf_eop_poll_prep_wait;
  myri_ops_snf.poll_wait = snf_eop_poll_wait;
}
