/******************************************************************************
*                                                                             *
* License Agreement                                                           *
*                                                                             *
* Copyright (c) 2003 Altera Corporation, San Jose, California, USA.           *
* All rights reserved.                                                        *
*                                                                             *
* Permission is hereby granted, free of charge, to any person obtaining a     *
* copy of this software and associated documentation files (the "Software"),  *
* to deal in the Software without restriction, including without limitation   *
* the rights to use, copy, modify, merge, publish, distribute, sublicense,    *
* and/or sell copies of the Software, and to permit persons to whom the       *
* Software is furnished to do so, subject to the following conditions:        *
*                                                                             *
* The above copyright notice and this permission notice shall be included in  *
* all copies or substantial portions of the Software.                         *
*                                                                             *
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  *
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    *
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE *
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      *
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     *
* FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         *
* DEALINGS IN THE SOFTWARE.                                                   *
*                                                                             *
* This agreement shall be governed in all respects by the laws of the State   *
* of California and by the laws of the United States of America.              *
*                                                                             *
* Modified:                                                                   *
*  9/29./14 G. Hannon (ghannon@cspi.com) This is now a combination of the     *
*  content of various Altera header files for interfacing with the Avalon     *
*  cores.                                                                     *
*  12/29/15 G. Hannon (ghannon@cspi.com) incorporated routines for Xilinx     *
*  fpga based board differences.                                              *
******************************************************************************/

#ifndef __PHX_AVL_H__
#define __PHX_AVL_H__

#include "mal_stbar.h"

static inline void
fpga_read8s(char* base, uint64_t address, uint8_t* buffer, unsigned int count)
{
        MAL_READBAR(); MAL_WRITEBAR();
        while (count) {
                void *p = base + address;
                *(uint8_t*)buffer = *(uint8_t*)p;
                address += 1;
                count -= 1;
                buffer += 1;
        }
        MAL_READBAR(); MAL_WRITEBAR();
}

static inline void
fpga_read32s(char* base, uint64_t address, uint32_t* buffer, unsigned int count)
{
        MAL_READBAR(); MAL_WRITEBAR();
        while (count) {
                void *p = base + address;
                *(uint32_t*)buffer = *(uint32_t*)p;
                address += 4;
                count -= 1;
                buffer += 4;
        }
        MAL_READBAR(); MAL_WRITEBAR();
}

static inline void
fpga_write8s(char* base, uint64_t address, uint8_t* buffer, unsigned int count)
{
        MAL_READBAR(); MAL_WRITEBAR();
        while (count) {
                void *p = base + address;
                *(uint8_t*)p = *(uint8_t*)buffer;
                address += 1;
                count -= 1;
                buffer += 1;
        }
        MAL_READBAR(); MAL_WRITEBAR();
}


static inline void
fpga_write32s(char* base, uint64_t address, uint32_t *buffer, size_t count)
{
    MAL_READBAR(); MAL_WRITEBAR();
    while (count) {
        void *p = base + address;
        *(uint32_t*)p = *(uint32_t*)buffer;
        address += 4;
        count -= 1;
        buffer += 4;
    }
    MAL_READBAR(); MAL_WRITEBAR();
}

static inline uint32_t IORD(char* base, uint32_t reg)
{
        uint32_t val;

        fpga_read32s(base, (reg << 2), &val, 1);
        return val;
}

static inline void IOWR(char* base, uint32_t reg, uint32_t val)
{
        fpga_write32s(base, (reg << 2), &val, 1);
}

/* 
 * Registers for PIO core
 */
#define IORD_ALTERA_AVALON_PIO_DATA(base)             IORD(base, 0) 
#define IOWR_ALTERA_AVALON_PIO_DATA(base, data)       IOWR(base, 0, data)

#define IORD_ALTERA_AVALON_PIO_DIRECTION(base)        IORD(base, 1) 
#define IOWR_ALTERA_AVALON_PIO_DIRECTION(base, data)  IOWR(base, 1, data)

#define IORD_ALTERA_AVALON_PIO_IRQ_MASK(base)         IORD(base, 2) 
#define IOWR_ALTERA_AVALON_PIO_IRQ_MASK(base, data)   IOWR(base, 2, data)

#define IORD_ALTERA_AVALON_PIO_EDGE_CAP(base)         IORD(base, 3) 
#define IOWR_ALTERA_AVALON_PIO_EDGE_CAP(base, data)   IOWR(base, 3, data)

#define IORD_ALTERA_AVALON_PIO_SET_BITS(base)         IORD(base, 4) 
#define IOWR_ALTERA_AVALON_PIO_SET_BITS(base, data)   IOWR(base, 4, data)

#define IORD_ALTERA_AVALON_PIO_CLEAR_BITS(base)         IORD(base, 5) 
#define IOWR_ALTERA_AVALON_PIO_CLEAR_BITS(base, data)   IOWR(base, 5, data)

/* Defintions for direction-register operation with bi-directional PIOs */
#define ALTERA_AVALON_PIO_DIRECTION_INPUT  0
#define ALTERA_AVALON_PIO_DIRECTION_OUTPUT 1


int avl_pio_init(char* base);
int avl_pio_read(char* base, uint32_t mask, uint32_t* value);
int avl_pio_write(char* base, uint32_t mask, uint32_t value);
int phx_avl_init(char* base, uint32_t* version);

/* 
 * Registers/commands for spi cores
 */

#define ALTERA_AVALON_SPI_RXDATA_REG                    0
#define IORD_ALTERA_AVALON_SPI_RXDATA(base)       IORD(base, ALTERA_AVALON_SPI_RXDATA_REG) 
#define IOWR_ALTERA_AVALON_SPI_RXDATA(base, data) IOWR(base, ALTERA_AVALON_SPI_RXDATA_REG, data)

#define ALTERA_AVALON_SPI_TXDATA_REG                    1
#define IORD_ALTERA_AVALON_SPI_TXDATA(base)       IORD(base, ALTERA_AVALON_SPI_TXDATA_REG) 
#define IOWR_ALTERA_AVALON_SPI_TXDATA(base, data) IOWR(base, ALTERA_AVALON_SPI_TXDATA_REG, data)

#define ALTERA_AVALON_SPI_STATUS_REG                    2
#define IORD_ALTERA_AVALON_SPI_STATUS(base)       IORD(base, ALTERA_AVALON_SPI_STATUS_REG) 
#define IOWR_ALTERA_AVALON_SPI_STATUS(base, data) IOWR(base, ALTERA_AVALON_SPI_STATUS_REG, data)

#define ALTERA_AVALON_SPI_STATUS_ROE_MSK              (0x8)
#define ALTERA_AVALON_SPI_STATUS_ROE_OFST             (3)
#define ALTERA_AVALON_SPI_STATUS_TOE_MSK              (0x10)
#define ALTERA_AVALON_SPI_STATUS_TOE_OFST             (4)
#define ALTERA_AVALON_SPI_STATUS_TMT_MSK              (0x20)
#define ALTERA_AVALON_SPI_STATUS_TMT_OFST             (5)
#define ALTERA_AVALON_SPI_STATUS_TRDY_MSK             (0x40)
#define ALTERA_AVALON_SPI_STATUS_TRDY_OFST            (6)
#define ALTERA_AVALON_SPI_STATUS_RRDY_MSK             (0x80)
#define ALTERA_AVALON_SPI_STATUS_RRDY_OFST            (7)
#define ALTERA_AVALON_SPI_STATUS_E_MSK                (0x100)
#define ALTERA_AVALON_SPI_STATUS_E_OFST               (8)

#define ALTERA_AVALON_SPI_CONTROL_REG                 3
#define IORD_ALTERA_AVALON_SPI_CONTROL(base)          IORD(base, ALTERA_AVALON_SPI_CONTROL_REG) 
#define IOWR_ALTERA_AVALON_SPI_CONTROL(base, data)    IOWR(base, ALTERA_AVALON_SPI_CONTROL_REG, data)

#define ALTERA_AVALON_SPI_CONTROL_IROE_MSK            (0x8)
#define ALTERA_AVALON_SPI_CONTROL_IROE_OFST           (3)
#define ALTERA_AVALON_SPI_CONTROL_ITOE_MSK            (0x10)
#define ALTERA_AVALON_SPI_CONTROL_ITOE_OFST           (4)
#define ALTERA_AVALON_SPI_CONTROL_ITRDY_MSK           (0x40)
#define ALTERA_AVALON_SPI_CONTROL_ITRDY_OFS           (6)
#define ALTERA_AVALON_SPI_CONTROL_IRRDY_MSK           (0x80)
#define ALTERA_AVALON_SPI_CONTROL_IRRDY_OFS           (7)
#define ALTERA_AVALON_SPI_CONTROL_IE_MSK              (0x100)
#define ALTERA_AVALON_SPI_CONTROL_IE_OFST             (8)
#define ALTERA_AVALON_SPI_CONTROL_SSO_MSK             (0x400)
#define ALTERA_AVALON_SPI_CONTROL_SSO_OFST            (10)

#define ALTERA_AVALON_SPI_SLAVE_SEL_REG               5
#define IORD_ALTERA_AVALON_SPI_SLAVE_SEL(base)        IORD(base, ALTERA_AVALON_SPI_SLAVE_SEL_REG) 
#define IOWR_ALTERA_AVALON_SPI_SLAVE_SEL(base, data)  IOWR(base, ALTERA_AVALON_SPI_SLAVE_SEL_REG, data)
/*
 * Macros used by alt_sys_init
 */

#define ALTERA_AVALON_SPI_INSTANCE(name, device) extern int alt_no_storage
#define ALTERA_AVALON_SPI_INIT(name, device) while (0)

/*
 * Use this function to perform one SPI access on your target.  'base' should
 * be the base address of your SPI peripheral, while 'slave' indicates which
 * bit in the slave select register should be set.
 */

/* If you need to make multiple accesses to the same slave then you should 
 * set the merge bit in the flags for all of them except the first.
 */
#define ALT_AVALON_SPI_COMMAND_MERGE (0x01)

/*
 * If you need the slave select line to be toggled between words then you
 * should set the toggle bit in the flag.
 */
#define ALT_AVALON_SPI_COMMAND_TOGGLE_SS_N (0x02)

int avl_spi_command(char* base, uint32_t slave,
                           uint32_t write_length, const uint8_t * write_data,
                           uint32_t read_length, uint8_t * read_data,
                           uint32_t flags);

/*
 * Definitions for I2C communications
 */

// Open Core Reg Offsets
#define I2C_CTR_LO      0x00
#define I2C_CTR_HI      0x04
#define I2C_CONTROL     0x08
#define I2C_RX_TX       0x0C
#define I2C_CMD_STAT    0x10
#define I2C_ENABLE      0x80
#define I2C_DISABLE     0x00

// Open Core Command Words
#define I2C_START_WR     0x90
#define I2C_STOP_WR      0x50
#define I2C_WR           0x10
#define I2C_RD           0x20
#define I2C_STOP_RD_NACK 0x68

// The driver maps a page starting at the base of the i2c cores and passes 
// the starting address of the mapped region to avl routines.
#define AVL_BASE_I2C_0             (0x0000) /* MAC-ID PROM */
#define AVL_BASE_I2C_1             (0x0020) /* QSFP-1 */
#define AVL_BASE_I2C_2             (0x0040) /* VSC7111-1 */
#define AVL_BASE_I2C_3             (0x0060) /* VSC7224-1 */
#define AVL_BASE_I2C_4             (0x0080) /* QSFP-2 */
#define AVL_BASE_I2C_5             (0x00A0) /* VSC7111-2 */
#define AVL_BASE_I2C_6             (0x00C0) /* VSC7224-2 */

// Known PIO cores
#define AVL_BASE_PIO_O             (0x0100) // Qsys control register (mapped at top level)
#define AVL_BASE_PIO_I             (0x0120) // Qsys status  register (mapped at top level)

// General Registers
#define AVL_BASE_ID                (0x0140) // ID register 
#define AVL_BASE_VER               (0x0160) // AVR Version register
#define AVL_SHA1_VER               (0x0180) // Firmware Version register

#define AVL_ID_REGVAL              0x12BA 	// BittWare's PCI Vendor ID

// Known I2C device addresses on Phoenix
#define DEV_7224   	0x20
#define DEV_7111   	0x00
#define DEV_QSFP   	0xA0
#define DEV_SPECS   	0xA0
#define DEV_LICENSE	0xA2
#define DEV_SERNUM 	0xB0

// Xilinx General Registers relative to I2C offset
#define XLX_PORT_LINK_LEDS         (0x1000) // Link LEDs (FPGA controlled)
#define XLX_PORT_ACTV_LEDS         (0x1004) // Activity LEDs (FPGA controlled)
#define XLX_STATUS_LEDS            (0x1008) // Green Status LEDs
#define XLX_SFP_PRESENT            (0x100C) // SFP presence detected register
#define XLX_SFP_LOS                (0x1010) // SFP LOS state register
#define XLX_SFP_TX_FAULT           (0x1014) // SFP TXFAULT condition register
#define XLX_SFP_TX_DISABLE         (0x1018) // SFP TX disable control register
#define XLX_FAN_PWM_PERCENT        (0x101C) // Control heatsink fan speed
#define XLX_FAN_PWM                (0x1020) // Read heatsink pwm fan speed
#define XLX_I2C_MUX_SELECT         (0x1024) // Allows remote i2c control on P9 connector (Aardvark)
#define XLX_NOR_FLASH_WP           (0x1028) // Flash write protection (1 = unprotected)
#define XLX_FPGA_PROG              (0x102C) // Force FPGA load from Flash
#define XLX_ASCII_ID_0             (0x1030) // "CSPi"
#define XLX_ASCII_ID_1             (0x1034) // "KU35" or "KU40"
#define XLX_ASCII_ID_2             (0x1038) // "DBLX" or "SNFX"
#define XLX_ASCII_ID_3             (0x103C) // "2P10"
#define XLX_ASCII_ID_4             (0x1040) // "G..."
#define XLX_FPGA_REV               (0x1044) // Board and Firmware Revisions
#define XLX_FPGA_REV_BRANCH        (0x1048) // Firmware Branch
#define XLX_MONTH_DAY              (0x104C) // Bitstream creation date
#define XLX_TIME                   (0x1050) // Bitstream creation time

// Used to differentiate between KCU105 eval board and K35.
#define KCU105_ID_REGVAL           0x4b553430   // KU40 - KCU105 eval board
#define K35_ID_REGVAL              0x4b553335   // KU35 - K35 boards

// Known I2C device addresses on K35.
#define K35_DEV_EEPROM   	0xA8
#define K35_DEV_MUXx8   	0xE2    // TI tca9548a - x8 mux

// Known I2C device addresses on KCU105.
#define KCU105_DEV_SPECS   	0xAC
#define KCU105_DEV_LICENSE	0xAE
#define KCU105_DEV_SERNUM 	0xA8
#define KCU105_DEV_MUXx4	0xEA    // TI pca9544a - x4 mux
#define KCU105_DEV_MUXx8   	0xE8    // TI tca9548a - x8 mux

// Known I2C device addresses for Finisar RJ45 transceiver
#define K35_DEV_MARVELL         0xAC

int avl_i2c_init(char* base, uint32_t avl_base, uint8_t high, uint8_t low);
int avl_i2c_read_bytes(char* base, uint32_t avl_base, uint8_t device, uint8_t reg_ofs, uint8_t* bytes, uint8_t count);
int avl_i2c_write_bytes(char* base, uint32_t avl_base, uint8_t device, uint8_t reg_ofs, uint8_t* bytes, uint8_t count);
int at24c64_i2c_read_bytes(char *base, uint32_t avl_base, uint8_t device, uint16_t reg_ofs, uint8_t* bytes, uint8_t count);
int avl_i2c_set_mux(char* base, uint8_t mux, int bus);
int phx_xilinx_init(char* base, uint32_t* version, uint32_t *xlx_board);

#endif /* __PHX_AVL_H__ */
