/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2011 by Myricom, Inc.  All rights reserved.          *
 * Copyright 2014 by CSP, Inc.  All rights reserved.                     *
 *************************************************************************/

#ifndef _PHX_DBL_FWD_H
#define _PHX_DBL_FWD_H

extern myri_endpt_ops_t phx_dbl_ops;
void phx_dbl_init_ops(void);

#define MYRI_DBL_POLLWAIT 0x1
#define MYRI_DBL_MXWAIT 0x2

typedef phx_pages_t phx_dbl_pages_t;

typedef struct {
  unsigned int size; /* all of gamma pages in bytes */
#if MAL_OS_WINDOWS
  /* Gamma pages are allocated differently on Windows. */
  void *gamma_mdl;
#endif
  phx_dbl_pages_t gamma; /* contains DMA data from NIC */
  phx_dbl_pages_t beta;	 /* 2nd level pages */
  phx_dbl_pages_t alpha; /* 1st level pages */
} phx_dbl_rx_ring_t;

struct phx_endpt_state_dbl {
  mx_spinlock_t sleep_lock;
  uint32_t dmasend_threshold;
  int rx_intr_requested;
  int tx_intr_requested;
  enum phx_dbl_filter_mode recvq_filter_mode;

  /* Channel IDs for data and descriptor APU channels */
  int apu_data_ch;
  int apu_desc_ch;

  /* soft endpoint support */
  int de_tag;
  mx_endpt_state_t *parent_es;
  mx_endpt_state_t **tag2es;
  int soft_intr_req_cnt;
  mx_endpt_state_t **soft_intr_req_es;
  int soft_data_pending;

  /* Below are rings and buffers that are mapped into the user. */

  /* Receive data ring in host memory */
  phx_dbl_rx_ring_t rx_data_ring;

  /* Receive descriptor ring in host memory */
  phx_dbl_rx_ring_t rx_desc_ring;

  /* Send DMA buffer in host memory */
  phx_dbl_pages_t tx_dma_buf;
#if MAL_OS_WINDOWS
  /* FIXME.  Should re-use an API like copyblock. */
  void *tx_dma_buf_mdl;
#endif

  /* Endpoint specific region in NIC */
  struct phx_nic_mem_endpoint *ep_mem;
  uint32_t ep_mem_size;

  /* Send PIO buffer on NIC */
  void *tx_pio_buf;
  uint32_t tx_pio_buf_size;

  /* Send descriptor buffer on NIC.  Its size is one descriptor. */
  void *tx_desc_buf;

};

struct phx_is_state_dbl {
  enum phx_dbl_filter_mode recvq_filter_mode;
  mx_endpt_state_t        *es_filter_owner;
};
#endif /* _PHX_DBL_FWD_H */

