/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2011 by Myricom, Inc.  All rights reserved.          *
 * Copyright 2015 - 2016 by CSP, Inc.  All rights reserved.              *
 *************************************************************************/

#ifndef _PHX_SNF_FWD_H
#define _PHX_SNF_FWD_H

#define SNF_PORTGROUP_MAX_MEMBERS 8

extern myri_endpt_ops_t phx_snf_ops;

void phx_snf_init_ops(void);

enum snf_rx_state { SNF_RX_S_FREE = 0,
                    SNF_RX_S_INIT = 1,
                    SNF_RX_S_STARTED = 2,
                    SNF_RX_S_STOPPED = 3 };

struct myri_snf_waiter;
STAILQ_HEAD(snf_waitq, myri_snf_waiter);

struct myri_snf_rx_state;
struct myri_snf_waiter {
  TAILQ_ENTRY(myri_snf_waiter) link;    /* waitq link */
  TAILQ_ENTRY(myri_snf_waiter) pg_link; /* portgroup link */
  TAILQ_ENTRY(myri_snf_waiter) wake_link; /* wakeup link */
  int                intr_requested;
  int                intr_wake_pending;
  struct myri_snf_rx_state *snf;
  mx_sync_t           *wait_sync;
  mx_endpt_state_t    *pg_attached_es_rx; /* ring added to a portgroup */
};

typedef
union snf_endpt_state {
  struct {
    int rx_is_attached;
    struct snf_dup_ma_app_state *app_info; /* DUP_MA mode. */
  } es_rx_h;

  struct {
    struct snf__rx_state  *rx_s;
    struct snf__pq_state  *pq_s;
    struct myri_snf_waiter rx_waitsync;
    struct myri_snf_waiter vq_waitsync;
    int                    rx_ring_id;
    mx_endpt_state_t      *rx_pg; /* If in a portgroup */
    struct snf_dup_ma_ring_state *app_ring_info; /* DUP_MA mode. */
  } es_rx;

  struct {
    mx_spinlock_t                tx_lock;
    int                          tx_intr_requested;
    union {
      struct es_tx_phx {
        struct phx_nic_mem_endpoint *tx_regs;       /* nic memory */
        phx_pages_t              tx_dma_pages;
#if MAL_OS_WINDOWS
        void                    *tx_dma_mdl;
#endif
        void                    *tx_pio_buf;    /* nic memory */
        uint32_t                 tx_pio_buf_size;
        void                    *tx_desc_buf;   /* nic memory */
      } phx;
      struct es_tx_k35snf {
        void             *tx_regs; /* nic memory */
        phx_rx_ring_t     desc_ring;
        phx_rx_ring_t     data_ring;
        struct {
          mx_page_pin_t   pin;
          char           *addr;
        } completion;
        int               apu_data_ch;
        int               apu_desc_ch;
#if K35SNF_FLUSH_STRATEGY == K35SNF_FLUSH_INTERRUPT
        struct k35snf__tx_shared_params sp;
#endif
      } k35snf;
    } u;
  } es_tx;

  struct snf_endpt_portgroup {
    /* One ring per board, max */
    struct myri_snf_waiter       pg_waiters[SNF_PORTGROUP_MAX_MEMBERS];
    TAILQ_HEAD(,myri_snf_waiter) pg_waitq;
    /* prehash + aggregation mode */
    struct myri_snf_rx_state    *pq_locked_port;
    struct myri_snf_waiter       vq_waitsync;
    uint32_t                     port_mask;
    /* The endpoint's user va. Used to check if the ring is empty.
     * The index is the port number.
     */
    void                        *evq_base[SNF_PORTGROUP_MAX_MEMBERS];
  } es_pg;
}
snf_endpt_state_t;

struct snf__rx;

/* Per-instance sniffer state, basically representing the single physical
 * receive ring.
 */
typedef struct myri_snf_rx_state {
  mx_instance_state_t   *is;
  mx_sync_t              sync;
  uint32_t               owner_pid;
  uint32_t               num_rings_attached;
  uint32_t		 num_rings_requested;
  int                    refcount;
  enum snf_rx_state      state;
  myri_snf_rx_params_t   params;
  int                    using_hw_hash; /* valid when not using APP_ID */

  struct snf__rx_state  *rx_s;
  unsigned               rx_s_length;
  mx_spinlock_t          pq_lock;
  struct mx_endpt_state *pq_locked_es;
  struct snf__pq_state  *pq_s;
  unsigned               pq_s_length;
  int                    pq_refcount;
  int                    pq_group_pq_waiter;
  uint32_t		 rx_desc_ring_count;

  /* Phoenix flow control registers */
  volatile uint32_t     *desc_ring_consume;
  volatile uint32_t     *data_ring_consume;

  const struct phx_physical_ep *phx_ep; /* phoenix endpoint for the receive ring */

  /* Channel IDs for data and descriptor APU channels */
  int apu_data_ch;
  int apu_desc_ch;

  /* Data and descriptor rings (pages) */
  phx_rx_ring_t desc_ring;
  phx_rx_ring_t data_ring;
  phx_rx_ring_t prehash_desc_ring;

  /* State for synchronizing with NIC */
  struct snf_sync_state {
    mx_spinlock_t             rs_lock;
    uint32_t		      attach_mask;
    uint32_t		      detach_mask;
    struct snf__rx_sync       rs;
    struct snf__rx_sync       rs_last_update; /* Phoenix needs deltas (pages) */
    struct snf__rx_sync	      r_rs[MYRI_SNF_MAX_RINGS];
    mx_spinlock_t             r_rs_lock[MYRI_SNF_MAX_RINGS];
  } sync_s;

  mx_endpt_state_t *ring_map_ep[MYRI_SNF_MAX_RINGS];

  /* RX interrupts and virtual queue waiters */
  TAILQ_HEAD(myri_snf_waitq, myri_snf_waiter)
                  rx_intr_waitq, vq_waitq;
  int             rx_intr_requested;
  mx_spinlock_t   rx_intr_lock;

  /* DUP_MA mode.  Per-process/application ring information. */
  struct snf_dup_ma_app_state {
    int app_id;
    int refcount;
    int using_hw_hash;
    uint32_t num_rings_requested;
    uint32_t rss_flags;
    uint32_t num_rings_attached;
    uint32_t attach_mask;
  } app_info[MYRI_SNF_MAX_RINGS];
  struct snf_dup_ma_ring_state {
    int app_info_idx;      /* index into app_info */
    uint32_t port_ring_id; /* port-wide id used in the driver */
    uint32_t hash_ring_id; /* per-application id, used for hashing only */
  } app_ring_info[MYRI_SNF_MAX_RINGS];

  /* Stats at the start of sniffer.  We return diff(current stats, start stats).
   * These are error prone, inaccurate.  FIXME.
   */
  phx_snf_stats_t stats_at_start;

  /* SNF and Ethernet kernel interface both enable/disable MAC promiscuous mode.
   * Use a spin lock and flags to serialize the access.
   */
#define SNF_RX_PROMISC_FLAG_SNF_OWNED       (1 << 0) /* SNF is using the port */
#define SNF_RX_PROMISC_FLAG_REQUEST_ENABLE  (1 << 1) /* Eth wants promisc mode */
  int promisc_flags;
  mx_spinlock_t promisc_lock;

  /* HW hash module status */
  struct {
    uint32_t rss_flags; /* snf.h rss_flags from the app */
    int in_use; /* SNF_RX endpoints currently using hw */
  } hw_hash;
} myri_snf_rx_state_t;

void     myri_snf_rx_teardown(mx_endpt_state_t *es);
uint32_t myri_snf_intr_req(mx_endpt_state_t *es, uint32_t intr_flags);
int      myri_snf_intr_tx_try_flush(mx_endpt_state_t *es);
int      myri_snf_find_node_id(mx_endpt_state_t *es);

/* OS-specific */
void myri_snf_handle_tx_intr(mx_endpt_state_t *es);

/* The Ethernet driver wants to change the promiscuous mode.
 * If sniffer is running, remember the requested mode but do not change it.
 * Later on when sniffer stops, we set the requested mode.
 */
void phx_snf_set_ether_promisc(mx_instance_state_t *is, int enable);

#endif /* _PHX_SNF_FWD_H */

