/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014 - 2016 by CSP, Inc.  All rights reserved.              *
 *************************************************************************/

#include "mx_arch.h"
#include "mx_misc.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "phx_common.h"
#include "phx_ether_common.h"
#include "mx_pio.h"
#include "phx_avl.h"

void
phx_ether_copy_pkthdr(phx_ether_pkt_frag_t frags[], int n_frags, char *dst, int len)
{
	int i, copy_len, off;

	i = 0;
	off = 0;
	while (len > 0 && i < n_frags) {
		copy_len = len;
		if (copy_len > frags[i].len)
			copy_len = frags[i].len;
		memcpy(dst, frags[i].vaddr + off, copy_len);
		dst += copy_len;
		len -= copy_len;
		off += copy_len;
		if (off == frags[i].len) {
			i++;
			off = 0;
		}
	}
}

/* FIXME.  PIO buffer.  Do we always write to the same offset?
 * At the start of the buffer region on the NIC?  If so, we do not head
 * and tail to compute free space and offset.
 */
/* It looks like we always have to write to offset 0.  Need to
 * confirm.  We do not know the actual buffer size.  Until we are
 * sure, keep the old code around.
 * FIXME.
 */
int
phx_ether_tx_alloc_piobuf(struct phx_ether *eth, int len, uint32_t *start_offset)
{
	phx_ether_tx_t *tx;
	uint32_t tail_space, head_space, required;
	int used;

	/* PIO buffer.  The packet may not wrap. */
	/* FIXME.  Clean up conditionals. */

	tx = &eth->myri_phx(tx);
	tail_space = tx->pio_buffer_size - (tx->pio_offset_host % tx->pio_buffer_size);
	head_space = tx->pio_offset_nic % tx->pio_buffer_size;
	/* A multiple of 64 to help WC performance */
	required = ((len + 64 - 1) & ~(64 - 1));

#if 0
	used = (int)tx->pio_offset_host - (int)tx->pio_offset_nic;
	if (used + required > tx->pio_buffer_size)
		return -1;
	tx->pio_offset_host += required;
	*start_offset = 0;
	return 0;
#else
	used = (int)tx->pio_offset_host - (int)tx->pio_offset_nic;
	if (used + required > tx->pio_buffer_size ||
	    (required > tail_space && required > head_space))
		return -1;

	if (required > tail_space) {
		/* Not enough space in the tail.  Start from the beginning. */
		*start_offset = 0;
		tx->pio_offset_host += tail_space + required;
	}
	else {
		*start_offset = tx->pio_offset_host % tx->pio_buffer_size;
		tx->pio_offset_host += required;
	}
	return 0;
#endif
}

int
phx_ether_tx_alloc_dmabuf(struct phx_ether *eth, int len, uint32_t *start_offset)
{
	phx_ether_tx_t *tx;
	uint32_t required, off;
	int used;

	/* DMA buffer.  The packet may not wrap around the buffer end. */

	tx = &eth->myri_phx(tx);
	/* FIXME.  Release2-rc3.  DMA offset is a multiple of 64. */
	required = ((len + 64 - 1) & ~(64 - 1));
	used = (int)tx->dma_offset_host - (int)tx->dma_offset_nic;
	if (used + required > tx->dma_buffer_size) {
		MX_INFO(("No DMA buffer. used=%u required=%u\n", used, required));
		return -1;
	}
	off = tx->dma_offset_host % tx->dma_buffer_size;
	if (off + required > tx->dma_buffer_size) {
		/* Wrap and start at offset 0 */
		required += (tx->dma_buffer_size - off);
		off = 0;
		if (used + required > tx->dma_buffer_size) {
			MX_INFO(("No DMA buffer after wrap. used=%u required=%u\n", used, required));
			return -1;
		}
	}
	*start_offset = off;
	tx->dma_offset_host += required;
	return 0;
}

int
phx_ether_tx_copy_pkt_piobuf(struct phx_ether *eth, phx_ether_pkt_frag_t frags[],
			     int n_frags, int data_offset, int len, int buf_offset)
{
	phx_ether_tx_t *tx;
	char *dst;
	int i, frag_off, err, staging_len;
	uint8_t *staging;

#if MX_DEBUG_ETHER
	MX_INFO(("%s: n_frags=%d data_offset=%d len=%d buf_offset=%d\n", __FUNCTION__,
		 n_frags, data_offset, len, buf_offset));
#endif

	tx = &eth->myri_phx(tx);
	/* Skip data_offset bytes */
	err = 0;
	i = 0;
	frag_off = 0;
	while (data_offset > 0 && i < n_frags) {
		if (data_offset < frags[i].len) {
			/* frag_off = frags[i].len - data_offset; */
			frag_off = data_offset;
			data_offset = 0;
		}
		else {
			data_offset -= frags[i].len;
			i++;
		}
	}
	if (data_offset > 0) {
		err = -1;
		goto exit;
	}

	/* Now copy */
	staging = tx->aligned_data;
	staging_len = len;
	while (len > 0 && i < n_frags) {
		int rem = frags[i].len - frag_off;
		int copy_len = len;
		if (len > rem)
			copy_len = rem;
		memcpy(staging, frags[i].vaddr + frag_off, copy_len);
		staging += copy_len;
		len -= copy_len;
		frag_off = 0;
		i++;
	}
	if (len > 0)
		err = -1;

	/* Now copy from the staging buffer to the NIC.
	 * PIO buffer is contiguous.  Copy 8B at a time.
	 * Flush every 64B.
	 */
	dst = tx->pio_fifo + buf_offset;
#if MX_DEBUG_ETHER
        MX_INFO(("%s: dst =%p\n", __FUNCTION__, dst));
#endif
	staging_len = (staging_len + 7) & ~7;
	staging = tx->aligned_data;
	while (staging_len > 0) {
		*(uint64_t*)dst = *(uint64_t*)staging;
		dst += 8;
		staging += 8;
		staging_len -= 8;
		if (((uintptr_t)staging & 63) == 0)
			MAL_STBAR();
	}
	/* The caller does MAL_STBAR() so skip it here. */

exit:
	return err;
}

void
phx_copy_from_page_ring(phx_ether_pages_t *pages, uint32_t start_off, uint32_t len, char *dst)
{
	mx_page_pin_t *p;
	uint32_t p_idx, p_off, copy_len;

	/* Start page */
	p_idx = (start_off / PAGE_SIZE) % pages->n_pages;
	p_off = (start_off & (PAGE_SIZE-1));

	/* The data may span across multiple pages.  Visit as many pages as required. */
	while (len > 0) {
		p = &pages->pages[p_idx];
		copy_len = PAGE_SIZE - p_off;
		if (copy_len > len)
			copy_len = len;
		/* Copy out of the page to the caller's buffer. */
		memcpy(dst, (char*)(p->va + p_off), copy_len);
		dst += copy_len;
		len -= copy_len;
		p_off = 0;
		p_idx = (p_idx + 1) % pages->n_pages;
	}
}

int
phx_ether_tx_copy_pkt_dmabuf(struct phx_ether *eth, phx_ether_pkt_frag_t frags[],
			     int n_frags, int data_offset, int len, int buf_offset)
{
	phx_ether_pages_t *pages;
	mx_page_pin_t *p;
	uint32_t p_idx, p_off, frag_off;
	int i, err;

#if MX_DEBUG_ETHER
	MX_INFO(("%s: n_frags=%d data_offset=%d len=%d buf_offset=%d\n",
		 __FUNCTION__, n_frags, data_offset, len, buf_offset));
#endif

	/* Packet buffer: skip data_offset bytes */
	err = 0;
	i = 0;
	frag_off = 0;
	while (data_offset > 0 && i < n_frags) {
		if (data_offset < frags[i].len) {
			frag_off = data_offset;
			data_offset = 0;
		}
		else {
			data_offset -= frags[i].len;
			i++;
		}
	}
	if (data_offset > 0) {
		err = -1;
		goto exit;
	}

	/* DMA buffer: start page */
	pages = &eth->myri_phx(tx).dma;

	p_idx = (buf_offset / PAGE_SIZE) % pages->n_pages;
	p_off = (buf_offset & (PAGE_SIZE-1));

	/* Now copy.  Both the packet and the DMA buffer may span across
	 * multiple frags or pages.
	 */
	while (len > 0 && i < n_frags) {
		int rem, copy_len;

		/* Want to copy min(len, frag, page) */
		/* Check frag and len */
		rem = frags[i].len - frag_off;
		copy_len = (len > rem) ? rem : len;
		/* Check page */
		p = &pages->pages[p_idx];
		rem = PAGE_SIZE - p_off;
		if (copy_len > rem)
			copy_len = rem;

		/* Copy from frags to page */
		memcpy((char*)(p->va + p_off), frags[i].vaddr + frag_off, copy_len);

		len -= copy_len;
		/* Advance frag pointer */
		frag_off += copy_len;
		if (frag_off == frags[i].len) {
			i++;
			frag_off = 0;
		}
		/* Advance page pointer */
		p_off += copy_len;
		if (p_off == PAGE_SIZE) {
			p_idx = (p_idx + 1) % pages->n_pages;
			p_off = 0;
		}
	}
	if (len > 0)
		err = -1;
exit:
	return err;
}

static uint32_t
roundup_rx_ring(uint32_t mb, uint32_t min, uint32_t max)
{
	uint32_t size;
	size = mb << 20;
	if (size <= min)
		size = min;
	if (size > max)
		size = max;
	size = MAL_ROUND_UP(size, min); /* assume min is a power of 2 */
	return size;
}

static void
free_phx_ether_pages(struct phx_ether *eth, phx_ether_pages_t *p)
{
	mx_page_pin_t *pages;
	char *addr;
	int i;

	if (p && p->pages) {
		pages = p->pages;
		for (i = 0; i < p->n_pages; i++) {
			if (pages[i].va) {
				addr = (char*)pages[i].va;
				mx_free_dma_page(eth->is, &addr, &pages[i]);
				memset(&pages[i], 0, sizeof(pages[i]));
			}
		}
		mx_kfree(pages);
		memset(p, 0, sizeof(*p));
	}
}

static int
alloc_phx_ether_pages(struct phx_ether *eth, phx_ether_pages_t *p, int n_pages)
{
	mx_page_pin_t *pages;
	char *addr;
	int i;

#if MX_DEBUG_ETHER
	MX_INFO(("%s: allocate rx pages. n_pages=%d\n", eth->is->is_name, n_pages));
#endif
	pages = mx_kmalloc(sizeof(mx_page_pin_t) * n_pages, MX_MZERO | MX_NOWAIT);
	if (!pages)
		goto abort;
	p->pages = pages;
	p->n_pages = n_pages;
	for (i = 0; i < n_pages; i++) {
		if (mx_alloc_dma_page(eth->is, &addr, &pages[i]))
			goto abort;
		memset(addr, 0, PAGE_SIZE);
	}
	return 0;

abort:
	free_phx_ether_pages(eth, p);
	return -ENOMEM;
}

static void
setup_apu_alpha_beta_pages(struct phx_ether *eth, phx_ether_rx_ring_t *ring)
{
	int n_beta, n_alpha, i, j, k, rem;
	mx_page_pin_t *page, *alpha_page;
	phx_dma_addr_t *addr;
	phx_apu_flp_t *flp, *prev_flp;

	/* Fill in beta pages */
	n_beta = ring->beta.n_pages;
	page = ring->gamma.pages;
	for (i = 0; i < n_beta; i++) {
		addr = (phx_dma_addr_t *)ring->beta.pages[i].va;
		for (j = 0; j < PHX_APU_GAMMA_PAGES_PER_BETA; j++) {
			addr->low = page->dma.low;
			addr->high = page->dma.high;
			addr++;
			page++;
		}
	}

	/* Fill in alpha pages */
	n_alpha = ring->alpha.n_pages;
	page = ring->beta.pages;
	rem = n_beta;
	prev_flp = NULL;
	for (i = 0; rem > 0 && i < n_alpha; i++) {
		alpha_page = &ring->alpha.pages[i];
		flp = (phx_apu_flp_t *)alpha_page->va;
		for (j = 0; rem > 0 && j < PHX_APU_FLPS_PER_ALPHA; j++) {
			for (k = 0; rem > 0 && k < PHX_APU_BETA_PAGES_PER_FLP; k++) {
				flp->beta[k].low = page->dma.low;
				flp->beta[k].high = page->dma.high;
				rem--;
				page++;
			}
			/* Link FLPs.  The previous FLP points to the current FLP. */
			if (prev_flp) {
				prev_flp->next.low = alpha_page->dma.low;
				prev_flp->next.high = alpha_page->dma.high;
				prev_flp->next.low += ((uintptr_t)flp - alpha_page->dma.low);
			}
			flp->opcode = 0;
			PHX_APU_FLP_OPCODE_SET_NUM_BETA(flp->opcode, k);
			prev_flp = flp;
			flp++;
		}
	}
	/* The last/final FLP points back to the first FLP. */
	alpha_page = &ring->alpha.pages[0];
	prev_flp->next.low = alpha_page->dma.low;
	prev_flp->next.high = alpha_page->dma.high;
}

static int
alloc_rx_ring(struct phx_ether *eth)
{
	int n_gamma, n_beta, n_alpha;
	int flp_per_page = PAGE_SIZE / sizeof(phx_apu_flp_t);

	/* First, allocate pages for the data ring */
	n_gamma = eth->rx_data.ring.size / PAGE_SIZE;
	n_beta = (n_gamma + PHX_APU_GAMMA_PAGES_PER_BETA - 1) / PHX_APU_GAMMA_PAGES_PER_BETA;
	n_alpha = (n_beta + flp_per_page - 1) / flp_per_page;
	if (alloc_phx_ether_pages(eth, &eth->rx_data.ring.gamma, n_gamma))
		goto abort;
	if (alloc_phx_ether_pages(eth, &eth->rx_data.ring.beta, n_beta))
		goto abort;
	if (alloc_phx_ether_pages(eth, &eth->rx_data.ring.alpha, n_alpha))
		goto abort;
	setup_apu_alpha_beta_pages(eth, &eth->rx_data.ring);

	/* Second, allocate pages for the descriptor ring */
	n_gamma = eth->rx_desc.ring.size / PAGE_SIZE;
	n_beta = (n_gamma + PHX_APU_GAMMA_PAGES_PER_BETA - 1) / PHX_APU_GAMMA_PAGES_PER_BETA;
	n_alpha = (n_beta + flp_per_page - 1) / flp_per_page;
	if (alloc_phx_ether_pages(eth, &eth->rx_desc.ring.gamma, n_gamma))
		goto abort;
	if (alloc_phx_ether_pages(eth, &eth->rx_desc.ring.beta, n_beta))
		goto abort;
	if (alloc_phx_ether_pages(eth, &eth->rx_desc.ring.alpha, n_alpha))
		goto abort;
	setup_apu_alpha_beta_pages(eth, &eth->rx_desc.ring);
	return 0;

abort:
	/* These check pointers internally.  It is okay to call them without checking
	 * NULL pointers.
	 */
	free_phx_ether_pages(eth, &eth->rx_data.ring.gamma);
	free_phx_ether_pages(eth, &eth->rx_data.ring.beta);
	free_phx_ether_pages(eth, &eth->rx_data.ring.alpha);
	free_phx_ether_pages(eth, &eth->rx_desc.ring.gamma);
	free_phx_ether_pages(eth, &eth->rx_desc.ring.beta);
	free_phx_ether_pages(eth, &eth->rx_desc.ring.alpha);
	return -ENOMEM;
}

static void
free_rx_ring(struct phx_ether *eth)
{
	free_phx_ether_pages(eth, &eth->rx_data.ring.gamma);
	free_phx_ether_pages(eth, &eth->rx_data.ring.beta);
	free_phx_ether_pages(eth, &eth->rx_data.ring.alpha);
	free_phx_ether_pages(eth, &eth->rx_desc.ring.gamma);
	free_phx_ether_pages(eth, &eth->rx_desc.ring.beta);
	free_phx_ether_pages(eth, &eth->rx_desc.ring.alpha);
}

static int
alloc_tx_ring(struct phx_ether *eth)
{
	phx_ether_tx_t *tx = &eth->myri_phx(tx);
	int i;

	tx->dma_buffer_size = PHX_TX_DMA_PAGES_PER_ENDPOINT * PAGE_SIZE;
	tx->dma_offset_host = 0;
	tx->dma_offset_nic = 0;

	tx->pio_buffer_size = eth->is->phx.pfi->tx_config.dbltx_pio_length;
	tx->pio_offset_host = 0;
	tx->pio_offset_nic = 0;
	tx->pio_fifo = (char *)eth->ep_pio_mem;

	tx->desc_fifo = phx_ep_desc_fifo(eth->ep_nic_mem);
	tx->desc_idx_host = 0;
	tx->desc_idx_nic = 0;

	tx->desc_count = phx_read_ep_reg_tx_desc_buffer_size(eth->ep_nic_mem);
#if MX_DEBUG_ETHER
	MX_INFO(("tx_desc_buffer_size=%d\n", tx->desc_count));
#endif
	tx->queued = mx_kmalloc(sizeof(*tx->queued) * tx->desc_count, MX_MZERO);
	if (tx->queued == NULL)
		goto abort;

	/* Send completion page shared by all endpoints */
	tx->completion = (phx_tx_completion_t*)eth->is->tx_done.addr;
	/* Set the starting count.  FIXME.  Remove later on.
	 * See phx_dbl_ops.c:dbl_phx_init().
	 */
	tx->completion->endpoint[eth->myri_phx(ep)->tx_ep].sent =
		phx_ep_tx_completion_count(eth->ep_nic_mem);

#if MX_DEBUG_ETHER
	MX_INFO(("Initial send completion count=%d\n",
		 tx->completion->endpoint[eth->ep.phx->tx_ep].sent));
#endif
	/* Set the start index numbers as well */
	tx->desc_idx_host = tx->desc_idx_nic = tx->completion->endpoint[eth->myri_phx(ep)->tx_ep].sent;
	/* Pages for the DMA buffer */
	if (alloc_phx_ether_pages(eth, &tx->dma, PHX_TX_DMA_PAGES_PER_ENDPOINT))
		goto abort;

	/* Push all DMA page addresses to the NIC. */
	phx_write_ep_reg_tx_addr_buffer_clear(eth->ep_nic_mem);
	for (i = 0; i < PHX_TX_DMA_PAGES_PER_ENDPOINT; i++) {
		phx_write_ep_reg_tx_addr_load(eth->ep_nic_mem, tx->dma.pages[i].dma.low,
					      tx->dma.pages[i].dma.high);
	}
	phx_write_ep_reg_tx_addr_buffer_finish(eth->ep_nic_mem);

	/* init_board() sets the completion page's DMA address as it
	 * is a per-instance register.  See phx_common.c:init_board().
	 */

	return 0;
abort:
	if (tx->queued) {
		mx_kfree(tx->queued);
		tx->queued = NULL;
	}
	return -ENOMEM;
}

static void
free_tx_ring(struct phx_ether *eth)
{
	phx_ether_tx_t *tx = &eth->myri_phx(tx);
	/* Free pages */
	free_phx_ether_pages(eth, &tx->dma);
	if (tx->queued) {
		mx_kfree(tx->queued);
		tx->queued = NULL;
	}
}

int
phx_ether_open_common(mx_instance_state_t *is, int use_interrupt, int mtu,
		      int rx_data_ring_size_mb, int rx_desc_ring_size_mb)
{
	struct phx_ether *eth = is->ether;
	struct phx_physical_ep *endpoint;
	int status;

	eth->rx_done_limit = myri_ether_rx_done_limit;
	eth->use_interrupt = use_interrupt;

	eth->myri_phx(tx).aligned_desc =
		(phx_tx_desc_t*)(((uintptr_t)eth->myri_phx(tx).desc_buf + 63) & ~63);
	eth->myri_phx(tx).aligned_data =
		(uint8_t*)(((uintptr_t)eth->myri_phx(tx).data_buf + 63) & ~63);

	endpoint = is->board_ops.ether_endpoint(is);
	if (phx_claim_fpga_endpoint(is, endpoint->id, 1))
		MX_INFO(("Could not claim endpoint %d for ethernet\n", endpoint->id));
	eth->myri_phx(ep) = endpoint;

	eth->ep_tx_stats = &is->phx.pfi->counters.ep_tx_stats[eth->myri_phx(ep)->id];

	MX_INFO(("%s: Using endpoint id=%d rx=%d tx=%d\n", is->is_name,
		 eth->myri_phx(ep)->id, eth->myri_phx(ep)->rx_ep, eth->myri_phx(ep)->tx_ep));

	eth->ep_nic_mem = phx_endpoint_nic_mem(is, eth->myri_phx(ep)->tx_ep);

	eth->ep_pio_mem = phx_ep_pio_fifo(is, eth->myri_phx(ep)->tx_ep);

	eth->apu_data_ch = phx_apu_channel_id(is, 1, eth->myri_phx(ep)->rx_ep, 0);
	eth->apu_desc_ch = phx_apu_channel_id(is, 0, eth->myri_phx(ep)->rx_ep, 0);

#if MX_DEBUG_ETHER
	MX_INFO(("APU channels: data=%d desc=%d\n",
		 eth->apu_data_ch, eth->apu_desc_ch));
#endif

	/* Round up rx_data_ring_mb to a sane value. */
	eth->rx_data.ring.size = roundup_rx_ring(rx_data_ring_size_mb,
						 MYRI_RX_DATA_RING_MB_MIN,
						 MYRI_RX_DATA_RING_MB_MAX);
#if MX_DEBUG_ETHER
	MX_INFO(("%s: rx_data_ring_size=%d\n", is->is_name, eth->rx_data.ring.size));
#endif

	/* Round up rx_desc_ring_mb to a sane value. */
	eth->rx_desc.ring.size = roundup_rx_ring(rx_desc_ring_size_mb,
						 MYRI_RX_DESC_RING_MB_MIN,
						 MYRI_RX_DESC_RING_MB_MAX);
#if MX_DEBUG_ETHER
	MX_INFO(("%s: rx_desc_ring_size=%d\n", is->is_name, eth->rx_desc.ring.size));
#endif

	/* Allocate receive ring: data and descriptor */
	if ((status = alloc_rx_ring(eth))) {
		MX_WARN(("%s: unable to allocate receive ring\n", is->is_name));
		goto abort_with_nothing;
	}

	/* Allocate send ring */
	if ((status = alloc_tx_ring(eth))) {
		MX_WARN(("%s: unable to allocate transmit ring\n", is->is_name));
		goto abort_with_receive_ring;
	}

	phx_ether_set_mtu(eth, mtu);

	mx_spin_lock_init(&eth->spinlock, is, -1, "ether spinlock");
	is->ether_is_open = 1;
	return 0;

abort_with_receive_ring:
	free_rx_ring(eth);
abort_with_nothing:
	return status;
}

int
phx_ether_start_common(struct phx_ether *eth)
{
	struct phx_nic_mem_global *nic;
	mx_page_pin_t *page;

	/* Set up the receive path.  Clear index vars first. */
	eth->rx_data.data_off = 0;
	eth->rx_data.cons_off = 0;
	eth->rx_desc.desc_idx = 0;
	eth->rx_desc.cons_idx = 0;
	eth->rx_desc.seqnum = 1;

	nic = eth->is->phx.nic_mem;

	/* Filter.  Nothing special to do for filters.
	 * init_board() initializes per-instance filter settings.
	 * FIXME.  Later on, add kernel_dup module param.
	 */
	/* phx_dbl_filter_init_hack(nic); */

#if MX_DEBUG_ETHER
	MX_INFO(("RX data page 0: high=0x%x low=0x%x\n",
		 eth->rx_data.ring.gamma.pages[0].dma.high,
		 eth->rx_data.ring.gamma.pages[0].dma.low));
	MX_INFO(("RX desc page 0: high=0x%x low=0x%x\n",
		 eth->rx_desc.ring.gamma.pages[0].dma.high,
		 eth->rx_desc.ring.gamma.pages[0].dma.low));
#endif

	/* following the reset procedure for endpoints from
	   the ReleaseNotes.txt */
	phx_stop_rx_endpoint(eth->is, eth->myri_phx(ep)->rx_ep);

	phx_apu_channel_reset(eth->is, eth->apu_data_ch);
	phx_apu_channel_reset(eth->is, eth->apu_desc_ch);

	page = eth->rx_data.ring.alpha.pages;
	phx_apu_channel_start(eth->is, eth->apu_data_ch,
			      page->dma.low, page->dma.high);
	page = eth->rx_desc.ring.alpha.pages;
	phx_apu_channel_start(eth->is, eth->apu_desc_ch,
			      page->dma.low, page->dma.high);
	/* Now the APUs are fetching addresses. */

	phx_reset_rx_endpoint(eth->is, eth->myri_phx(ep)->rx_ep);
	phx_start_rx_endpoint(eth->is, eth->myri_phx(ep)->rx_ep,
			      eth->rx_data.ring.gamma.n_pages,
			      eth->rx_desc.ring.gamma.n_pages);

	return 0;
}

void
phx_ether_close_common(struct phx_ether *eth)
{
	mx_instance_state_t *is = eth->is;
	const struct phx_physical_ep *ep = eth->myri_phx(ep);

	/* Stop receive */
	phx_stop_rx_endpoint(is, ep->rx_ep);
	phx_reset_rx_endpoint(is, ep->rx_ep);

	phx_apu_channel_reset(is, eth->apu_data_ch);
	phx_apu_channel_reset(is, eth->apu_desc_ch);

	phx_release_fpga_endpoint(is, ep->id)

	mx_spin_lock_destroy(&eth->spinlock);
	/* Free both send and receive rings */
	free_rx_ring(eth);
	free_tx_ring(eth);
	is->ether_is_open = 0;
}

void
phx_ether_set_mtu(struct phx_ether *eth, int mtu)
{
	eth->mtu = mtu;
	eth->is->mac_ops.mac_set_max_frame_length(eth->is, mtu);
}

/* Pointer to the first 16B of the next receive packet */
static int
phx_check_rx_frame_length(struct phx_ether *eth, uint32_t pkt_len)
{
	phx_ether_pages_t *pages = &eth->rx_data.ring.gamma;
	uint32_t start_off = eth->rx_data.data_off;
	uint32_t p_idx, p_off, mtu = eth->mtu + 14;
	char *pkt;

	/* First, peek into the first 16B and check if the packet is
	 * VLAN tagged.  If so, increase mtu by 4.
	 */

	/* Start page */
	p_idx = (start_off / PAGE_SIZE) % pages->n_pages;
	p_off = (start_off & (PAGE_SIZE-1));

	pkt = (char*)(pages->pages[p_idx].va + p_off);
#if MAL_CPU_BIGENDIAN
#define ETHER_TYPE_VLAN 0x8100
#else
#define ETHER_TYPE_VLAN 0x0081
#endif
	if (*((uint16_t*)(pkt + 12)) == ETHER_TYPE_VLAN)
		mtu += 4;
	if (pkt_len > mtu)
		return -1;
	return 0;
}

void
phx_ether_process_rx_desc(struct phx_ether *eth, phx_rx_desc_t *rx_desc)
{
	uint32_t pkt_len, total_length;
	int i, pkt_cnt, mac_error, drop;

#if MX_DEBUG_ETHER
	/* Have a new descriptor */
	MX_INFO(("new descriptor. seqnum=%u pkt_cnt=%u"
		 " tstamp_high=%u tstamp_low=%u"
		 " pkt1_length=%u pkt1_tag=%u\n",
		 rx_desc->seqnum, rx_desc->pkt_cnt,
		 rx_desc->timestamp_high, rx_desc->timestamp_low,
		 rx_desc->pkt1_length, rx_desc->pkt1_tag));
#endif

	eth->ep_tx_stats->rx_packets++;

	/* Flags are per-descriptor, not per-packet. */
	mac_error = !!(rx_desc->desc_flags & PHX_DESC_FLAG_MAC_ERROR);
	if (mac_error)
		eth->ep_tx_stats->rx_packets_mac_error++;
	drop = mac_error;

	/* The first packet */
	pkt_len = rx_desc->pkt1_length;
#ifdef PHX_WORKAROUND_MAC_OVERSIZE_FRAMES
	if (!drop && phx_check_rx_frame_length(eth, pkt_len)) {
		/* Check the frame size only if the packet does not have
		 * MAC errors, to avoid double counting.
		 */
		eth->stats_dropped_oversized++;
		drop = 1;
	}
#endif
	if (!drop)
		phx_ether_receive_packet(eth, pkt_len,
					 rx_desc->timestamp_high,
					 rx_desc->timestamp_low);

	total_length = phx_ether_rx_advance_data_offset(eth, pkt_len);

	/* Possibly remaining packets */
	pkt_cnt = rx_desc->pkt_cnt - 1;
	for (i = 0; i < pkt_cnt; i++) {
		/* Additional packets start from the bottom of pkts[] */
		int pkt_idx = PHX_RX_DESC_PKT_MAX - 2 - i;
		pkt_len = rx_desc->pkts[pkt_idx].length;
		eth->ep_tx_stats->rx_packets++;

		/* Per-descriptor MAC error bit applies to all packets */
		if (mac_error){
			eth->ep_tx_stats->rx_packets_mac_error++;
			drop = 1;
		}

#ifdef PHX_WORKAROUND_MAC_OVERSIZE_FRAMES
		if (!drop && phx_check_rx_frame_length(eth, pkt_len)) {
			eth->stats_dropped_oversized++;
			drop = 1;
		}
#endif
		if (!drop) {
			uint32_t ts_seconds, ts_nanos;
			ts_seconds = rx_desc->timestamp_high;
			ts_nanos = rx_desc->pkts[pkt_idx].nano_timestamp;
#ifdef PHX_WORKAROUND_PHXSNF_26
			if ((int)ts_nanos < 0)
				ts_nanos += 1000000000; /* overflow */
#endif
			ts_nanos += rx_desc->timestamp_low;
			if (ts_nanos > 1000000000) {
				ts_nanos -= 1000000000;
				ts_seconds += 1;
			}
			phx_ether_receive_packet(eth, pkt_len,
						 ts_seconds, ts_nanos);
		}
		total_length += phx_ether_rx_advance_data_offset(eth, pkt_len);
	}

	phx_ether_consumed_rx_desc(eth);

	/* Flow control: return data/desc pages back to the NIC */
	phx_ether_rx_advance_cons(eth);
}

int
phx_ether_rx_done(mx_instance_state_t *is)
{
	struct phx_ether *eth = is->ether;
	phx_rx_desc_t *rx_desc;
	uint8_t seqnum;
	int count = 0;
	static int broken = 0;
	int limit = eth->rx_done_limit;

	rx_desc = phx_ether_get_rx_desc_from_ring(eth);
	while ((seqnum = *((volatile uint8_t*)&rx_desc->seqnum)) == eth->rx_desc.seqnum) {
		phx_ether_process_rx_desc(eth, rx_desc);

		/* Next descriptor. */
		rx_desc = phx_ether_get_rx_desc_from_ring(eth);
		count++;
		if (limit > 0 && count >= limit)
			break;
	}

	/* if we get woken up to process packets, but don't find any, this may indicate a flooded descriptor ring */
	/* the outcome of that is an interrupt storm that will never end  */
	if (count == 0) {
		if (broken < 10){
		    broken++;
		} else {
			MX_WARN(("FATAL ERROR: Disabling RX interrupts on kernel endpoint\n"));
			MX_WARN((" rx desc seq num = %d  expecting  %d\n", seqnum, eth->rx_desc.seqnum));
			goto dont_interrupt;    /* do not reenable interrupts..stops the madness */
		}
	}
	else
		broken = 0;

	if (eth->use_interrupt) {
		/* Request interrupt for future packets */
		phx_rx_wake_request(eth->is, eth->myri_phx(ep)->rx_ep,
				    eth->rx_desc.desc_idx);
		eth->rx_desc.stat_wake_req_idx = eth->rx_desc.desc_idx;
	}
dont_interrupt:
	return count;
}

void
phx_ether_clean_sent_packets(struct phx_ether *eth)
{
	phx_ether_tx_t *tx = &eth->myri_phx(tx);
	phx_ether_tx_queued_packet_t *info;
	uint32_t sent;

	sent = tx->completion->endpoint[eth->myri_phx(ep)->tx_ep].sent;

	eth->debug_last_tx_completion = sent;
	if (sent != tx->desc_idx_nic) {
#if MX_DEBUG_ETHER
		MX_INFO(("send completion. new_idx=%u old_idx=%u\n",
			 sent, tx->desc_idx_nic));
#endif
		/* Since we copy packet to DMA or PIO buffer, we do not need
		 * to keep skbs or NBLs around.	Only need lengths.
		 */
		while (tx->desc_idx_nic != sent) {
			info = &tx->queued[(tx->desc_idx_nic % tx->desc_count)];
			if ((info->flags & PHX_TX_DESC_FLAG_DMA))
				tx->dma_offset_nic = info->end_offset;
			else
				tx->pio_offset_nic = info->end_offset;
			tx->desc_idx_nic++;
		}
	}
}

void
phx_ether_irq(mx_instance_state_t *is, uint32_t intr_flags)
{
#if MX_DEBUG_ETHER
	MX_INFO(("%s: intr_flags=0x%x\n", __FUNCTION__, intr_flags));
#endif
	is->ether->stat_irq++;
	if ((intr_flags & PHX_INTR_ENDPT_WAKE_RX)) {
		is->ether->stat_rx_irq++;
#ifdef HAVE_NEW_NAPI
		napi_schedule(&is->ether->arch.napi);
#else
		phx_ether_rx_done(is);
#endif
	}

	if ((intr_flags & PHX_INTR_ENDPT_WAKE_TX)) {
		is->ether->stat_tx_irq++;
		phx_ether_tx_done(is);
	}
}

void
phx_ether_set_mac_address_common(struct phx_ether *eth, uint8_t *addr)
{
	mx_instance_state_t *is = eth->is;

	/* copy new mac address */
	memcpy(is->mac_addr, addr, 8);
	is->board_ops.set_mac_addr(is, is->mac_addr);
}

int
phx_ether_set_promisc_common(struct phx_ether *eth, int promisc)
{
	mx_instance_state_t *is = eth->is;
	/* If sniffer is running, we do not want to disturb.  Let the sniffer code decide
	 * what to do with promisc requests.
	 */
	phx_snf_set_ether_promisc(is, promisc);

	return 0;
}

int
phx_ether_set_pause_common(struct phx_ether *eth, int pause)
{
	/* MX_WARN(("%s %d: Phoenix: unimpl\n", __FUNCTION__, __LINE__)); */
	return 0;
}

void
phx_ether_rx_advance_cons(struct phx_ether *eth)
{
	int advanced_bytes;
	uint32_t pages;

	uint32_t min_pages;
	int desc_size;
	/* Should use 16 for both? K35SNF needs a throttling. Otherwise, it may
	 * crash the chip.
	 */
	min_pages = (ETHER_K35SNF(eth)) ? 16 : 1;

	/* Data ring.  Move pages, not individual bytes */
	advanced_bytes = (int)(eth->rx_data.data_off - eth->rx_data.cons_off);
	pages = (uint32_t)advanced_bytes / PAGE_SIZE;
	if (pages >= min_pages) {
		phx_rx_data_cons_advance(eth->is, eth->ep.phx->rx_ep, pages);
		eth->rx_data.cons_off += (pages * PAGE_SIZE);
#if MX_DEBUG_ETHER
		MX_INFO(("%s: advance data cons (pages)=%u cons_off=%u data_off=%u\n",
			 __FUNCTION__, pages, eth->rx_data.cons_off, eth->rx_data.data_off));
#endif
	}

	/* Now descriptor ring */
	advanced_bytes = ((int)(eth->rx_desc.desc_idx - eth->rx_desc.cons_idx));
	desc_size = ETHER_K35SNF(eth) ? sizeof(k35snf_rx_desc_t) : sizeof(phx_rx_desc_t);
	advanced_bytes *= desc_size;
	pages = (uint32_t)advanced_bytes / PAGE_SIZE;
	if (pages >= min_pages) {
		phx_rx_desc_cons_advance(eth->is, eth->ep.phx->rx_ep, pages);
		eth->rx_desc.cons_idx += (pages * PAGE_SIZE / desc_size);
#if MX_DEBUG_ETHER
		MX_INFO(("%s: advance desc cons (pages)=%u cons_idx=%u data_idx=%u\n",
			 __FUNCTION__, pages, eth->rx_desc.cons_idx, eth->rx_desc.desc_idx));
#endif
	}

	/* FIXME.  Ugly calculations.  Use byte offset throughout or something easier to understand */
}

int
phx_ether_enable_all_multicast(struct phx_ether *eth)
{
	mx_instance_state_t *is = eth->is;
	return is->board_ops.enable_all_multicast(is);
}

int
phx_ether_disable_all_multicast(struct phx_ether *eth)
{
	mx_instance_state_t *is = eth->is;
	return is->board_ops.disable_all_multicast(is);
}

int
phx_ether_leave_all_multicast(struct phx_ether *eth)
{
	mx_instance_state_t *is = eth->is;
	struct myri_rx_filterq_entry *entry;
	int rc;

	/* Remove all multicast filters using our list */
	while (!TAILQ_EMPTY(&eth->ether_mc_filter_list)) {
		entry = TAILQ_FIRST(&eth->ether_mc_filter_list);
		entry->filt.remove = 1;
		rc = is->board_ops.remove_filter(is, entry);
		if (rc != 0) {
			MX_WARN(("Failed to remove multicast filter!\n"));
			return rc;
		}

		TAILQ_REMOVE(&eth->ether_mc_filter_list, entry, entries);
		mx_kfree(entry);
	}

	return 0;
}

int
phx_ether_join_multicast(struct phx_ether *eth, uint32_t addr_hi, uint32_t addr_lo)
{
	mx_instance_state_t *is = eth->is;
	struct myri_rx_filterq_entry *entry;
	int rc;

	/* allocate the new filter */
	entry = mx_kmalloc(sizeof(*entry), MX_WAITOK|MX_MZERO);
	if (entry == NULL) {
		MX_WARN(("Failed to allocate filter for ethernet device due to lack of memory!\n"));
		return -ENOMEM;
	}

        entry->data[0] = addr_hi;
        entry->data[1] = addr_lo;

	rc = is->board_ops.add_filter(is, entry);
	if (rc != 0) {
		mx_kfree(entry);
		if (rc == -EINVAL){
			rc = 0;  /* we did not create an ipv4 filter for it, but OK */
		}
		else
			MX_WARN(("Failed to add multicast filter!\n"));
	}
	else {
		/* keep track of filter */
		TAILQ_INSERT_TAIL(&eth->ether_mc_filter_list, entry, entries);
	}

	return rc;
}

void
k35snf_ether_irq(mx_instance_state_t *is, uint32_t intr_flags)
{
	/* MX_INFO(("k35snf_ether_irq: intr_flags=0x%x\n", intr_flags)); */

	/* We have separate RX and TX interrupts.  Their handlers may run
	 * concurrently.  So we should not assume only one interrupt thread
	 * runs and take care to lock where necessary.
	 */

	if (ep_enter(&is->ether->arch, EP_FLAGS_IRQ_ENTER))
		return;

	if ((intr_flags & PHX_INTR_ENDPT_WAKE_RX)) {
#ifdef HAVE_NEW_NAPI
		napi_schedule(&is->ether->arch.napi);
#else
		k35snf_ether_rx_done(is);
#endif
		goto exit;
	}

	if ((intr_flags & PHX_INTR_ENDPT_WAKE_TX)) {
		k35snf_ether_tx_done(is);
		goto exit;
	}

exit:
	ep_leave(&is->ether->arch, EP_FLAGS_IRQ_ENTER);
}

void
k35snf_ether_clean_sent_packets(struct phx_ether *eth)
{
	k35snf_ether_tx_t *tx = &eth->tx.k35snf;
	uint32_t packets, bytes;

	packets = tx->completion.count->packets;
	bytes = tx->completion.count->bytes;
	tx->desc_idx_nic += packets;
	tx->data_offset_nic += bytes;
	/*
	MX_INFO(("k35snf_ether_clean_sent_packets: packets=%u,%u"
		 " bytes=%u,%u\n",
		 packets, tx->desc_idx_nic, bytes, tx->data_offset_nic));
	*/
	tx->completion.count->packets = 0;
	tx->completion.count->bytes = 0;
	eth->debug_last_tx_completion_packets = packets;
	eth->debug_last_tx_completion_bytes = bytes;
}

void
k35snf_ether_process_rx_desc(struct phx_ether *eth, k35snf_rx_desc_t *desc)
{
	uint32_t pkt_len;
	int drop;
	/*
	MX_INFO(("k35snf_ether_process_rx_desc: seqnum=%u"
		 " tstamp=%llu pktlen=%u tag=%u flags=%u\n",
		 desc->seqnum, desc->timestamp,
		 desc->pktlen, desc->tag, desc->flags));
	*/
	eth->ep_tx_stats->rx_packets++;

	pkt_len = desc->pktlen;
	drop = 0;
#ifdef PHX_WORKAROUND_MAC_OVERSIZE_FRAMES
	if (phx_check_rx_frame_length(eth, pkt_len)) {
		/* Check the frame size only if the packet does not have
		 * MAC errors, to avoid double counting.
		 */
		eth->stats_dropped_oversized++;
		drop = 1;
	}
#endif
	if (!drop)
		k35snf_ether_receive_packet(eth, pkt_len, desc->timestamp);
	/* When advancing the head, we want the whole area occupied by this
	 * packet.
	 */
	k35snf_ether_rx_advance_data_offset(eth, pkt_len);
	phx_ether_consumed_rx_desc(eth); /* same as PHX */
	/* Flow control: return data/desc pages back to the NIC */
	phx_ether_rx_advance_cons(eth); /* same as PHX */
}

int
k35snf_ether_open_common(mx_instance_state_t *is, int mtu,
			 int rx_data_ring_size_mb, int rx_desc_ring_size_mb)
{
	struct phx_ether *eth = is->ether;
	struct phx_physical_ep *rx_ep, *tx_ep;
	int status, tx_claimed;
	uint32_t desc_ring_size;
	k35snf_ether_tx_t *tx;

	eth->rx_done_limit = myri_ether_rx_done_limit;
	eth->use_interrupt = 1;
	/* Claim both RX and TX endpoints.  If we cannot, proceed and allocate
	 * rings and set flags so that we attach the endpoints when SNF apps
	 * quit.
	 */
	tx_ep = is->board_ops.physical_endpoint(is, 0);
	rx_ep = is->board_ops.physical_endpoint(is, 1);
	if ((tx_claimed = phx_claim_fpga_endpoint(is, tx_ep->id, 1)) ||
	    phx_claim_fpga_endpoint(is, rx_ep->id, 1)) {
		MX_INFO(("%s: starting the kernel Ethernet interface without"
			 " endpoints\n", is->is_name));
		if (!tx_claimed)
			phx_release_fpga_endpoint(is, tx_ep->id);
		eth->ep.k35snf.stolen = 1;
	}

	eth->ep.k35snf.tx = tx_ep;
	eth->ep.k35snf.rx = rx_ep;
	eth->ep_tx_stats = &is->phx.pfi->counters.ep_tx_stats[tx_ep->id];
	eth->ep_nic_mem = NULL;

	eth->apu_data_ch = phx_apu_channel_id(is, 1, 0 /* unused */, 0 /* not tx */);
	eth->apu_desc_ch = phx_apu_channel_id(is, 0, 0, 0);
#if MX_DEBUG_ETHER
	MX_INFO(("APU channels: data=%d desc=%d\n",
		 eth->apu_data_ch, eth->apu_desc_ch));
#endif

	/* Round up rx_data_ring_mb to a sane value. */
	eth->rx_data.ring.size = roundup_rx_ring(rx_data_ring_size_mb,
						 MYRI_RX_DATA_RING_MB_MIN,
						 MYRI_RX_DATA_RING_MB_MAX);
#if MX_DEBUG_ETHER
	MX_INFO(("%s: rx_data_ring_size=%d\n", is->is_name, eth->rx_data.ring.size));
#endif

	/* Round up rx_desc_ring_mb to a sane value. */
	eth->rx_desc.ring.size = roundup_rx_ring(rx_desc_ring_size_mb,
						 MYRI_RX_DESC_RING_MB_MIN,
						 MYRI_RX_DESC_RING_MB_MAX);
#if MX_DEBUG_ETHER
	MX_INFO(("%s: rx_desc_ring_size=%d\n", is->is_name, eth->rx_desc.ring.size));
#endif

	/* Allocate receive ring: data and descriptor */
	if ((status = alloc_rx_ring(eth))) {
		MX_WARN(("%s: unable to allocate receive ring\n", is->is_name));
		goto abort_with_nothing;
	}

	/* Now TX rings */
	tx = &eth->tx.k35snf;
	tx->prod_doorbell = (volatile uint32_t*)
		((uintptr_t)is->phx.nic_mem + K35SNF_TX_BASE +
		 K35SNF_TX_DESC_PRODW_ADVANCE);

	/* Allocate the data ring.  FIXME: make size tunable. */
	tx->data_ring_size = (2 << 20);
	tx->data_offset_host = 0;
	tx->data_offset_nic = 0;
	status = phx_alloc_rx_ring(is, &tx->data_ring, tx->data_ring_size, 0);
	if (status) {
		MX_WARN(("%s: failed to allocate the TX data ring\n", is->is_name));
		status = ENOMEM;
		goto abort_with_receive_ring;
	}

	/* Allocate the descriptor ring */
	desc_ring_size = (2 << 20); /* FIXME */
	tx->desc_count = desc_ring_size / sizeof(k35snf_tx_desc_t);
	tx->desc_idx_host = 0;
	tx->desc_idx_nic = 0;
	status = phx_alloc_rx_ring(is, &tx->desc_ring, desc_ring_size, 0);
	if (status) {
		MX_WARN(("%s: failed to allocate SNF TX descriptor ring\n", is->is_name));
		status = ENOMEM;
		goto abort_with_tx_data_ring;
	}

	/* Completion page */
	status = mx_alloc_zeroed_dma_page_relax_order(is, &tx->completion.addr,
						  &tx->completion.pin);
	if (status) {
		MX_WARN(("%s: failed to allocate TX completion page\n", is->is_name));
		status = ENOMEM;
		goto abort_with_tx_desc_ring;
	}
	tx->completion.count = (volatile k35snf_tx_completion_t*)tx->completion.addr;

	/* Set up APUs */
	tx->apu_data_ch = phx_apu_channel_id(is, 1, 0, 1);
	tx->apu_desc_ch = phx_apu_channel_id(is, 0, 0, 1);
	MX_INFO(("%s: TX APU channels: data=%d desc=%d\n", is->is_name,
		 tx->apu_data_ch, tx->apu_desc_ch));

	phx_ether_set_mtu(eth, mtu);

	mx_spin_lock_init(&eth->spinlock, is, -1, "ether spinlock");
	is->ether_is_open = 1;
	return 0;

abort_with_tx_desc_ring:
	phx_free_rx_ring(is, &tx->desc_ring);
abort_with_tx_data_ring:
	phx_free_rx_ring(is, &tx->data_ring);
abort_with_receive_ring:
	free_rx_ring(eth);
abort_with_nothing:
	if (!eth->ep.k35snf.stolen) {
		phx_release_fpga_endpoint(is, tx_ep->id);
		phx_release_fpga_endpoint(is, rx_ep->id);
	}
	return status;
}

int
k35snf_ether_start_common(struct phx_ether *eth)
{
	mx_instance_state_t *is = eth->is;
	k35snf_ether_tx_t *tx = &eth->tx.k35snf;
	mx_page_pin_t *page;
	phx_ether_pages_t *p;
	int i;

	if (eth->ep.k35snf.stolen)
		return 0;

	/* Zero DMA buffers */
	memset(tx->completion.addr, 0, PAGE_SIZE);
	p = &eth->rx_data.ring.gamma;
	for (i = 0; i < p->n_pages; i++)
		memset((void*)(uintptr_t)p->pages[i].va, 0, PAGE_SIZE);
	p = &eth->rx_desc.ring.gamma;
	for (i = 0; i < p->n_pages; i++)
		memset((void*)(uintptr_t)p->pages[i].va, 0, PAGE_SIZE);
	p = &tx->data_ring.gamma;
	for (i = 0; i < p->n_pages; i++)
		memset((void*)(uintptr_t)p->pages[i].va, 0, PAGE_SIZE);
	p = &tx->desc_ring.gamma;
	for (i = 0; i < p->n_pages; i++)
		memset((void*)(uintptr_t)p->pages[i].va, 0, PAGE_SIZE);

	/* Reset pointers */
	eth->rx_data.data_off = 0;
	eth->rx_data.cons_off = 0;
	eth->rx_desc.desc_idx = 0;
	eth->rx_desc.cons_idx = 0;
	eth->rx_desc.seqnum = 1;
	tx->desc_idx_host = 0;
	tx->desc_idx_nic = 0;
	tx->data_offset_host = 0;
	tx->data_offset_nic = 0;

	/* RX endpoint */
	MX_INFO(("%s: starting RX endpoint\n", is->is_name));
	phx_stop_rx_endpoint(is, 0 /* unused */);
	phx_apu_channel_reset(is, eth->apu_data_ch);
	phx_apu_channel_reset(is, eth->apu_desc_ch);
	page = eth->rx_data.ring.alpha.pages;
	phx_apu_channel_start(is, eth->apu_data_ch, page->dma.low, page->dma.high);
	page = eth->rx_desc.ring.alpha.pages;
	phx_apu_channel_start(is, eth->apu_desc_ch, page->dma.low, page->dma.high);
	phx_reset_rx_endpoint(is, 0 /* unused */);
	phx_start_rx_endpoint(is, 0 /* unused */,
			      eth->rx_data.ring.gamma.n_pages,
			      eth->rx_desc.ring.gamma.n_pages);

	/* TX endpoint */
	MX_INFO(("%s: starting TX endpoint\n", is->is_name));
	phx_apu_channel_reset(is, tx->apu_data_ch);
	phx_apu_channel_reset(is, tx->apu_desc_ch);
	page = tx->data_ring.alpha.pages;
	phx_apu_channel_start(is, tx->apu_data_ch, page->dma.low, page->dma.high);
	page = tx->desc_ring.alpha.pages;
	phx_apu_channel_start(is, tx->apu_desc_ch, page->dma.low, page->dma.high);
	k35snf_tx_start(is, tx->completion.pin.dma.low,
			tx->completion.pin.dma.high,
			tx->data_ring.gamma.n_pages,
			tx->desc_count);
	return 0;
}

void
k35snf_ether_close_common(struct phx_ether *eth)
{
	mx_instance_state_t *is = eth->is;
	k35snf_ether_tx_t *tx = &eth->tx.k35snf;

	MX_INFO(("k35snf_ether_close_common\n"));

	if (!eth->ep.k35snf.stolen) {
		/* Stop receive */
		phx_stop_rx_endpoint(is, 0 /* unused */);
		phx_reset_rx_endpoint(is, 0 /* unused */);

		phx_apu_channel_reset(is, eth->apu_data_ch);
		phx_apu_channel_reset(is, eth->apu_desc_ch);

		/* Stop send */
		phx_apu_channel_reset(is, tx->apu_data_ch);
		phx_apu_channel_reset(is, tx->apu_desc_ch);
		k35snf_tx_stop(is);
	}

	/* Free everything */
	mx_spin_lock_destroy(&eth->spinlock);

	free_rx_ring(eth);
	if (tx->completion.addr)
		mx_free_dma_page(is, &tx->completion.addr, &tx->completion.pin);
	phx_free_rx_ring(is, &tx->data_ring);
	phx_free_rx_ring(is, &tx->desc_ring);

	is->ether_is_open = 0;

	/* Release the endpoints as the last thing.  SNF would start accessing
	 * APU and such as soon as it claims the endpoints.  See phx_snf_ops.c
	 */
	if (!eth->ep.k35snf.stolen) {
		phx_release_fpga_endpoint(is, eth->ep.k35snf.rx->id);
		phx_release_fpga_endpoint(is, eth->ep.k35snf.tx->id);
	}
	eth->ep.k35snf.stolen = 0;
}

int
k35snf_ether_tx_copy_pkt(struct phx_ether *eth, phx_ether_pkt_frag_t frags[],
			 int n_frags, uint32_t len, uint32_t buf_offset)
{
	k35snf_ether_tx_t *tx = &eth->tx.k35snf;
	phx_ether_pages_t *pages;
	mx_page_pin_t *p;
	uint32_t p_idx, p_off, frag_off;
	int i, err;
	/*
	MX_INFO(("k35snf_ether_tx_copy_pkt n_frags=%d len=%d buf_offset=%d\n",
		 n_frags, len, buf_offset));
	*/
	err = 0;
	i = 0;
	frag_off = 0;

	/* DMA buffer: start page */
	pages = &tx->data_ring.gamma;
	p_idx = (buf_offset / PAGE_SIZE) % pages->n_pages;
	p_off = (buf_offset & (PAGE_SIZE-1));

	/* Now copy.  Both the packet and the DMA buffer may span across
	 * multiple frags or pages.
	 */
	while (len > 0 && i < n_frags) {
		int rem, copy_len;

		/* Want to copy min(len, frag, page) */
		/* Check frag and len */
		rem = frags[i].len - frag_off;
		copy_len = (len > rem) ? rem : len;
		/* Check page */
		p = &pages->pages[p_idx];
		rem = PAGE_SIZE - p_off;
		if (copy_len > rem)
			copy_len = rem;

		/* Copy from frags to page */
		memcpy((char*)(p->va + p_off), frags[i].vaddr + frag_off, copy_len);

		len -= copy_len;
		/* Advance frag pointer */
		frag_off += copy_len;
		if (frag_off == frags[i].len) {
			i++;
			frag_off = 0;
		}
		/* Advance page pointer */
		p_off += copy_len;
		if (p_off == PAGE_SIZE) {
			p_idx = (p_idx + 1) % pages->n_pages;
			p_off = 0;
		}
	}
	if (len > 0)
		err = -1;
	return err;
}

#if MYRI_ENABLE_PTP
#error "PTP not supported yet"
#endif

/*
  This file uses MX driver indentation.

  Local Variables:
  c-file-style:"linux"
  tab-width:8
  indent-tabs-mode:t
  End:
*/
