/************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET		*
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the	*
 * "License"); User may not use this file except in compliance with the	*
 * License.  The full text of the License can found in LICENSE.TXT	*
 *									*
 * Software distributed under the License is distributed on an "AS IS"	*
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See	*
 * the License for the specific language governing rights and		*
 * limitations under the License.					*
 *									*
 * Copyright 2011 by Myricom, Inc.  All rights reserved.		*
 ************************************************************************/

#include <linux/version.h>

#ifndef pr_fmt
#define pr_fmt(fmt)			KBUILD_MODNAME ": " fmt
#endif

#ifndef pr_err
#define  pr_err(fmt, ...)		printk(KERN_ERR pr_fmt(fmt), ##__VA_ARGS__)
#endif

#ifndef netdev_err
#define netdev_err(dev, fmt, ...)	printk(KERN_ERR "%s: " fmt, dev->name, ##__VA_ARGS__)
#endif

#ifndef netdev_info
#define netdev_info(dev, fmt, ...)	printk(KERN_INFO "%s: " fmt, dev->name, ##__VA_ARGS__)
#endif

#ifndef netif_info
#define netif_info(priv, type, dev, fmt, args...)	\
do {							\
	if (netif_msg_##type(priv))			\
		netdev_info((dev), fmt, ##args);	\
} while (0)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
#define cancel_delayed_work_sync(work)	cancel_rearming_delayed_work(work)

#define alloc_etherdev_mq(size, count)	alloc_etherdev(size)
#define NUM_TX_QUEUES(num)		1
#else
#define NUM_TX_QUEUES(num)		num
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
#define skb_get_queue_mapping(skb)	0
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30)
#define skb_record_rx_queue(skb, q)	do { } while (0)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,35)
#define netif_set_real_num_tx_queues(dev, txq)	do { } while (0)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,37)
#define netif_set_real_num_rx_queues(dev, rxq)	do { } while (0)
#endif

#ifndef CHECKSUM_PARTIAL
#define CHECKSUM_PARTIAL		CHECKSUM_HW
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static inline int myri_skb_checksum_help(struct sk_buff *skb)
{
	return skb_checksum_help(skb, 0);
}

#define skb_checksum_help(skb)		myri_skb_checksum_help(skb)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
#define delayed_work 			work_struct
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
#ifdef CONFIG_NUMA
#define cpumask_of_pcibus(bus)		&pcibus_to_cpumask(bus)
#else
#define cpumask_of_pcibus(bus)		&cpu_online_map
#endif
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
#define cpumask_weight(mask)		cpus_weight(*mask)
#define cpumask_copy(dst, src)		cpus_or(*dst, *src, *src)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
#define dev_getbyhwaddr(a, b, c)	dev_getbyhwaddr(b, c)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38)
#define dev_getbyhwaddr_rcu(a, b, c)	dev_getbyhwaddr(a, b, c)
#endif

/* newer kernels stopped defining the HAVE_NET* for netdevice changes as
 * the transition to the new structures is now considered complete. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
#define HAVE_NET_DEVICE_OPS
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,2,0)
#define skb_frag_address(frag) (page_address(frag->page) + frag->page_offset)
#endif
