/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2014 by CSP, Inc.  All rights reserved.                     *
 *************************************************************************/

#ifndef _phx_ether_h_
#define _phx_ether_h_

#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/string.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/etherdevice.h>
#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/pci.h>
#include <linux/if_ether.h>
#include <linux/if_vlan.h>
#include <linux/ip.h>
#include <linux/inet.h>
#include <linux/ethtool.h>
#include <linux/in.h>
#include <net/checksum.h>

struct phx_ether_arch {
	struct net_device *dev;
	struct net_device_stats stats;
#ifdef HAVE_NEW_NAPI
	struct napi_struct napi;
#endif
	struct task_struct *kthread;
	/* Use this to synchronize between xmit() and tx_done().
	 * See comments in xmit().
	 */
	spinlock_t tx_intr_lock;
	unsigned long stats_tx_false_fullness;
	unsigned long stats_tx_try_again;
	unsigned long stats_tx_stop_queue;
	unsigned long stats_tx_wake_queue;
	unsigned long stats_napi_complete;
	unsigned long stats_tx_timestamp;
	unsigned long stats_no_ep_tx_wake_queue;
	unsigned long stats_no_ep_tx_dropped;
	unsigned long stats_no_ep_napi_complete;
	struct sk_buff *tx_ts_skb;  /* only support one outstanding 
				     tx timestamp packet at a time for now */

	spinlock_t ep_lock;
	unsigned long ep_irqsave;
	int ep_flags;
#define EP_FLAGS_REMOVE_ENTER     (1 << 0)
#define EP_FLAGS_TX_ENTER         (1 << 1)
#define EP_FLAGS_IRQ_ENTER        (1 << 2)
#define EP_FLAGS_POLL_ENTER       (1 << 3)
#define EP_FLAGS_OPEN_ENTER       (1 << 4)
#define EP_FLAGS_CLOSE_ENTER      (1 << 5)
#define EP_FLAGS_RETURN_ENTER     (1 << 6)
#define EP_FLAGS_REMOVED          (1 << 7)
};

static inline int
ep_flags_isset(struct phx_ether_arch *arch, int flag)
{
	/* True if any flag bit is set */
	return !!(arch->ep_flags & flag);
}

static inline void
ep_flags_set(struct phx_ether_arch *arch, int flag)
{
	arch->ep_flags |= flag;
}

static inline void
ep_flags_clear(struct phx_ether_arch *arch, int flag)
{
	arch->ep_flags &= ~flag;
}

static inline void
ep_lock(struct phx_ether_arch *arch)
{
	spin_lock_irqsave(&arch->ep_lock, arch->ep_irqsave);
}

static inline void
ep_unlock(struct phx_ether_arch *arch)
{
	spin_unlock_irqrestore(&arch->ep_lock, arch->ep_irqsave);
}

static inline int
ep_enter(struct phx_ether_arch *arch, int flag)
{
	int stopped = 1;

	ep_lock(arch);
	if (!ep_flags_isset(arch, EP_FLAGS_REMOVE_ENTER | EP_FLAGS_REMOVED)) {
		ep_flags_set(arch, flag);
		stopped = 0;
	}
	ep_unlock(arch);
	return stopped;
}

static inline void
ep_enter_wait(struct phx_ether_arch *arch, int clear_flag, int set_flag)
{
	ep_lock(arch);
	while ((arch->ep_flags & clear_flag)) {
		ep_unlock(arch);
		msleep_interruptible(1);
		ep_lock(arch);
	}
	ep_flags_set(arch, set_flag);
	ep_unlock(arch);
}

static inline void
ep_leave(struct phx_ether_arch *arch, int flag)
{
	ep_lock(arch);
	ep_flags_clear(arch, flag);
	ep_unlock(arch);
}

 /* newer kernels stopped defining the HAVE_NET* for netdevice changes as
 * the transition to the new structures is now considered complete. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
#define HAVE_NET_DEVICE_OPS
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#define phx_netdev_priv(d) *(struct phx_ether **)(netdev_priv(d))
#else
#define phx_netdev_priv(d) ((d)->priv)
#endif

#endif /* _phx_ether_h_ */

/*
  This file uses MX driver indentation.

  Local Variables:
  c-file-style:"linux"
  tab-width:8
  indent-tabs-mode:t
  End:
*/
